/**
 * @file      sensor_focus_wnd.cpp
 * @brief     Window of SensorFocus plugin.
 * @author    Vision Processing Community
 * @copyright 2016 Vision Processing Community. All rights reserved.
 */

#include "./sensor_focus_wnd.h"
#include "./sensor_focus.h"

BEGIN_EVENT_TABLE(SensorFocusWnd, wxFrame)
EVT_CLOSE(SensorFocusWnd::OnClose)
EVT_BUTTON(kButtonInitId, SensorFocusWnd::OnInit)
EVT_BUTTON(kButtonDemoId, SensorFocusWnd::OnInitDemo)
EVT_BUTTON(kButtonFocusModeApplyId, SensorFocusWnd::OnFocusModeApply)
EVT_BUTTON(kButtonOISModeApplyId, SensorFocusWnd::OnOISModeApply)
EVT_BUTTON(kButtonOISModeDemoStartId, SensorFocusWnd::OnOISModeDemoStart)
EVT_BUTTON(kButtonOISModeDemoStopId, SensorFocusWnd::OnOISModeDemoStop)
END_EVENT_TABLE();

// I2C definition
extern "C" {
int __CCIRegReadBySlaveAddress(int CCISlaveAddress, int RegAddress, int *data);
int __CCIRegWriteBySlaveAddress(int CCISlaveAddress, int RegAddress, int data);
}

/**
 * @brief
 * Constructor for this window.
 */
SensorFocusWnd::SensorFocusWnd(SensorFocus *parent)
    : wxFrame(NULL, kWndId, wxT(kWndTitle), wxPoint(kWndPointX, kWndPointY),
              wxSize(kWndSizeW, kWndSizeH)) {
  // Initialize
  parent_ = parent;

  // Create Radio box
  wxString focus_tbl[3];
  wxString ois_tbl[7];

  // Create Init button
  button_image_Init_apply_ = new wxButton(
      this, kButtonInitId, wxT(kButtonInitName),
      wxPoint(kButtonInitPointX, kButtonInitPointY),
      wxSize(kButtonInitSizeX, kButtonInitSizeY));

  // Create Demo button
  button_image_InitDemo_apply_ = new wxButton(
      this, kButtonDemoId, wxT(kButtonDemoName),
      wxPoint(kButtonDemoPointX, kButtonDemoPointY),
      wxSize(kButtonDemoSizeX, kButtonDemoSizeY));
  // Create focus mode TBL
  focus_tbl[kFocusMode1] = wxT("Direct");
  focus_tbl[kFocusMode2] = wxT("Inf");
  focus_tbl[kFocusMode3] = wxT("Macro");
  radio_box_focus_mode_ = new wxRadioBox(
      this, kRboxFocusModeId, wxT("Focus Mode"),
      wxPoint(kRadioFocusModePointX, kRadioFocusModePointY),
      wxSize(kRadioFocusModeSizeW, kRadioFocusModeSizeH), 3, focus_tbl, 0,
      wxRA_SPECIFY_ROWS, wxDefaultValidator, wxT(kRadioFocusModeText));

  static_text_focus_code_ = new wxStaticText(
      this, kStaticTextFocusCodeId, wxT(kStaticTextFocusCodeName),
      wxPoint(kStaticTextFocusCodePointX, kStaticTextFocusCodePointY),
      wxSize(kStaticTextFocusCodeSizeW, kStaticTextFocusCodeSizeH));

  text_ctrl_focus_code_ = new wxTextCtrl(
      this, kTextFocusCodeId, wxT("0x0000"),
	  wxPoint(kTextFocusCodePointX,kTextFocusCodePointY),
	  wxSize(kTextFocusCodeSizeW, kTextFocusCodeSizeH),
	  wxTE_PROCESS_TAB | wxTE_LEFT);

  // Create Apply button
  button_image_focus_apply_ = new wxButton(
      this, kButtonFocusModeApplyId, wxT(kButtonFocusModeApplyName),
      wxPoint(kButtonFocusModeApplyPointX, kButtonFocusModeApplyPointY),
      wxSize(kButtonFocusModeApplySizeX, kButtonFocusModeApplySizeY));

  // Create OIS mode TBL
  ois_tbl[kOisModeMode1] = wxT("OFF");
  ois_tbl[kOisModeMode2] = wxT("Zero shutter");
  ois_tbl[kOisModeMode3] = wxT("Exposure/Shake eval.");
  ois_tbl[kOisModeMode4] = wxT("Movie");
  ois_tbl[kOisModeMode5] = wxT("High SR Movie");
  ois_tbl[kOisModeMode6] = wxT("View Finder");
  ois_tbl[kOisModeMode7] = wxT("Demo");

  radio_box_ois_mode_ = new wxRadioBox(
      this, kRboxOISModeId, wxT("OIS Mode"),
      wxPoint(kRadioOISModePointX, kRadioOISModePointY),
      wxSize(kRadioOISModeSizeW, kRadioOISModeSizeH), 7, ois_tbl, 0,
      wxRA_SPECIFY_ROWS, wxDefaultValidator, wxT(kRadioOISModeText));

  // Create Apply button
  button_image_OIS_apply_ = new wxButton(this,
	  kButtonOISModeApplyId, wxT(kButtonOISModeApplyName),
      wxPoint(kButtonOISModeApplyPointX, kButtonOISModeApplyPointY),
      wxSize(kButtonOISModeApplySizeX, kButtonOISModeApplySizeY));

  // Create Demo Start button
  button_image_OIS_DemoStart_ = new wxButton(this,
	  kButtonOISModeDemoStartId, wxT(kButtonOISModeDemoStartName),
      wxPoint(kButtonOISModeDemoStartPointX, kButtonOISModeDemoStartPointY),
      wxSize(kButtonOISModeDemoStartSizeX, kButtonOISModeDemoStartSizeY));
  // Create Demo Stop button
  button_image_OIS_DemoStop_ = new wxButton(this,
	  kButtonOISModeDemoStopId, wxT(kButtonOISModeDemoStopName),
      wxPoint(kButtonOISModeDemoStopPointX, kButtonOISModeDemoStopPointY),
      wxSize(kButtonOISModeDemoStopSizeX, kButtonOISModeDemoStopSizeY));
}

/**
 * @brief
 * Destructor for this window.
 */
SensorFocusWnd::~SensorFocusWnd() {}

/**
 * @brief
 * The handler function for EVT_CLOSE.
 */
void SensorFocusWnd::OnClose(wxCloseEvent &event) { Show(false); }

/**
 * @brief
 * Sensor Focus/OIS Init
 */
void SensorFocusWnd::OnInit(wxCommandEvent& event) {
  DEBUG_PRINT("SensorFocusWnd::OnInit\n");

  int slave_address = (0x7c) >> 1;
  int sensor_slave_address = (0x34) >> 1;

  //@OIS init
  int ois_init_address = 0x30AC;
  int ois_init_regdata = 0x01;
  if (__CCIRegWriteBySlaveAddress(sensor_slave_address, ois_init_address, ois_init_regdata) > 0) {
     printf("set: [Init:OIS] slave=%02x, address=%04x, RegData = %02x\n", sensor_slave_address, ois_init_address, ois_init_regdata);
  } else {
    printf("set: error(Init:OIS)!\n");
  }

  //@Gyro Activation
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_gyro_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_gyro_[i].adr, i2cdat_gyro_[i].dat) > 0) {
      printf("set: [Init:i2cdat_gyro] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_gyro_[i].adr, i2cdat_gyro_[i].dat);
    } else {
      printf("set: error(Init:i2cdat_gyro)!\n");
    }
  }

  //@Start download
  int start_address = 0xF010;
  int start_regdata = 0x00;
  if (__CCIRegWriteBySlaveAddress(slave_address, start_address, start_regdata) > 0) {
     printf("set: [Init:Start download] slave=%02x, address=%04x, RegData = %02x\n", slave_address, start_address, start_regdata);
  } else {
    printf("set: error(Init:Start download)!\n");
  }

  //@DownloadProgram1
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_dp1_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_dp1_[i].adr, i2cdat_dp1_[i].dat) > 0) {
      printf("set: [Init:DownloadProgram1] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_dp1_[i].adr, i2cdat_dp1_[i].dat);
    } else {
      printf("set: error(Init:DownloadProgram1)!\n");
    }
  }

  //@DownloadProgram2
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_dp2_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_dp2_[i].adr, i2cdat_dp2_[i].dat) > 0) {
      printf("set: [Init:DownloadProgram2] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_dp2_[i].adr, i2cdat_dp2_[i].dat);
    } else {
      printf("set: error(Init:DownloadProgram2)!\n");
    }
  }

  //@Calibration
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_calb_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_calb_[i].adr, i2cdat_calb_[i].dat) > 0) {
      printf("set: [Init:Calibration] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_calb_[i].adr, i2cdat_calb_[i].dat);
    } else {
      printf("set: error(Init:Calibration)!\n");
    }
  }

  //@Set OIS complete DL
  int ois_comp_address = 0xF006;
  int ois_com_regdata = 0x00;
  if (__CCIRegWriteBySlaveAddress(slave_address, ois_comp_address, ois_com_regdata) > 0) {
     printf("set: [Init:Set OIS complete DL] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_comp_address, ois_com_regdata);
  } else {
    printf("set: error(Init:Set OIS complete DL)!\n");
  }

  //@Servo on
  int servo_address = 0x6020;
  int servo_regdata = 0x01;
  if (__CCIRegWriteBySlaveAddress(slave_address, servo_address, servo_regdata) > 0) {
     printf("set: [Init:Servo on] slave=%02x, address=%04x, RegData = %02x\n", slave_address, servo_address, servo_regdata);
  }
  else {
    printf("set: error(Init:Servo on)!\n");
  }

  //@Gyro on
  int gyro_address = 0x6023;
  int gyro_regdata = 0x00;
  if (__CCIRegWriteBySlaveAddress(slave_address, gyro_address, gyro_regdata) > 0) {
     printf("set: [Init:Gyro on] slave=%02x, address=%04x, RegData = %02x\n", slave_address, gyro_address, gyro_regdata);
  } else {
    printf("set: error(Init:Gyro on)!\n");
  }

  //@GPO to Hsync
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_gpo_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(sensor_slave_address, i2cdat_gpo_[i].adr, i2cdat_gpo_[i].dat) > 0) {
      printf("set: [Init:GPO] slave=%02x, address=%04x, RegData = %02x\n", sensor_slave_address, i2cdat_gpo_[i].adr, i2cdat_gpo_[i].dat);
    } else {
      printf("set: error(Init:GPO)!\n");
    }
  }

  //@PWM frequency adjustment setting - About 4.5MHz
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_pwmfrequency_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_pwmfrequency_[i].adr, i2cdat_pwmfrequency_[i].dat) > 0) {
      printf("set: [Init:PWM frequency] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_pwmfrequency_[i].adr, i2cdat_pwmfrequency_[i].dat);
    } else {
      printf("set: error(Init:PWM frequency)!\n");
    }
  }

  // EEPROM to DriverRegister
  int regdata = 0x00;
  int regdata1 = 0x00;
  // 1. OIS Hall current X DataWidthDetai:8bit  EEPROM:1DC1h Source:60A2h
  //int ois_hall_cur_x_eeprom = 0x1DC1;
  int ois_hall_cur_x_source = 0x60A2;
  //regdata = 0x32; //No.24
  regdata = 0x37; //No.14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_hall_cur_x_eeprom, &regdata) > 0) {
  //  printf("read: [1. OIS Hall current X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_cur_x_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_hall_cur_x_source, regdata) > 0) {
       printf("set: [1. OIS Hall current X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_cur_x_source, regdata);
    }
  //}
  // 2. OIS Hall current Y DataWidthDetai:8bit EEPROM:1DC3h Source:60A3h
  //int ois_hall_cur_y_eeprom = 0x1DC3;
  int ois_hall_cur_y_source = 0x60A3;
  //regdata = 0x38; //No.24
  regdata = 0x3C; //No.14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_hall_cur_y_eeprom, &regdata) > 0) {
  //  printf("read: [2. OIS Hall current Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_cur_y_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_hall_cur_y_source, regdata) > 0) {
       printf("set: [2. OIS Hall current Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_cur_y_source, regdata);
    }
  //}
  // 3. OIS Hall offset X DataWidthDetai:10bit EEPROM:1DC4h Source:60A4h
  //int ois_hall_off_x_eeprom = 0x1DC4;
  int ois_hall_off_x_source = 0x60A4;
  int ois_hall_off_x_source1 = 0x60A5;
  //regdata = 0x02; //No.24
  //regdata1 = 0x0B; //No.24
  regdata = 0x02; //No.14
  regdata1 = 0x18; //No.14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_hall_off_x_eeprom, &regdata) > 0) {
  //  printf("read: [3. OIS Hall offset X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_off_x_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_hall_off_x_source, regdata) > 0) {
       printf("set: [3. OIS Hall offset X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_off_x_source, regdata);
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_hall_off_x_source1, regdata1) > 0) {
       printf("set: [3. OIS Hall offset X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_off_x_source1, regdata1);
    }
  //}
  // 4. OIS Hall offset Y DataWidthDetai:10bit EEPROM:1DC6h Source:60A6h
  //int ois_hall_off_y_eeprom = 0x1DC6;
  int ois_hall_off_y_source = 0x60A6;
  int ois_hall_off_y_source1 = 0x60A7;
  //regdata = 0x01; //No.24
  //regdata1 = 0xD5; //No.24
  regdata = 0x01; //No.14
  regdata1 = 0xB4; //No.14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_hall_off_y_eeprom, &regdata) > 0) {
  //  printf("read: [4. OIS Hall offset Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_off_y_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_hall_off_y_source, regdata) > 0) {
       printf("set: [4. OIS Hall offset Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_off_y_source, regdata);
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_hall_off_y_source1, regdata1) > 0) {
       printf("set: [4. OIS Hall offset Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_hall_off_y_source1, regdata1);
    }
  //}
  // 5. OIS optical offset X DataWidthDetai:12bit EEPROM:1DD0h Source:6062h
  //int ois_optical_off_x_eeprom = 0x1DD0;
  //int ois_optical_off_x_source = 0x6062;
  //regdata = 0x00;
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_optical_off_x_eeprom, &regdata) > 0) {
  //  printf("read: [5. OIS optical offset X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_optical_off_x_eeprom, regdata);
  //  if (__CCIRegWriteBySlaveAddress(slave_address, ois_optical_off_x_source, regdata) > 0) {
  //     printf("set: [5. OIS optical offset X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_optical_off_x_source, regdata);
  //  }
  //}
  // 6. OIS optical offset Y DataWidthDetai:12bit EEPROM:1DD2h Source:6062h
  //int ois_optical_off_y_eeprom = 0x1DD0;
  //int ois_optical_off_y_source = 0x6062;
  //regdata = 0x00;
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_optical_off_y_eeprom, &regdata) > 0) {
  //  printf("read: [6. OIS optical offset Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_optical_off_y_eeprom, regdata);
  //  if (__CCIRegWriteBySlaveAddress(slave_address, ois_optical_off_y_source, regdata) > 0) {
  //     printf("set: [6. OIS optical offset Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_optical_off_y_source, regdata);
  //  }
  //}
  // 7. OIS Servo loop gain X DataWidthDetai:16bit EEPROM:1DCCh Source:6090h
  //int ois_servo_gain_x_eeprom = 0x1DCC;
  //int ois_servo_gain_x_source = 0x6090;
  //regdata = 0x00;
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_servo_gain_x_eeprom, &regdata) > 0) {
  //  printf("read: [7. OIS Servo loop gain X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_servo_gain_x_eeprom, regdata);
  //  if (__CCIRegWriteBySlaveAddress(slave_address, ois_servo_gain_x_source, regdata) > 0) {
  //     printf("set: [7. OIS Servo loop gain X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_servo_gain_x_source, regdata);
  //  }
  //}
  // 8. OIS Servo loop gain Y DataWidthDetai:16bit EEPROM:1DCEh Source:6090h
  //int ois_servo_gain_y_eeprom = 0x1DCE;
  //int ois_servo_gain_y_source = 0x6090;
  //regdata = 0x00;
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_servo_gain_y_eeprom, &regdata) > 0) {
  //  printf("read: [8. OIS Servo loop gain Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_servo_gain_y_eeprom, regdata);
  //  if (__CCIRegWriteBySlaveAddress(slave_address, ois_servo_gain_y_source, regdata) > 0) {
  //     printf("set: [8. OIS Servo loop gain Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_servo_gain_y_source, regdata);
  //  }
  //}
  // 9. OIS Gyro gain X DataWidthDetai:16bit EEPROM:1DD6h Source:607Ah
  //int ois_gyro_gain_x_eeprom = 0x1DD6;
  int ois_gyro_gain_x_source = 0x607A;
  int ois_gyro_gain_x_source1 = 0x607B;
  //regdata = 0x14; //No.24
  //regdata1 = 0xB3; //No.24
  regdata = 0x17; //No.14
  regdata1 = 0x76; //No.14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_gyro_gain_x_eeprom, &regdata) > 0) {
  //  printf("read: [9. OIS Gyro gain X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_gain_x_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_gyro_gain_x_source, regdata) > 0) {
       printf("set: [9. OIS Gyro gain X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_gain_x_source, regdata);
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_gyro_gain_x_source1, regdata1) > 0) {
       printf("set: [9. OIS Gyro gain X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_gain_x_source1, regdata1);
    }
  //}
  // 10. OIS Gyro gain Y DataWidthDetai:16bit EEPROM:1DD8h Source:607Ch
  //int ois_gyro_gain_y_eeprom = 0x1DD8;
  int ois_gyro_gain_y_source = 0x607C;
  int ois_gyro_gain_y_source1 = 0x607D;
  //regdata = 0x38; //No.24
  //regdata1 = 0xD5; //No.24
  regdata = 0x18; //No.14
  regdata1 = 0x27; //No.14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_gyro_gain_y_eeprom, &regdata) > 0) {
  //  printf("read: [10. OIS Gyro gain Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_gain_y_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_gyro_gain_y_source, regdata) > 0) {
       printf("set: [10. OIS Gyro gain Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_gain_y_source, regdata);
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_gyro_gain_y_source1, regdata1) > 0) {
       printf("set: [10. OIS Gyro gain Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_gain_y_source1, regdata1);
    }
  //}
  // 11. OIS Gyro offset X DataWidthDetai:16bit EEPROM:1DDAh Source:6076h
  //int ois_gyro_off_x_eeprom = 0x1DDA;
  int ois_gyro_off_x_source = 0x6076;
  int ois_gyro_off_x_source1 = 0x6077;
  regdata = 0x29; //No.24,14
  regdata1 = 0x24; //No.24,14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_gyro_off_x_eeprom, &regdata) > 0) {
  //  printf("read: [11. OIS Gyro offset X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_off_x_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_gyro_off_x_source, regdata) > 0) {
       printf("set: [11. OIS Gyro offset X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_off_x_source, regdata);
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_gyro_off_x_source1, regdata1) > 0) {
       printf("set: [11. OIS Gyro offset X] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_off_x_source1, regdata1);
    }
  //}
  // 12. OIS Gyro offset Y DataWidthDetai:16bit EEPROM:1DDCh Source:6078h
  //int ois_gyro_off_y_eeprom = 0x1DDC;
  int ois_gyro_off_y_source = 0x6078;
  int ois_gyro_off_y_source1 = 0x6079;
  regdata = 0xD6; //No.24,14
  regdata1 = 0xDC; //No.24,14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_gyro_off_y_eeprom, &regdata) > 0) {
  //  printf("read: [12. OIS Gyro offset Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_off_y_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_gyro_off_y_source, regdata) > 0) {
       printf("set: [12. OIS Gyro offset Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_off_y_source, regdata);
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_gyro_off_y_source1, regdata1) > 0) {
       printf("set: [12. OIS Gyro offset Y] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_gyro_off_y_source1, regdata1);
    }
  //}
  // 13. Oscillator DataWidthDetai:8bit EEPROM:1DD5h Source:607Eh
  //int ois_oscillator_eeprom = 0x1DD5;
  //int ois_oscillator_source = 0x607E;
  //regdata = 0x78;
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_oscillator_eeprom, &regdata) > 0) {
  //  printf("read: [13. Oscillator] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_oscillator_eeprom, regdata);
  //  if (__CCIRegWriteBySlaveAddress(slave_address, ois_oscillator_source, regdata) > 0) {
  //     printf("set: [13. Oscillator] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_oscillator_source, regdata);
  //  }
  //}
  // 14. CLAF Hall current DataWidthDetai:8bit unsigned EEPROM:1DE7h Source:610Ah
  //int ois_claf_hall_cur_eeprom = 0x1DE7;
  int ois_claf_hall_cur_source = 0x610A;
  //regdata = 0x57; //No.24
  regdata = 0x4E; //No.14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_claf_hall_cur_eeprom, &regdata) > 0) {
  //  printf("read: [14. CLAF Hall current] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_claf_hall_cur_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_claf_hall_cur_source, regdata) > 0) {
       printf("set: [14. CLAF Hall current] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_claf_hall_cur_source, regdata);
    }
  //}
  // 15. CLAF Hall offset DataWidthDetai:10bit EEPROM:1DE8h Source:60E0h
  //int ois_claf_hall_off_eeprom = 0x1DE8;
  int ois_claf_hall_off_source = 0x60E0;
  int ois_claf_hall_off_source1 = 0x60E1;
  //regdata = 0x01; //No.24
  //regdata1 = 0x7A; //No.24
  regdata = 0x01; //No.14
  regdata1 = 0x8D; //No.14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_claf_hall_off_eeprom, &regdata) > 0) {
  //  printf("read: [15. CLAF Hall offset] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_claf_hall_off_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_claf_hall_off_source, regdata) > 0) {
       printf("set: [15. CLAF Hall offset] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_claf_hall_off_source, regdata);
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_claf_hall_off_source1, regdata1) > 0) {
       printf("set: [15. CLAF Hall offset] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_claf_hall_off_source1, regdata1);
    }
  //}
  // 16. CLAF Servo loop gain DataWidthDetai:16bit EEPROM:1DEAh Source:60E2h
  //int ois_claf_servo_gain_eeprom = 0x1DEA;
  int ois_claf_servo_gain_source = 0x60E2;
  int ois_claf_servo_gain_source1 = 0x60E3;
  regdata = 0x00; //No.24,14
  regdata1 = 0x00; //No.24,14
  //if (__CCIRegReadBySlaveAddress(slave_address, ois_claf_servo_gain_eeprom, &regdata) > 0) {
  //  printf("read: [16. CLAF Servo loop gain] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_claf_servo_gain_eeprom, regdata);
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_claf_servo_gain_source, regdata) > 0) {
       printf("set: [16. CLAF Servo loop gain] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_claf_servo_gain_source, regdata);
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, ois_claf_servo_gain_source1, regdata1) > 0) {
       printf("set: [16. CLAF Servo loop gain] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_claf_servo_gain_source1, regdata1);
    }
  //}

}
/**
 * @brief
 * Sensor Focus/OIS Demo
 */
void SensorFocusWnd::OnInitDemo(wxCommandEvent& event) {
  DEBUG_PRINT("SensorFocusWnd::OnInit\n");

  int slave_address = (0x7c) >> 1;

  //@Gyro Activation
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_gyro_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_gyro_[i].adr, i2cdat_gyro_[i].dat) > 0) {
      printf("set: [Demo:i2cdat_gyro] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_gyro_[i].adr, i2cdat_gyro_[i].dat);
    } else {
      printf("set: error(Demo:i2cdat_gyro)!\n");
    }
  }

  //@Start download
  int start_address = 0xF010;
  int start_regdata = 0x00;
  if (__CCIRegWriteBySlaveAddress(slave_address, start_address, start_regdata) > 0) {
     printf("set: [Demo:Start download] slave=%02x, address=%04x, RegData = %02x\n", slave_address, start_address, start_regdata);
  } else {
    printf("set: error(Demo:Start download)!\n");
  }

  //@DownloadProgram1(SineTest)
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_SineTest_dp1_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_SineTest_dp1_[i].adr, i2cdat_SineTest_dp1_[i].dat) > 0) {
      printf("set: [Demo:DownloadProgram1] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_SineTest_dp1_[i].adr, i2cdat_SineTest_dp1_[i].dat);
    } else {
      printf("set: error(Demo:DownloadProgram1)!\n");
    }
  }

  //@DownloadProgram2(SineTest)
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_SineTest_dp2_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_SineTest_dp2_[i].adr, i2cdat_SineTest_dp2_[i].dat) > 0) {
      printf("set: [Demo:DownloadProgram2] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_SineTest_dp2_[i].adr, i2cdat_SineTest_dp2_[i].dat);
    } else {
      printf("set: error(Demo:DownloadProgram2)!\n");
    }
  }

  //@Calibration
  for (int i = 0; i < 65535; i++) {
    if (i2cdat_calb_[i].adr == 0xffff) {
      break;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_calb_[i].adr, i2cdat_calb_[i].dat) > 0) {
      printf("set: [Demo:Calibration] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_calb_[i].adr, i2cdat_calb_[i].dat);
    } else {
      printf("set: error(Demo:Calibration)!\n");
    }
  }

  //@Set OIS complete DL
  int ois_comp_address = 0xF006;
  int ois_com_regdata = 0x00;
  if (__CCIRegWriteBySlaveAddress(slave_address, ois_comp_address, ois_com_regdata) > 0) {
     printf("set: [Demo:Set OIS complete DL] slave=%02x, address=%04x, RegData = %02x\n", slave_address, ois_comp_address, ois_com_regdata);
  } else {
    printf("set: error(Demo:Set OIS complete DL)!\n");
  }
}

/**
 * @brief
 * Sensor Focus Mode Apply
 */
void SensorFocusWnd::OnFocusModeApply(wxCommandEvent& event) {
  DEBUG_PRINT("SensorFocusWnd::OnFocusModeApply\n");
  int regdata = 0x0000;
  int serection_value;
  int slave_address = (0x7c) >> 1;

  serection_value = radio_box_focus_mode_->GetSelection();

  if (serection_value == kFocusMode1) {
    // Direct
	int set_address = 0x60D6;
	int set_regdata = 0x04;
	if (__CCIRegWriteBySlaveAddress(slave_address, set_address, set_regdata) > 0) {
	  printf("set: [Direct(0x60D6)] slave=%02x, address=%04x, RegData = %02x\n", slave_address, set_address, set_regdata);
	} else {
	  printf("set: error(Direct(0x60D6))!\n");
	}

	int direct_address0 = 0x60E4;
	int direct_address1 = 0x60E5;
	unsigned char direct_regdata[2];
	sscanf(text_ctrl_focus_code_->GetLineText(0).ToAscii(), "%x", &regdata);
	//printf("text_ctrl_focus_code_ %x\n",regdata);
	if (regdata > 0x03ff || regdata < 0) {
	  printf("write: invalid data\n");
	  return;
	}
    direct_regdata[0] = (unsigned char)((regdata >> 8) & 0xFF);
    direct_regdata[1] = (unsigned char)( regdata       & 0xFF);
    if (__CCIRegWriteBySlaveAddress(slave_address, direct_address0, direct_regdata[0]) > 0) {
      printf("set: [FocusMode Direct(0x60E4)] slave=%02x, address=%04x, RegData = %02x\n", slave_address, direct_address0, direct_regdata[0]);
    } else {
      printf("set: error(FocusMode Direct(0x60E4))!\n");
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, direct_address1, direct_regdata[1]) > 0) {
      printf("set: [FocusMode Direct(0x60E5)] slave=%02x, address=%04x, RegData = %02x\n", slave_address, direct_address1, direct_regdata[1]);
    } else {
      printf("set: error(FocusMode Direct(0x60E5))!\n");
    }

  } else if (serection_value == kFocusMode2) {
    // Inf
    bool retval = false;
    wxTextFile read_focusmode_file;
    wxString wxLineStr, wxTokenStr;
    wxStringTokenizer wxTokenizer;
    int get_line_count = 0;
    int token_count = 0;
    char value[128];
    int i2cdat_focusmodeinf_adr1 = 0x0000;
    int i2cdat_focusmodeinf_adr2 = 0x0000;
    int i2cdat_focusmodeinf_dat1 = 0x00;
    int i2cdat_focusmodeinf_dat2 = 0x00;

    retval = read_focusmode_file.Open(wxT(kImx378FocusModeInfFile));
    if (retval == false) {
      printf("Failed to read imx378 focus mode Inf file\n");
    } else {
      for (wxLineStr = read_focusmode_file.GetFirstLine(); !read_focusmode_file.Eof();
           wxLineStr = read_focusmode_file.GetNextLine()) {
        memset(value, 0, sizeof(value));
        wxTokenizer.SetString(wxLineStr, wxT(","), wxTOKEN_DEFAULT);
        token_count = wxTokenizer.CountTokens();
        if (get_line_count == 0){
          for (int i = 0; i < token_count; i++) {
            wxTokenStr = wxTokenizer.GetNextToken();
            if (i == 0) {
              memset(value, 0, sizeof(value));
              strncpy(value, (const char*)wxTokenStr.mb_str(), static_cast<int>(wxTokenStr.length()));
              i2cdat_focusmodeinf_adr1 = strtol(value, NULL, 16);
            } else if (i == 1) {
              memset(value, 0, sizeof(value));
              strncpy(value, (const char*)wxTokenStr.mb_str(), static_cast<int>(wxTokenStr.length()));
              i2cdat_focusmodeinf_dat1 = strtol(value, NULL, 16);
            }
          }
        } else {
          for (int i = 0; i < token_count; i++) {
            wxTokenStr = wxTokenizer.GetNextToken();
            if (i == 0) {
              memset(value, 0, sizeof(value));
              strncpy(value, (const char*)wxTokenStr.mb_str(), static_cast<int>(wxTokenStr.length()));
              i2cdat_focusmodeinf_adr2 = strtol(value, NULL, 16);
            } else if (i == 1) {
              memset(value, 0, sizeof(value));
              strncpy(value, (const char*)wxTokenStr.mb_str(), static_cast<int>(wxTokenStr.length()));
              i2cdat_focusmodeinf_dat2 = strtol(value, NULL, 16);
            }
          }
        }
        get_line_count++;
      }
  	  int set_address = 0x60D6;
  	  int set_regdata = 0x04;
  	  int inf_regdata = 0x0000;
  	  unsigned char infregdata[2];
  	  if (__CCIRegWriteBySlaveAddress(slave_address, set_address, set_regdata) > 0) {
  	    printf("set: [FocusMode Inf(0x60D6)] slave=%02x, address=%04x, RegData = %02x\n", slave_address, set_address, set_regdata);
  	  } else {
  	    printf("set: error(FocusMode Inf(0x60D6))!\n");
  	  }
      if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_focusmodeinf_adr1, i2cdat_focusmodeinf_dat1) > 0) {
        printf("set: [FocusMode Inf] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_focusmodeinf_adr1, i2cdat_focusmodeinf_dat1);
        infregdata[0] = (unsigned char)( i2cdat_focusmodeinf_dat1 & 0xFF);
      } else {
        printf("set: error(FocusMode Inf)!\n");
      }
      if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_focusmodeinf_adr2, i2cdat_focusmodeinf_dat2) > 0) {
        printf("set: [FocusMode Inf] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_focusmodeinf_adr2, i2cdat_focusmodeinf_dat2);
        infregdata[1] = (unsigned char)( i2cdat_focusmodeinf_dat2 & 0xFF);
      } else {
        printf("set: error(FocusMode Inf)!\n");
      }
      inf_regdata = ((infregdata[0] & 0xFF) << 8) + (infregdata[1] & 0xFF);
      text_ctrl_focus_code_->SetValue(wxString::Format(wxT("0x%04x"), inf_regdata));
    }
    read_focusmode_file.Close();
  } else if (serection_value == kFocusMode3) {
    // Macro
    bool retval = false;
    wxTextFile read_focusmode_file;
    wxString wxLineStr, wxTokenStr;
    wxStringTokenizer wxTokenizer;
    int get_line_count = 0;
    int token_count = 0;
    char value[128];
    int i2cdat_focusmodemacro_adr1 = 0x0000;
    int i2cdat_focusmodemacro_adr2 = 0x0000;
    int i2cdat_focusmodemacro_dat1 = 0x00;
    int i2cdat_focusmodemacro_dat2 = 0x00;

    retval = read_focusmode_file.Open(wxT(kImx378FocusModeMacroFile));
    if (retval == false) {
      printf("Failed to read imx378 focus mode Macro file\n");
    } else {
      for (wxLineStr = read_focusmode_file.GetFirstLine(); !read_focusmode_file.Eof();
           wxLineStr = read_focusmode_file.GetNextLine()) {
        memset(value, 0, sizeof(value));
        wxTokenizer.SetString(wxLineStr, wxT(","), wxTOKEN_DEFAULT);
        token_count = wxTokenizer.CountTokens();
        if (get_line_count == 0){
          for (int i = 0; i < token_count; i++) {
            wxTokenStr = wxTokenizer.GetNextToken();
            if (i == 0) {
              memset(value, 0, sizeof(value));
              strncpy(value, (const char*)wxTokenStr.mb_str(), static_cast<int>(wxTokenStr.length()));
              i2cdat_focusmodemacro_adr1 = strtol(value, NULL, 16);
            } else if (i == 1) {
               memset(value, 0, sizeof(value));
              strncpy(value, (const char*)wxTokenStr.mb_str(), static_cast<int>(wxTokenStr.length()));
              i2cdat_focusmodemacro_dat1 = strtol(value, NULL, 16);
            }
          }
        } else {
          for (int i = 0; i < token_count; i++) {
            wxTokenStr = wxTokenizer.GetNextToken();
            if (i == 0) {
               memset(value, 0, sizeof(value));
              strncpy(value, (const char*)wxTokenStr.mb_str(), static_cast<int>(wxTokenStr.length()));
              i2cdat_focusmodemacro_adr2 = strtol(value, NULL, 16);
            } else if (i == 1) {
               memset(value, 0, sizeof(value));
              strncpy(value, (const char*)wxTokenStr.mb_str(), static_cast<int>(wxTokenStr.length()));
              i2cdat_focusmodemacro_dat2 = strtol(value, NULL, 16);
            }
          }
        }
        get_line_count++;
      }

  	  int set_address = 0x60D6;
  	  int set_regdata = 0x04;
  	  int macro_regdata = 0x0000;
  	  unsigned char macroregdata[2];
  	  if (__CCIRegWriteBySlaveAddress(slave_address, set_address, set_regdata) > 0) {
  	    printf("set: [FocusMode Macro(0x60D6)] slave=%02x, address=%04x, RegData = %02x\n", slave_address, set_address, set_regdata);
  	  } else {
  	      printf("set: error(FocusMode Macro(0x60D6))!\n");
  	  }
      if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_focusmodemacro_adr1, i2cdat_focusmodemacro_dat1) > 0) {
        printf("set: [FocusMode Macro] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_focusmodemacro_adr1, i2cdat_focusmodemacro_dat1);
        macroregdata[0] = (unsigned char)( i2cdat_focusmodemacro_dat1 & 0xFF);
      } else {
        printf("set: error(FocusMode Macro)!\n");
      }
      if (__CCIRegWriteBySlaveAddress(slave_address, i2cdat_focusmodemacro_adr2, i2cdat_focusmodemacro_dat2) > 0) {
        printf("set: [FocusMode Macro] slave=%02x, address=%04x, RegData = %02x\n", slave_address, i2cdat_focusmodemacro_adr2, i2cdat_focusmodemacro_dat2);
        macroregdata[1] = (unsigned char)( i2cdat_focusmodemacro_dat2 & 0xFF);
      } else {
        printf("set: error(FocusMode Macro)!\n");
      }
      macro_regdata = ((macroregdata[0] & 0xFF) << 8) + (macroregdata[1] & 0xFF);
      text_ctrl_focus_code_->SetValue(wxString::Format(wxT("0x%04x"), macro_regdata));
    }
    read_focusmode_file.Close();
  }
}

/**
 * @brief
 * Sensor OIS Mode Apply
 */
void SensorFocusWnd::OnOISModeApply(wxCommandEvent& event) {
  DEBUG_PRINT("SensorFocusWnd::OnOISModeApply\n");

  int serection_value;
  int slave_address = (0x7c) >> 1;

  int off_address = 0x6020;
  int off_regdata = 0x01;

  serection_value = radio_box_ois_mode_->GetSelection();
  if (serection_value == kOisModeMode1) {
    // OFF
    if (__CCIRegWriteBySlaveAddress(slave_address, off_address, off_regdata) > 0) {
      printf("set: [OIS OFF] slave=%02x, address=%04x, RegData = %02x\n", slave_address, off_address, off_regdata);
    } else {
      printf("set: error(OIS OFF)!\n");
    }
  } else {
    // OFF
    if (__CCIRegWriteBySlaveAddress(slave_address, off_address, off_regdata) > 0) {
      printf("set: [1.OIS OFF] slave=%02x, address=%04x, RegData = %02x\n", slave_address, off_address, off_regdata);
    } else {
      printf("set: error(1.OIS OFF)!\n");
    }

    int mode_address = 0x6021;
    int mode_regdata;
    if (serection_value == kOisModeMode2) {
      // Zero shutter
      mode_regdata = 0x6B;
    } else if (serection_value == kOisModeMode3) {
      // Exposure/Shake eval.
      mode_regdata = 0x03;
    } else if (serection_value == kOisModeMode4) {
      // Moviemovie_
      mode_regdata = 0x61;
    } else if (serection_value == kOisModeMode5) {
      // High SR Movie
      mode_regdata = 0x63;
    } else if (serection_value == kOisModeMode6) {
      // View Finder
      mode_regdata = 0x79;
    }
    if (__CCIRegWriteBySlaveAddress(slave_address, mode_address, mode_regdata) > 0) {
      printf("set: [2.Mode setting] slave=%02x, address=%04x, RegData = %02x\n", slave_address, mode_address, mode_regdata);
      if (__CCIRegReadBySlaveAddress(slave_address, mode_address, &mode_regdata) > 0) {
        printf("read: [2.Mode setting] slave=%02x, address=%04x, RegData = %02x\n", slave_address, mode_address, mode_regdata);
      }
    } else {
      printf("set: error(2.Mode setting)!\n");
    }

    // ON
    int on_address = 0x6020;
    int on_regdata = 0x02;
    if (__CCIRegWriteBySlaveAddress(slave_address, on_address, on_regdata) > 0) {
      printf("set: [3.OIS ON] slave=%02x, address=%04x, RegData = %02x\n", slave_address, on_address, on_regdata);
    } else {
      printf("set: error(3.OIS ON)!\n");
    }
  }
}

/**
 * @brief
 * Sensor OIS Mode Demo Start
 */
void SensorFocusWnd::OnOISModeDemoStart(wxCommandEvent& event) {
  DEBUG_PRINT("SensorFocusWnd::OnOISModeDemoStart\n");

  int serection_value;
  int slave_address = (0x7c) >> 1;

  int off_address = 0x6020;
  int off_regdata = 0x01;

  serection_value = radio_box_ois_mode_->GetSelection();
  if (serection_value == kOisModeMode7) {
    // Demo
    // OFF
    if (__CCIRegWriteBySlaveAddress(slave_address, off_address, off_regdata) > 0) {
      printf("set: [Demo:1.OIS OFF] slave=%02x, address=%04x, RegData = %02x\n", slave_address, off_address, off_regdata);
    } else {
       printf("set: error(Demo:1.OIS OFF)!\n");
    }
    // Sine frequency
    int sine_f_address = 0x6131;
    int sine_f_regdata = 0x02;
    if (__CCIRegWriteBySlaveAddress(slave_address, sine_f_address, sine_f_regdata) > 0) {
      printf("set: [Demo:2.Sine frequency] slave=%02x, address=%04x, RegData = %02x\n", slave_address, sine_f_address, sine_f_regdata);
    } else {
       printf("set: error(Demo:2.Sine frequency)!\n");
    }
    // Sine vibration
    int sine_v_address = 0x6132;
    int sine_v_regdata = 0x80;
    if (__CCIRegWriteBySlaveAddress(slave_address, sine_v_address, sine_v_regdata) > 0) {
      printf("set: [Demo:3.Sine vibration] slave=%02x, address=%04x, RegData = %02x\n", slave_address, sine_v_address, sine_v_regdata);
    } else {
       printf("set: error(Demo:3.Sine vibration)!\n");
    }
    // Manual mode
    int manual_address = 0x6020;
    int Manual_regdata = 0x03;
    if (__CCIRegWriteBySlaveAddress(slave_address, manual_address, Manual_regdata) > 0) {
      printf("set: [Demo:4.Manual mode] slave=%02x, address=%04x, RegData = %02x\n", slave_address, manual_address, Manual_regdata);
    } else {
       printf("set: error(Demo:4.Manual mode)!\n");
    }
    // Sine Wave Drive Start
    int sine_wd_start_address = 0x6130;
    int sine_wd_start_regdata = 0x03;
    if (__CCIRegWriteBySlaveAddress(slave_address, sine_wd_start_address, sine_wd_start_regdata) > 0) {
      printf("set: [Demo:5.Sine Wave Drive Start] slave=%02x, address=%04x, RegData = %02x\n", slave_address, sine_wd_start_address, sine_wd_start_regdata);
    } else {
       printf("set: error(Demo:5.Sine Wave Drive Start)!\n");
    }
  }
}

/**
 * @brief
 * Sensor OIS Mode Demo Stop
 */
void SensorFocusWnd::OnOISModeDemoStop(wxCommandEvent& event) {
  DEBUG_PRINT("SensorFocusWnd::OnOISModeDemoStop\n");

  int serection_value;
  int slave_address = (0x7c) >> 1;

  int off_address = 0x6020;
  int off_regdata = 0x01;

  serection_value = radio_box_ois_mode_->GetSelection();
  if (serection_value == kOisModeMode7) {
    // Demo
    // Sine Wave Drive Stop
    int sine_wd_stop_address = 0x6130;
    int sine_wd_stop_regdata = 0x00;
    if (__CCIRegWriteBySlaveAddress(slave_address, sine_wd_stop_address, sine_wd_stop_regdata) > 0) {
      printf("set: [Demo:6.Sine Wave Drive Stop] slave=%02x, address=%04x, RegData = %02x\n", slave_address, sine_wd_stop_address, sine_wd_stop_regdata);
    } else {
    	printf("set: error(Demo:6.Sine Wave Drive Stop)!\n");
    }

    // OFF
    if (__CCIRegWriteBySlaveAddress(slave_address, off_address, off_regdata) > 0) {
      printf("set: [Demo:7.OIS OFF] slave=%02x, address=%04x, RegData = %02x\n", slave_address, off_address, off_regdata);
    } else {
       printf("set: error(Demo:7.OIS OFF)!\n");
    }
  }
}

const T_I2CRegWriteData SensorFocusWnd::i2cdat_gyro_[] = {
  0xF020, 0x01,
  0xF021, 0x01,
  0xF022, 0x18,
  0xF023, 0x10,
  0xF024, 0x00,
  0xF025, 0x01,
  0xF026, 0x00,
  0xF027, 0x02,
  0xF028, 0x00,
  0xF029, 0x00,
  0xF02A, 0x00,
  0xF02B, 0xE1,
  0xF02C, 0x0B,
  0xF02D, 0x13,
  0xF02E, 0x00,
  0xF02F, 0x00,
  0xF02C, 0x0C,
  0xF02D, 0x80,
  0xF02E, 0x00,
  0xF02F, 0x00,
  0xffff, 0xff,
};

const T_I2CRegWriteData SensorFocusWnd::i2cdat_dp1_[] = {
  0x0000, 0x24,
  0x0001, 0x00,
  0x0002, 0xB5,
  0x0003, 0x70,
  0x0004, 0x28,
  0x0005, 0x00,
  0x0006, 0x46,
  0x0007, 0x22,
  0x0008, 0x4A,
  0x0009, 0x35,
  0x000A, 0xDA,
  0x000B, 0x10,
  0x000C, 0xD0,
  0x000D, 0x04,
  0x000E, 0x18,
  0x000F, 0x80,
  0x0010, 0xD0,
  0x0011, 0x04,
  0x0012, 0x28,
  0x0013, 0x01,
  0x0014, 0xD1,
  0x0015, 0x60,
  0x0016, 0x28,
  0x0017, 0x04,
  0x0018, 0x4A,
  0x0019, 0x32,
  0x001A, 0xE0,
  0x001B, 0x03,
  0x001C, 0x4A,
  0x001D, 0x32,
  0x001E, 0xE0,
  0x001F, 0x02,
  0x0020, 0x4A,
  0x0021, 0x32,
  0x0022, 0xE0,
  0x0023, 0x00,
  0x0024, 0x47,
  0x0025, 0x90,
  0x0026, 0x46,
  0x0027, 0x08,
  0x0028, 0xE0,
  0x0029, 0x56,
  0x002A, 0x46,
  0x002B, 0x04,
  0x002C, 0x33,
  0x002D, 0x40,
  0x002E, 0x23,
  0x002F, 0xFF,
  0x0030, 0x42,
  0x0031, 0x98,
  0x0032, 0x1A,
  0x0033, 0xC5,
  0x0034, 0xDC,
  0x0035, 0x1F,
  0x0036, 0xD0,
  0x0037, 0x4B,
  0x0038, 0xD0,
  0x0039, 0x3D,
  0x003A, 0x28,
  0x003B, 0xCC,
  0x003C, 0x28,
  0x003D, 0x94,
  0x003E, 0xDC,
  0x003F, 0x0C,
  0x0040, 0xDC,
  0x0041, 0x04,
  0x0042, 0xD0,
  0x0043, 0x45,
  0x0044, 0xD0,
  0x0045, 0x42,
  0x0046, 0x28,
  0x0047, 0x92,
  0x0048, 0xD1,
  0x0049, 0x46,
  0x004A, 0x28,
  0x004B, 0x93,
  0x004C, 0x28,
  0x004D, 0xA0,
  0x004E, 0xE0,
  0x004F, 0x28,
  0x0050, 0x28,
  0x0051, 0xA1,
  0x0052, 0xD0,
  0x0053, 0x3D,
  0x0054, 0xE0,
  0x0055, 0x25,
  0x0056, 0xD1,
  0x0057, 0x41,
  0x0058, 0x35,
  0x0059, 0x0D,
  0x005A, 0x25,
  0x005B, 0xFF,
  0x005C, 0x42,
  0x005D, 0xA8,
  0x005E, 0x1B,
  0x005F, 0x43,
  0x0060, 0xDC,
  0x0061, 0x04,
  0x0062, 0xD0,
  0x0063, 0x35,
  0x0064, 0xD0,
  0x0065, 0x32,
  0x0066, 0x33,
  0x0067, 0x0C,
  0x0068, 0xD1,
  0x0069, 0x36,
  0x006A, 0x2B,
  0x006B, 0x01,
  0x006C, 0x2B,
  0x006D, 0x01,
  0x006E, 0xE0,
  0x006F, 0x20,
  0x0070, 0x2B,
  0x0071, 0x02,
  0x0072, 0xD0,
  0x0073, 0x20,
  0x0074, 0xE0,
  0x0075, 0x2A,
  0x0076, 0xD1,
  0x0077, 0x31,
  0x0078, 0xD0,
  0x0079, 0x25,
  0x007A, 0x2D,
  0x007B, 0x0E,
  0x007C, 0x3D,
  0x007D, 0x09,
  0x007E, 0xDC,
  0x007F, 0x07,
  0x0080, 0xF0,
  0x0081, 0x00,
  0x0082, 0x00,
  0x0083, 0x2B,
  0x0084, 0x24,
  0x0085, 0x05,
  0x0086, 0xFB,
  0x0087, 0x29,
  0x0088, 0x1F,
  0x0089, 0x1D,
  0x008A, 0x24,
  0x008B, 0x1B,
  0x008C, 0x3D,
  0x008D, 0x19,
  0x008E, 0x00,
  0x008F, 0x2A,
  0x0090, 0xF0,
  0x0091, 0x00,
  0x0092, 0x00,
  0x0093, 0x2B,
  0x0094, 0x1C,
  0x0095, 0x07,
  0x0096, 0xFB,
  0x0097, 0x21,
  0x0098, 0x22,
  0x0099, 0x0B,
  0x009A, 0x1C,
  0x009B, 0x09,
  0x009C, 0x00,
  0x009D, 0x22,
  0x009E, 0x1B,
  0x009F, 0x22,
  0x00A0, 0xE0,
  0x00A1, 0x14,
  0x00A2, 0x4A,
  0x00A3, 0x13,
  0x00A4, 0xE0,
  0x00A5, 0x12,
  0x00A6, 0x4A,
  0x00A7, 0x13,
  0x00A8, 0xE0,
  0x00A9, 0x10,
  0x00AA, 0x4A,
  0x00AB, 0x13,
  0x00AC, 0xE0,
  0x00AD, 0x0E,
  0x00AE, 0x4A,
  0x00AF, 0x13,
  0x00B0, 0xE0,
  0x00B1, 0x0C,
  0x00B2, 0x4A,
  0x00B3, 0x13,
  0x00B4, 0xE0,
  0x00B5, 0x0A,
  0x00B6, 0x4A,
  0x00B7, 0x13,
  0x00B8, 0xE0,
  0x00B9, 0x08,
  0x00BA, 0x4A,
  0x00BB, 0x13,
  0x00BC, 0xE0,
  0x00BD, 0x06,
  0x00BE, 0x4A,
  0x00BF, 0x13,
  0x00C0, 0xE0,
  0x00C1, 0x04,
  0x00C2, 0x4A,
  0x00C3, 0x13,
  0x00C4, 0xE0,
  0x00C5, 0x02,
  0x00C6, 0x4A,
  0x00C7, 0x13,
  0x00C8, 0xE0,
  0x00C9, 0x00,
  0x00CA, 0x4A,
  0x00CB, 0x13,
  0x00CC, 0x4B,
  0x00CD, 0x14,
  0x00CE, 0x4A,
  0x00CF, 0x13,
  0x00D0, 0x2A,
  0x00D1, 0x00,
  0x00D2, 0x54,
  0x00D3, 0x19,
  0x00D4, 0xB2,
  0x00D5, 0xC8,
  0x00D6, 0xD0,
  0x00D7, 0x01,
  0x00D8, 0x46,
  0x00D9, 0x20,
  0x00DA, 0x47,
  0x00DB, 0x90,
  0x00DC, 0x00,
  0x00DD, 0x00,
  0x00DE, 0xBD,
  0x00DF, 0x70,
  0x00E0, 0x7F,
  0x00E1, 0xFF,
  0x00E2, 0xFF,
  0x00E3, 0xD6,
  0x00E4, 0x10,
  0x00E5, 0x00,
  0x00E6, 0x06,
  0x00E7, 0x1B,
  0x00E8, 0x10,
  0x00E9, 0x00,
  0x00EA, 0x05,
  0x00EB, 0xCD,
  0x00EC, 0x10,
  0x00ED, 0x00,
  0x00EE, 0x04,
  0x00EF, 0xD5,
  0x00F0, 0x10,
  0x00F1, 0x00,
  0x00F2, 0x0A,
  0x00F3, 0xC5,
  0x00F4, 0x10,
  0x00F5, 0x00,
  0x00F6, 0x0A,
  0x00F7, 0xD7,
  0x00F8, 0x10,
  0x00F9, 0x00,
  0x00FA, 0x0A,
  0x00FB, 0x49,
  0x00FC, 0x10,
  0x00FD, 0x00,
  0x00FE, 0x0A,
  0x00FF, 0x69,
  0x0100, 0x10,
  0x0101, 0x00,
  0x0102, 0x08,
  0x0103, 0xD1,
  0x0104, 0x10,
  0x0105, 0x00,
  0x0106, 0x08,
  0x0107, 0xDD,
  0x0108, 0x10,
  0x0109, 0x00,
  0x010A, 0x0A,
  0x010B, 0xE5,
  0x010C, 0x10,
  0x010D, 0x00,
  0x010E, 0x08,
  0x010F, 0xED,
  0x0110, 0x10,
  0x0111, 0x00,
  0x0112, 0x09,
  0x0113, 0x59,
  0x0114, 0x10,
  0x0115, 0x00,
  0x0116, 0x09,
  0x0117, 0xA5,
  0x0118, 0x10,
  0x0119, 0x00,
  0x011A, 0x09,
  0x011B, 0xFD,
  0x011C, 0x10,
  0x011D, 0x00,
  0x011E, 0x0A,
  0x011F, 0x89,
  0x0120, 0x10,
  0x0121, 0x00,
  0x0122, 0x19,
  0x0123, 0x40,
  0x0124, 0x46,
  0x0125, 0x0C,
  0x0126, 0xB5,
  0x0127, 0x70,
  0x0128, 0xD0,
  0x0129, 0x15,
  0x012A, 0x00,
  0x012B, 0x05,
  0x012C, 0xD0,
  0x012D, 0x15,
  0x012E, 0x2D,
  0x012F, 0x01,
  0x0130, 0x3C,
  0x0131, 0xFF,
  0x0132, 0x21,
  0x0133, 0xFF,
  0x0134, 0x3C,
  0x0135, 0xFF,
  0x0136, 0x31,
  0x0137, 0x05,
  0x0138, 0x20,
  0x0139, 0x19,
  0x013A, 0x3C,
  0x013B, 0x02,
  0x013C, 0x18,
  0x013D, 0x08,
  0x013E, 0x06,
  0x013F, 0xC0,
  0x0140, 0x49,
  0x0141, 0xBE,
  0x0142, 0x60,
  0x0143, 0x04,
  0x0144, 0x68,
  0x0145, 0x09,
  0x0146, 0x48,
  0x0147, 0xBC,
  0x0148, 0x47,
  0x0149, 0x88,
  0x014A, 0x68,
  0x014B, 0x40,
  0x014C, 0xD0,
  0x014D, 0x07,
  0x014E, 0x2C,
  0x014F, 0x00,
  0x0150, 0xF0,
  0x0151, 0x00,
  0x0152, 0x46,
  0x0153, 0x28,
  0x0154, 0xBD,
  0x0155, 0x70,
  0x0156, 0xF9,
  0x0157, 0xA1,
  0x0158, 0xE7,
  0x0159, 0xEE,
  0x015A, 0x21,
  0x015B, 0x58,
  0x015C, 0xE7,
  0x015D, 0xEC,
  0x015E, 0x21,
  0x015F, 0x5C,
  0x0160, 0xBD,
  0x0161, 0x70,
  0x0162, 0x20,
  0x0163, 0x00,
  0x0164, 0x46,
  0x0165, 0x06,
  0x0166, 0xB5,
  0x0167, 0xF3,
  0x0168, 0xB0,
  0x0169, 0x89,
  0x016A, 0x48,
  0x016B, 0xB5,
  0x016C, 0x32,
  0x016D, 0xF9,
  0x016E, 0x1D,
  0x016F, 0xC2,
  0x0170, 0x21,
  0x0171, 0x00,
  0x0172, 0x30,
  0x0173, 0xA0,
  0x0174, 0x90,
  0x0175, 0x03,
  0x0176, 0x92,
  0x0177, 0x04,
  0x0178, 0xD0,
  0x0179, 0x00,
  0x017A, 0x2E,
  0x017B, 0x00,
  0x017C, 0x70,
  0x017D, 0x41,
  0x017E, 0x46,
  0x017F, 0x10,
  0x0180, 0x68,
  0x0181, 0x28,
  0x0182, 0x4D,
  0x0183, 0xAD,
  0x0184, 0x28,
  0x0185, 0x00,
  0x0186, 0x4F,
  0x0187, 0xAF,
  0x0188, 0xF0,
  0x0189, 0x00,
  0x018A, 0xD1,
  0x018B, 0x19,
  0x018C, 0x46,
  0x018D, 0x3C,
  0x018E, 0xF9,
  0x018F, 0x93,
  0x0190, 0x60,
  0x0191, 0x20,
  0x0192, 0x3C,
  0x0193, 0x0C,
  0x0194, 0xF0,
  0x0195, 0x00,
  0x0196, 0x20,
  0x0197, 0x01,
  0x0198, 0x60,
  0x0199, 0x60,
  0x019A, 0xF9,
  0x019B, 0x8D,
  0x019C, 0xF0,
  0x019D, 0x00,
  0x019E, 0x20,
  0x019F, 0x02,
  0x01A0, 0x60,
  0x01A1, 0xA0,
  0x01A2, 0xF9,
  0x01A3, 0x89,
  0x01A4, 0x20,
  0x01A5, 0x06,
  0x01A6, 0x4C,
  0x01A7, 0xA8,
  0x01A8, 0x60,
  0x01A9, 0x38,
  0x01AA, 0x47,
  0x01AB, 0xA0,
  0x01AC, 0x47,
  0x01AD, 0xA0,
  0x01AE, 0x20,
  0x01AF, 0x04,
  0x01B0, 0x20,
  0x01B1, 0x2A,
  0x01B2, 0x60,
  0x01B3, 0x78,
  0x01B4, 0x60,
  0x01B5, 0xB8,
  0x01B6, 0x47,
  0x01B7, 0xA0,
  0x01B8, 0x1C,
  0x01B9, 0x40,
  0x01BA, 0x68,
  0x01BB, 0x28,
  0x01BC, 0x21,
  0x01BD, 0x0F,
  0x01BE, 0x60,
  0x01BF, 0x28,
  0x01C0, 0xD0,
  0x01C1, 0x01,
  0x01C2, 0x2E,
  0x01C3, 0x00,
  0x01C4, 0xE0,
  0x01C5, 0x00,
  0x01C6, 0x98,
  0x01C7, 0x04,
  0x01C8, 0x78,
  0x01C9, 0x02,
  0x01CA, 0x98,
  0x01CB, 0x03,
  0x01CC, 0x28,
  0x01CD, 0x01,
  0x01CE, 0x98,
  0x01CF, 0x0A,
  0x01D0, 0x06,
  0x01D1, 0x10,
  0x01D2, 0xD1,
  0x01D3, 0x7E,
  0x01D4, 0x2E,
  0x01D5, 0x00,
  0x01D6, 0x0F,
  0x01D7, 0x00,
  0x01D8, 0x28,
  0x01D9, 0x03,
  0x01DA, 0xD0,
  0x01DB, 0x0C,
  0x01DC, 0x28,
  0x01DD, 0x00,
  0x01DE, 0xD8,
  0x01DF, 0x79,
  0x01E0, 0x24,
  0x01E1, 0x01,
  0x01E2, 0xD0,
  0x01E3, 0x06,
  0x01E4, 0x02,
  0x01E5, 0xA4,
  0x01E6, 0x01,
  0x01E7, 0xC3,
  0x01E8, 0x01,
  0x01E9, 0xC0,
  0x01EA, 0x1A,
  0x01EB, 0xE3,
  0x01EC, 0xE0,
  0x01ED, 0x0B,
  0x01EE, 0x93,
  0x01EF, 0x00,
  0x01F0, 0xE7,
  0x01F1, 0xFA,
  0x01F2, 0x4B,
  0x01F3, 0x96,
  0x01F4, 0xD8,
  0x01F5, 0x6D,
  0x01F6, 0x28,
  0x01F7, 0x07,
  0x01F8, 0xD0,
  0x01F9, 0x19,
  0x01FA, 0x28,
  0x01FB, 0x00,
  0x01FC, 0x90,
  0x01FD, 0x00,
  0x01FE, 0x02,
  0x01FF, 0x00,
  0x0200, 0x1C,
  0x0201, 0x40,
  0x0202, 0x43,
  0x0203, 0xC0,
  0x0204, 0x0D,
  0x0205, 0x00,
  0x0206, 0x05,
  0x0207, 0x00,
  0x0208, 0x07,
  0x0209, 0x10,
  0x020A, 0x90,
  0x020B, 0x05,
  0x020C, 0x28,
  0x020D, 0x07,
  0x020E, 0x0F,
  0x020F, 0x00,
  0x0210, 0x02,
  0x0211, 0x40,
  0x0212, 0xD8,
  0x0213, 0x5E,
  0x0214, 0x90,
  0x0215, 0x07,
  0x0216, 0x4A,
  0x0217, 0x8E,
  0x0218, 0x68,
  0x0219, 0x12,
  0x021A, 0x21,
  0x021B, 0x00,
  0x021C, 0x47,
  0x021D, 0x90,
  0x021E, 0x20,
  0x021F, 0x02,
  0x0220, 0x43,
  0x0221, 0x48,
  0x0222, 0x21,
  0x0223, 0x64,
  0x0224, 0x60,
  0x0225, 0x68,
  0x0226, 0x24,
  0x0227, 0x00,
  0x0228, 0xD0,
  0x0229, 0x03,
  0x022A, 0x2E,
  0x022B, 0x00,
  0x022C, 0xE0,
  0x022D, 0x05,
  0x022E, 0x24,
  0x022F, 0x02,
  0x0230, 0xE7,
  0x0231, 0xE4,
  0x0232, 0x20,
  0x0233, 0x80,
  0x0234, 0xD0,
  0x0235, 0x12,
  0x0236, 0x2C,
  0x0237, 0x00,
  0x0238, 0xD0,
  0x0239, 0x12,
  0x023A, 0x2C,
  0x023B, 0x01,
  0x023C, 0x4A,
  0x023D, 0x85,
  0x023E, 0x21,
  0x023F, 0x90,
  0x0240, 0x68,
  0x0241, 0x12,
  0x0242, 0x20,
  0x0243, 0x04,
  0x0244, 0x21,
  0x0245, 0x01,
  0x0246, 0x47,
  0x0247, 0x90,
  0x0248, 0x1A,
  0x0249, 0x08,
  0x024A, 0x02,
  0x024B, 0xC9,
  0x024C, 0x0D,
  0x024D, 0x00,
  0x024E, 0x05,
  0x024F, 0x00,
  0x0250, 0x2E,
  0x0251, 0x00,
  0x0252, 0x90,
  0x0253, 0x02,
  0x0254, 0x48,
  0x0255, 0x7B,
  0x0256, 0xD0,
  0x0257, 0x07,
  0x0258, 0x58,
  0x0259, 0x45,
  0x025A, 0x00,
  0x025B, 0xA1,
  0x025C, 0x21,
  0x025D, 0x8C,
  0x025E, 0xE0,
  0x025F, 0x17,
  0x0260, 0x21,
  0x0261, 0x92,
  0x0262, 0xE7,
  0x0263, 0xED,
  0x0264, 0x46,
  0x0265, 0x20,
  0x0266, 0xE7,
  0x0267, 0xEB,
  0x0268, 0xF9,
  0x0269, 0x16,
  0x026A, 0xF0,
  0x026B, 0x00,
  0x026C, 0x1A,
  0x026D, 0x40,
  0x026E, 0x99,
  0x026F, 0x02,
  0x0270, 0xF0,
  0x0271, 0x00,
  0x0272, 0x21,
  0x0273, 0x0A,
  0x0274, 0x49,
  0x0275, 0x73,
  0x0276, 0xFA,
  0x0277, 0x25,
  0x0278, 0x58,
  0x0279, 0x89,
  0x027A, 0x00,
  0x027B, 0xA2,
  0x027C, 0x46,
  0x027D, 0x29,
  0x027E, 0x1A,
  0x027F, 0x0D,
  0x0280, 0xF0,
  0x0281, 0x00,
  0x0282, 0x46,
  0x0283, 0x20,
  0x0284, 0x49,
  0x0285, 0x6D,
  0x0286, 0xF8,
  0x0287, 0xF9,
  0x0288, 0x68,
  0x0289, 0x09,
  0x028A, 0x20,
  0x028B, 0x01,
  0x028C, 0x46,
  0x028D, 0x20,
  0x028E, 0x47,
  0x028F, 0x88,
  0x0290, 0xF7,
  0x0291, 0xFF,
  0x0292, 0x99,
  0x0293, 0x00,
  0x0294, 0x90,
  0x0295, 0x01,
  0x0296, 0xFF,
  0x0297, 0x47,
  0x0298, 0x99,
  0x0299, 0x05,
  0x029A, 0x46,
  0x029B, 0x20,
  0x029C, 0xFF,
  0x029D, 0x42,
  0x029E, 0xF7,
  0x029F, 0xFF,
  0x02A0, 0x98,
  0x02A1, 0x01,
  0x02A2, 0x46,
  0x02A3, 0x07,
  0x02A4, 0xD0,
  0x02A5, 0x26,
  0x02A6, 0x2E,
  0x02A7, 0x00,
  0x02A8, 0xD5,
  0x02A9, 0x00,
  0x02AA, 0x1B,
  0x02AB, 0xC0,
  0x02AC, 0x10,
  0x02AD, 0x41,
  0x02AE, 0x42,
  0x02AF, 0x40,
  0x02B0, 0x42,
  0x02B1, 0xB8,
  0x02B2, 0x98,
  0x02B3, 0x01,
  0x02B4, 0x46,
  0x02B5, 0x07,
  0x02B6, 0xD2,
  0x02B7, 0x00,
  0x02B8, 0x99,
  0x02B9, 0x02,
  0x02BA, 0x19,
  0x02BB, 0xC8,
  0x02BC, 0x21,
  0x02BD, 0x0A,
  0x02BE, 0x1A,
  0x02BF, 0x40,
  0x02C0, 0xF9,
  0x02C1, 0xFE,
  0x02C2, 0xF0,
  0x02C3, 0x00,
  0x02C4, 0x48,
  0x02C5, 0x5F,
  0x02C6, 0x1A,
  0x02C7, 0x2D,
  0x02C8, 0x50,
  0x02C9, 0x45,
  0x02CA, 0x00,
  0x02CB, 0xA1,
  0x02CC, 0xE0,
  0x02CD, 0x02,
  0x02CE, 0x46,
  0x02CF, 0x29,
  0x02D0, 0xE0,
  0x02D1, 0x9D,
  0x02D2, 0xE0,
  0x02D3, 0x9C,
  0x02D4, 0x46,
  0x02D5, 0x20,
  0x02D6, 0xE0,
  0x02D7, 0xAB,
  0x02D8, 0xF8,
  0x02D9, 0xCE,
  0x02DA, 0xF0,
  0x02DB, 0x00,
  0x02DC, 0x20,
  0x02DD, 0x01,
  0x02DE, 0x49,
  0x02DF, 0x57,
  0x02E0, 0x47,
  0x02E1, 0x88,
  0x02E2, 0x68,
  0x02E3, 0x09,
  0x02E4, 0xF0,
  0x02E5, 0x00,
  0x02E6, 0x46,
  0x02E7, 0x20,
  0x02E8, 0x46,
  0x02E9, 0x07,
  0x02EA, 0xF8,
  0x02EB, 0xD7,
  0x02EC, 0x99,
  0x02ED, 0x00,
  0x02EE, 0x46,
  0x02EF, 0x20,
  0x02F0, 0xFF,
  0x02F1, 0x18,
  0x02F2, 0xF7,
  0x02F3, 0xFF,
  0x02F4, 0x1B,
  0x02F5, 0xC1,
  0x02F6, 0x90,
  0x02F7, 0x01,
  0x02F8, 0x42,
  0x02F9, 0x49,
  0x02FA, 0xD5,
  0x02FB, 0x00,
  0x02FC, 0x00,
  0x02FD, 0xA7,
  0x02FE, 0x48,
  0x02FF, 0x51,
  0x0300, 0x97,
  0x0301, 0x06,
  0x0302, 0x38,
  0x0303, 0x0C,
  0x0304, 0x9A,
  0x0305, 0x07,
  0x0306, 0x59,
  0x0307, 0xC0,
  0x0308, 0xF0,
  0x0309, 0x00,
  0x030A, 0x43,
  0x030B, 0x50,
  0x030C, 0x49,
  0x030D, 0x4D,
  0x030E, 0xF9,
  0x030F, 0xDF,
  0x0310, 0x39,
  0x0311, 0x0C,
  0x0312, 0x90,
  0x0313, 0x01,
  0x0314, 0x46,
  0x0315, 0x20,
  0x0316, 0x51,
  0x0317, 0xC8,
  0x0318, 0xF0,
  0x0319, 0x00,
  0x031A, 0x99,
  0x031B, 0x01,
  0x031C, 0x49,
  0x031D, 0x47,
  0x031E, 0xF8,
  0x031F, 0x9D,
  0x0320, 0x68,
  0x0321, 0x09,
  0x0322, 0x48,
  0x0323, 0x45,
  0x0324, 0x47,
  0x0325, 0x88,
  0x0326, 0x68,
  0x0327, 0x40,
  0x0328, 0xD0,
  0x0329, 0x20,
  0x032A, 0x2E,
  0x032B, 0x00,
  0x032C, 0xF0,
  0x032D, 0x00,
  0x032E, 0x46,
  0x032F, 0x20,
  0x0330, 0x46,
  0x0331, 0x07,
  0x0332, 0xF8,
  0x0333, 0xB3,
  0x0334, 0x99,
  0x0335, 0x05,
  0x0336, 0x46,
  0x0337, 0x20,
  0x0338, 0xFE,
  0x0339, 0xF4,
  0x033A, 0xF7,
  0x033B, 0xFF,
  0x033C, 0xD5,
  0x033D, 0x00,
  0x033E, 0x1A,
  0x033F, 0x39,
  0x0340, 0x10,
  0x0341, 0x49,
  0x0342, 0x42,
  0x0343, 0x49,
  0x0344, 0xD2,
  0x0345, 0x00,
  0x0346, 0x42,
  0x0347, 0x87,
  0x0348, 0x18,
  0x0349, 0x08,
  0x034A, 0x46,
  0x034B, 0x38,
  0x034C, 0x1A,
  0x034D, 0x40,
  0x034E, 0x99,
  0x034F, 0x02,
  0x0350, 0xF0,
  0x0351, 0x00,
  0x0352, 0x21,
  0x0353, 0x0A,
  0x0354, 0x1A,
  0x0355, 0x2D,
  0x0356, 0xF9,
  0x0357, 0xB5,
  0x0358, 0x99,
  0x0359, 0x06,
  0x035A, 0x48,
  0x035B, 0x3A,
  0x035C, 0x46,
  0x035D, 0x29,
  0x035E, 0x50,
  0x035F, 0x45,
  0x0360, 0xF0,
  0x0361, 0x00,
  0x0362, 0x46,
  0x0363, 0x20,
  0x0364, 0x49,
  0x0365, 0x35,
  0x0366, 0xF8,
  0x0367, 0x89,
  0x0368, 0x68,
  0x0369, 0x09,
  0x036A, 0x20,
  0x036B, 0x01,
  0x036C, 0x46,
  0x036D, 0x20,
  0x036E, 0x47,
  0x036F, 0x88,
  0x0370, 0xF8,
  0x0371, 0x92,
  0x0372, 0xF0,
  0x0373, 0x00,
  0x0374, 0x46,
  0x0375, 0x20,
  0x0376, 0x46,
  0x0377, 0x07,
  0x0378, 0xF7,
  0x0379, 0xFF,
  0x037A, 0x99,
  0x037B, 0x00,
  0x037C, 0x4A,
  0x037D, 0x31,
  0x037E, 0xFE,
  0x037F, 0xD3,
  0x0380, 0x3A,
  0x0381, 0x24,
  0x0382, 0x00,
  0x0383, 0xE1,
  0x0384, 0x18,
  0x0385, 0x89,
  0x0386, 0x50,
  0x0387, 0x50,
  0x0388, 0x2E,
  0x0389, 0x00,
  0x038A, 0x60,
  0x038B, 0x4F,
  0x038C, 0x1B,
  0x038D, 0xC1,
  0x038E, 0xD1,
  0x038F, 0x18,
  0x0390, 0x42,
  0x0391, 0x49,
  0x0392, 0xD5,
  0x0393, 0x00,
  0x0394, 0x42,
  0x0395, 0xB8,
  0x0396, 0x10,
  0x0397, 0x49,
  0x0398, 0x46,
  0x0399, 0x38,
  0x039A, 0xD3,
  0x039B, 0x00,
  0x039C, 0x99,
  0x039D, 0x02,
  0x039E, 0x18,
  0x039F, 0x08,
  0x03A0, 0x21,
  0x03A1, 0x0A,
  0x03A2, 0x1A,
  0x03A3, 0x40,
  0x03A4, 0xF9,
  0x03A5, 0x8C,
  0x03A6, 0xF0,
  0x03A7, 0x00,
  0x03A8, 0x48,
  0x03A9, 0x26,
  0x03AA, 0x1A,
  0x03AB, 0x2D,
  0x03AC, 0x50,
  0x03AD, 0x45,
  0x03AE, 0x99,
  0x03AF, 0x06,
  0x03B0, 0x46,
  0x03B1, 0x20,
  0x03B2, 0x46,
  0x03B3, 0x29,
  0x03B4, 0xF8,
  0x03B5, 0x60,
  0x03B6, 0xF0,
  0x03B7, 0x00,
  0x03B8, 0x20,
  0x03B9, 0x01,
  0x03BA, 0x49,
  0x03BB, 0x20,
  0x03BC, 0x47,
  0x03BD, 0x88,
  0x03BE, 0x68,
  0x03BF, 0x09,
  0x03C0, 0xD0,
  0x03C1, 0x27,
  0x03C2, 0x2C,
  0x03C3, 0x02,
  0x03C4, 0x46,
  0x03C5, 0x20,
  0x03C6, 0x21,
  0x03C7, 0x00,
  0x03C8, 0xFE,
  0x03C9, 0xAC,
  0x03CA, 0xF7,
  0x03CB, 0xFF,
  0x03CC, 0x46,
  0x03CD, 0x20,
  0x03CE, 0x46,
  0x03CF, 0x29,
  0x03D0, 0xF8,
  0x03D1, 0x52,
  0x03D2, 0xF0,
  0x03D3, 0x00,
  0x03D4, 0xD0,
  0x03D5, 0x20,
  0x03D6, 0x2C,
  0x03D7, 0x00,
  0x03D8, 0xD0,
  0x03D9, 0x21,
  0x03DA, 0x2C,
  0x03DB, 0x01,
  0x03DC, 0x30,
  0x03DD, 0x05,
  0x03DE, 0x20,
  0x03DF, 0xFF,
  0x03E0, 0x4F,
  0x03E1, 0x17,
  0x03E2, 0x1C,
  0x03E3, 0x41,
  0x03E4, 0x54,
  0x03E5, 0x3B,
  0x03E6, 0x0A,
  0x03E7, 0x2B,
  0x03E8, 0x46,
  0x03E9, 0x20,
  0x03EA, 0x54,
  0x03EB, 0x7D,
  0x03EC, 0xF0,
  0x03ED, 0x00,
  0x03EE, 0x99,
  0x03EF, 0x01,
  0x03F0, 0x2C,
  0x03F1, 0x00,
  0x03F2, 0xF8,
  0x03F3, 0x33,
  0x03F4, 0x2C,
  0x03F5, 0x01,
  0x03F6, 0xD0,
  0x03F7, 0x17,
  0x03F8, 0x20,
  0x03F9, 0xFF,
  0x03FA, 0xD0,
  0x03FB, 0x17,
  0x03FC, 0x99,
  0x03FD, 0x01,
  0x03FE, 0x30,
  0x03FF, 0x03,
  0x0400, 0x21,
  0x0401, 0x01,
  0x0402, 0x54,
  0x0403, 0x39,
  0x0404, 0x2C,
  0x0405, 0x02,
  0x0406, 0x1C,
  0x0407, 0x64,
  0x0408, 0xE7,
  0x0409, 0x0D,
  0x040A, 0xD2,
  0x040B, 0x00,
  0x040C, 0xD0,
  0x040D, 0x0E,
  0x040E, 0x2E,
  0x040F, 0x00,
  0x0410, 0xE0,
  0x0411, 0x0D,
  0x0412, 0x98,
  0x0413, 0x04,
  0x0414, 0x02,
  0x0415, 0x49,
  0x0416, 0x21,
  0x0417, 0x01,
  0x0418, 0x20,
  0x0419, 0xA4,
  0x041A, 0xE7,
  0x041B, 0xD5,
  0x041C, 0xE7,
  0x041D, 0xE0,
  0x041E, 0x21,
  0x041F, 0xA5,
  0x0420, 0x21,
  0x0421, 0xA7,
  0x0422, 0x20,
  0x0423, 0xA6,
  0x0424, 0x20,
  0x0425, 0xA2,
  0x0426, 0xE7,
  0x0427, 0xDD,
  0x0428, 0x20,
  0x0429, 0xA3,
  0x042A, 0xE7,
  0x042B, 0xE9,
  0x042C, 0x98,
  0x042D, 0x03,
  0x042E, 0xE7,
  0x042F, 0xE7,
  0x0430, 0xB0,
  0x0431, 0x0B,
  0x0432, 0x70,
  0x0433, 0x41,
  0x0434, 0x00,
  0x0435, 0x00,
  0x0436, 0xBD,
  0x0437, 0xF0,
  0x0438, 0x10,
  0x0439, 0x00,
  0x043A, 0x0F,
  0x043B, 0x84,
  0x043C, 0x10,
  0x043D, 0x00,
  0x043E, 0x0E,
  0x043F, 0xF8,
  0x0440, 0x10,
  0x0441, 0x00,
  0x0442, 0x19,
  0x0443, 0x40,
  0x0444, 0x10,
  0x0445, 0x00,
  0x0446, 0x0F,
  0x0447, 0xBC,
  0x0448, 0x00,
  0x0449, 0x00,
  0x044A, 0x07,
  0x044B, 0x7B,
  0x044C, 0x00,
  0x044D, 0x00,
  0x044E, 0x03,
  0x044F, 0xFF,
  0x0450, 0x10,
  0x0451, 0x00,
  0x0452, 0x0E,
  0x0453, 0xF4,
  0x0454, 0x10,
  0x0455, 0x00,
  0x0456, 0x0E,
  0x0457, 0xF0,
  0x0458, 0xD0,
  0x0459, 0x09,
  0x045A, 0x28,
  0x045B, 0x00,
  0x045C, 0xD0,
  0x045D, 0x09,
  0x045E, 0x28,
  0x045F, 0x01,
  0x0460, 0x05,
  0x0461, 0x89,
  0x0462, 0x20,
  0x0463, 0x44,
  0x0464, 0x0D,
  0x0465, 0x89,
  0x0466, 0x22,
  0x0467, 0x19,
  0x0468, 0x18,
  0x0469, 0x80,
  0x046A, 0x06,
  0x046B, 0xD2,
  0x046C, 0x47,
  0x046D, 0x70,
  0x046E, 0x60,
  0x046F, 0x01,
  0x0470, 0xE7,
  0x0471, 0xF6,
  0x0472, 0x20,
  0x0473, 0x18,
  0x0474, 0xE7,
  0x0475, 0xF4,
  0x0476, 0x20,
  0x0477, 0x1C,
  0x0478, 0xD0,
  0x0479, 0x09,
  0x047A, 0x28,
  0x047B, 0x00,
  0x047C, 0xD0,
  0x047D, 0x09,
  0x047E, 0x28,
  0x047F, 0x01,
  0x0480, 0x05,
  0x0481, 0x89,
  0x0482, 0x20,
  0x0483, 0x48,
  0x0484, 0x0D,
  0x0485, 0x89,
  0x0486, 0x22,
  0x0487, 0x19,
  0x0488, 0x18,
  0x0489, 0x80,
  0x048A, 0x06,
  0x048B, 0xD2,
  0x048C, 0x47,
  0x048D, 0x70,
  0x048E, 0x60,
  0x048F, 0x01,
  0x0490, 0xE7,
  0x0491, 0xF6,
  0x0492, 0x20,
  0x0493, 0x20,
  0x0494, 0xE7,
  0x0495, 0xF4,
  0x0496, 0x20,
  0x0497, 0x24,
  0x0498, 0xD0,
  0x0499, 0x07,
  0x049A, 0x28,
  0x049B, 0x00,
  0x049C, 0xD0,
  0x049D, 0x07,
  0x049E, 0x28,
  0x049F, 0x01,
  0x04A0, 0x21,
  0x04A1, 0x19,
  0x04A2, 0x20,
  0x04A3, 0x68,
  0x04A4, 0x18,
  0x04A5, 0x40,
  0x04A6, 0x06,
  0x04A7, 0xC9,
  0x04A8, 0x47,
  0x04A9, 0x70,
  0x04AA, 0x68,
  0x04AB, 0x00,
  0x04AC, 0xE7,
  0x04AD, 0xF8,
  0x04AE, 0x20,
  0x04AF, 0x60,
  0x04B0, 0xE7,
  0x04B1, 0xF6,
  0x04B2, 0x20,
  0x04B3, 0x64,
  0x04B4, 0xD0,
  0x04B5, 0x04,
  0x04B6, 0x28,
  0x04B7, 0x00,
  0x04B8, 0xD0,
  0x04B9, 0x04,
  0x04BA, 0x28,
  0x04BB, 0x01,
  0x04BC, 0x49,
  0x04BD, 0x03,
  0x04BE, 0x20,
  0x04BF, 0x24,
  0x04C0, 0x20,
  0x04C1, 0x02,
  0x04C2, 0x47,
  0x04C3, 0x08,
  0x04C4, 0x20,
  0x04C5, 0x00,
  0x04C6, 0xE7,
  0x04C7, 0xFB,
  0x04C8, 0x00,
  0x04C9, 0x00,
  0x04CA, 0xE7,
  0x04CB, 0xF9,
  0x04CC, 0x00,
  0x04CD, 0x00,
  0x04CE, 0x07,
  0x04CF, 0x89,
  0x04D0, 0x47,
  0x04D1, 0x70,
  0x04D2, 0xBA,
  0x04D3, 0x40,
  0x04D4, 0x4D,
  0x04D5, 0x6C,
  0x04D6, 0xB5,
  0x04D7, 0xF8,
  0x04D8, 0x4E,
  0x04D9, 0x6B,
  0x04DA, 0x46,
  0x04DB, 0x04,
  0x04DC, 0xF0,
  0x04DD, 0x00,
  0x04DE, 0x35,
  0x04DF, 0xB4,
  0x04E0, 0x4F,
  0x04E1, 0x6A,
  0x04E2, 0xF9,
  0x04E3, 0xC3,
  0x04E4, 0xD0,
  0x04E5, 0x15,
  0x04E6, 0x2C,
  0x04E7, 0x00,
  0x04E8, 0x60,
  0x04E9, 0x38,
  0x04EA, 0x6B,
  0x04EB, 0x60,
  0x04EC, 0x60,
  0x04ED, 0x78,
  0x04EE, 0x6B,
  0x04EF, 0xA0,
  0x04F0, 0x60,
  0x04F1, 0xB8,
  0x04F2, 0x69,
  0x04F3, 0x60,
  0x04F4, 0x60,
  0x04F5, 0xF8,
  0x04F6, 0x68,
  0x04F7, 0xA0,
  0x04F8, 0x61,
  0x04F9, 0x38,
  0x04FA, 0x68,
  0x04FB, 0xE0,
  0x04FC, 0x61,
  0x04FD, 0x78,
  0x04FE, 0x69,
  0x04FF, 0x20,
  0x0500, 0x61,
  0x0501, 0xB8,
  0x0502, 0x6B,
  0x0503, 0x20,
  0x0504, 0x61,
  0x0505, 0xF8,
  0x0506, 0x69,
  0x0507, 0xE0,
  0x0508, 0x62,
  0x0509, 0x38,
  0x050A, 0x6A,
  0x050B, 0xA0,
  0x050C, 0x62,
  0x050D, 0x78,
  0x050E, 0x6A,
  0x050F, 0x20,
  0x0510, 0x62,
  0x0511, 0xB8,
  0x0512, 0x6A,
  0x0513, 0xE0,
  0x0514, 0x78,
  0x0515, 0x00,
  0x0516, 0x48,
  0x0517, 0x5E,
  0x0518, 0xD1,
  0x0519, 0x0B,
  0x051A, 0x28,
  0x051B, 0x19,
  0x051C, 0x69,
  0x051D, 0x7A,
  0x051E, 0x21,
  0x051F, 0x1E,
  0x0520, 0x47,
  0x0521, 0x90,
  0x0522, 0x20,
  0x0523, 0x00,
  0x0524, 0x21,
  0x0525, 0x1D,
  0x0526, 0x02,
  0x0527, 0x04,
  0x0528, 0x20,
  0x0529, 0x00,
  0x052A, 0x69,
  0x052B, 0x7A,
  0x052C, 0x49,
  0x052D, 0x59,
  0x052E, 0x47,
  0x052F, 0x90,
  0x0530, 0x60,
  0x0531, 0x88,
  0x0532, 0x43,
  0x0533, 0x20,
  0x0534, 0x21,
  0x0535, 0x20,
  0x0536, 0x48,
  0x0537, 0x58,
  0x0538, 0x4F,
  0x0539, 0x54,
  0x053A, 0x81,
  0x053B, 0x01,
  0x053C, 0x69,
  0x053D, 0x7A,
  0x053E, 0x21,
  0x053F, 0x02,
  0x0540, 0x47,
  0x0541, 0x90,
  0x0542, 0x20,
  0x0543, 0x01,
  0x0544, 0x06,
  0x0545, 0xC9,
  0x0546, 0x21,
  0x0547, 0x19,
  0x0548, 0x49,
  0x0549, 0x54,
  0x054A, 0x63,
  0x054B, 0xC8,
  0x054C, 0x61,
  0x054D, 0x08,
  0x054E, 0x20,
  0x054F, 0x00,
  0x0550, 0x31,
  0x0551, 0x80,
  0x0552, 0x49,
  0x0553, 0x52,
  0x0554, 0x49,
  0x0555, 0x51,
  0x0556, 0x62,
  0x0557, 0x48,
  0x0558, 0x39,
  0x0559, 0xC0,
  0x055A, 0x48,
  0x055B, 0x51,
  0x055C, 0x48,
  0x055D, 0x4C,
  0x055E, 0x61,
  0x055F, 0x88,
  0x0560, 0x21,
  0x0561, 0x0C,
  0x0562, 0x78,
  0x0563, 0x04,
  0x0564, 0x20,
  0x0565, 0x01,
  0x0566, 0x69,
  0x0567, 0x7A,
  0x0568, 0x07,
  0x0569, 0x07,
  0x056A, 0x47,
  0x056B, 0x90,
  0x056C, 0x0F,
  0x056D, 0x3F,
  0x056E, 0x06,
  0x056F, 0x00,
  0x0570, 0x28,
  0x0571, 0x02,
  0x0572, 0x0F,
  0x0573, 0x00,
  0x0574, 0x2C,
  0x0575, 0x13,
  0x0576, 0xD1,
  0x0577, 0x28,
  0x0578, 0x2C,
  0x0579, 0x41,
  0x057A, 0xD0,
  0x057B, 0x04,
  0x057C, 0x2C,
  0x057D, 0x42,
  0x057E, 0xD0,
  0x057F, 0x06,
  0x0580, 0xE0,
  0x0581, 0x0C,
  0x0582, 0xD0,
  0x0583, 0x09,
  0x0584, 0x4E,
  0x0585, 0x47,
  0x0586, 0x4D,
  0x0587, 0x47,
  0x0588, 0xE0,
  0x0589, 0x08,
  0x058A, 0x35,
  0x058B, 0xB4,
  0x058C, 0x4D,
  0x058D, 0x45,
  0x058E, 0x4E,
  0x058F, 0x45,
  0x0590, 0x35,
  0x0591, 0xBC,
  0x0592, 0x36,
  0x0593, 0x08,
  0x0594, 0x4E,
  0x0595, 0x3C,
  0x0596, 0xE0,
  0x0597, 0x03,
  0x0598, 0x36,
  0x0599, 0x14,
  0x059A, 0x4D,
  0x059B, 0x42,
  0x059C, 0x4C,
  0x059D, 0x3B,
  0x059E, 0x3D,
  0x059F, 0x84,
  0x05A0, 0x69,
  0x05A1, 0x62,
  0x05A2, 0x21,
  0x05A3, 0x0D,
  0x05A4, 0x47,
  0x05A5, 0x90,
  0x05A6, 0x20,
  0x05A7, 0x01,
  0x05A8, 0x47,
  0x05A9, 0x88,
  0x05AA, 0x69,
  0x05AB, 0xA1,
  0x05AC, 0x47,
  0x05AD, 0xB0,
  0x05AE, 0x20,
  0x05AF, 0x0B,
  0x05B0, 0x47,
  0x05B1, 0xA8,
  0x05B2, 0x20,
  0x05B3, 0x13,
  0x05B4, 0x47,
  0x05B5, 0xB0,
  0x05B6, 0x20,
  0x05B7, 0x0C,
  0x05B8, 0x47,
  0x05B9, 0xA8,
  0x05BA, 0x20,
  0x05BB, 0x80,
  0x05BC, 0xD1,
  0x05BD, 0x03,
  0x05BE, 0x2F,
  0x05BF, 0x01,
  0x05C0, 0x47,
  0x05C1, 0xB0,
  0x05C2, 0x20,
  0x05C3, 0x1F,
  0x05C4, 0x47,
  0x05C5, 0xA8,
  0x05C6, 0x20,
  0x05C7, 0x08,
  0x05C8, 0xBD,
  0x05C9, 0xF8,
  0x05CA, 0x20,
  0x05CB, 0x00,
  0x05CC, 0x4B,
  0x05CD, 0x2F,
  0x05CE, 0xB5,
  0x05CF, 0x00,
  0x05D0, 0x6A,
  0x05D1, 0xD9,
  0x05D2, 0xB0,
  0x05D3, 0x87,
  0x05D4, 0xD0,
  0x05D5, 0x01,
  0x05D6, 0x28,
  0x05D7, 0x00,
  0x05D8, 0xD1,
  0x05D9, 0x1B,
  0x05DA, 0x28,
  0x05DB, 0x04,
  0x05DC, 0xD1,
  0x05DD, 0x19,
  0x05DE, 0x29,
  0x05DF, 0x01,
  0x05E0, 0x07,
  0x05E1, 0x89,
  0x05E2, 0x20,
  0x05E3, 0x00,
  0x05E4, 0x07,
  0x05E5, 0x89,
  0x05E6, 0x68,
  0x05E7, 0x09,
  0x05E8, 0x49,
  0x05E9, 0x2F,
  0x05EA, 0xD0,
  0x05EB, 0x0E,
  0x05EC, 0x49,
  0x05ED, 0x2F,
  0x05EE, 0x91,
  0x05EF, 0x04,
  0x05F0, 0x90,
  0x05F1, 0x01,
  0x05F2, 0x91,
  0x05F3, 0x05,
  0x05F4, 0x90,
  0x05F5, 0x02,
  0x05F6, 0x49,
  0x05F7, 0x2E,
  0x05F8, 0x90,
  0x05F9, 0x03,
  0x05FA, 0x91,
  0x05FB, 0x06,
  0x05FC, 0x22,
  0x05FD, 0x01,
  0x05FE, 0x68,
  0x05FF, 0x1B,
  0x0600, 0xA8,
  0x0601, 0x04,
  0x0602, 0xA9,
  0x0603, 0x01,
  0x0604, 0xE0,
  0x0605, 0x05,
  0x0606, 0x47,
  0x0607, 0x98,
  0x0608, 0x61,
  0x0609, 0x08,
  0x060A, 0x49,
  0x060B, 0x24,
  0x060C, 0x63,
  0x060D, 0x88,
  0x060E, 0x49,
  0x060F, 0x29,
  0x0610, 0x63,
  0x0611, 0x88,
  0x0612, 0x49,
  0x0613, 0x29,
  0x0614, 0xB0,
  0x0615, 0x07,
  0x0616, 0x20,
  0x0617, 0x00,
  0x0618, 0xB5,
  0x0619, 0x70,
  0x061A, 0xBD,
  0x061B, 0x00,
  0x061C, 0x4C,
  0x061D, 0x27,
  0x061E, 0x49,
  0x061F, 0x1C,
  0x0620, 0x78,
  0x0621, 0x09,
  0x0622, 0x31,
  0x0623, 0x20,
  0x0624, 0x07,
  0x0625, 0x49,
  0x0626, 0xB0,
  0x0627, 0x88,
  0x0628, 0x4D,
  0x0629, 0x18,
  0x062A, 0x0F,
  0x062B, 0x49,
  0x062C, 0x62,
  0x062D, 0xE9,
  0x062E, 0x28,
  0x062F, 0x01,
  0x0630, 0x29,
  0x0631, 0x00,
  0x0632, 0xD1,
  0x0633, 0x26,
  0x0634, 0x29,
  0x0635, 0x04,
  0x0636, 0xD0,
  0x0637, 0x01,
  0x0638, 0x20,
  0x0639, 0x0E,
  0x063A, 0xD1,
  0x063B, 0x22,
  0x063C, 0x46,
  0x063D, 0x06,
  0x063E, 0x47,
  0x063F, 0xA0,
  0x0640, 0x47,
  0x0641, 0xA0,
  0x0642, 0x20,
  0x0643, 0x0C,
  0x0644, 0x43,
  0x0645, 0x30,
  0x0646, 0x04,
  0x0647, 0x00,
  0x0648, 0x07,
  0x0649, 0x89,
  0x064A, 0x21,
  0x064B, 0x01,
  0x064C, 0x07,
  0x064D, 0x8A,
  0x064E, 0x68,
  0x064F, 0x09,
  0x0650, 0x2A,
  0x0651, 0x00,
  0x0652, 0x21,
  0x0653, 0x00,
  0x0654, 0x4A,
  0x0655, 0x14,
  0x0656, 0xD0,
  0x0657, 0x0E,
  0x0658, 0x4A,
  0x0659, 0x14,
  0x065A, 0x92,
  0x065B, 0x04,
  0x065C, 0x4A,
  0x065D, 0x14,
  0x065E, 0x92,
  0x065F, 0x05,
  0x0660, 0x91,
  0x0661, 0x02,
  0x0662, 0x92,
  0x0663, 0x06,
  0x0664, 0x90,
  0x0665, 0x01,
  0x0666, 0x91,
  0x0667, 0x03,
  0x0668, 0x22,
  0x0669, 0x03,
  0x066A, 0x68,
  0x066B, 0x2B,
  0x066C, 0xA8,
  0x066D, 0x04,
  0x066E, 0xA9,
  0x066F, 0x01,
  0x0670, 0xE0,
  0x0671, 0x05,
  0x0672, 0x47,
  0x0673, 0x98,
  0x0674, 0x61,
  0x0675, 0x10,
  0x0676, 0x4A,
  0x0677, 0x09,
  0x0678, 0x63,
  0x0679, 0x81,
  0x067A, 0x48,
  0x067B, 0x0E,
  0x067C, 0x63,
  0x067D, 0x81,
  0x067E, 0x48,
  0x067F, 0x0E,
  0x0680, 0xB0,
  0x0681, 0x08,
  0x0682, 0x20,
  0x0683, 0x00,
  0x0684, 0x47,
  0x0685, 0x70,
  0x0686, 0xBD,
  0x0687, 0x70,
  0x0688, 0x00,
  0x0689, 0x00,
  0x068A, 0x29,
  0x068B, 0xD9,
  0x068C, 0x10,
  0x068D, 0x00,
  0x068E, 0x0E,
  0x068F, 0xE0,
  0x0690, 0x10,
  0x0691, 0x00,
  0x0692, 0x19,
  0x0693, 0x40,
  0x0694, 0xC8,
  0x0695, 0x00,
  0x0696, 0x01,
  0x0697, 0x80,
  0x0698, 0x10,
  0x0699, 0x00,
  0x069A, 0x1A,
  0x069B, 0x80,
  0x069C, 0x50,
  0x069D, 0x00,
  0x069E, 0x60,
  0x069F, 0xC0,
  0x06A0, 0x30,
  0x06A1, 0x48,
  0x06A2, 0x30,
  0x06A3, 0x48,
  0x06A4, 0x00,
  0x06A5, 0x00,
  0x06A6, 0x2B,
  0x06A7, 0x25,
  0x06A8, 0x00,
  0x06A9, 0x00,
  0x06AA, 0x60,
  0x06AB, 0xD0,
  0x06AC, 0x00,
  0x06AD, 0x00,
  0x06AE, 0x41,
  0x06AF, 0x38,
  0x06B0, 0x00,
  0x06B1, 0x00,
  0x06B2, 0x51,
  0x06B3, 0x38,
  0x06B4, 0x50,
  0x06B5, 0x00,
  0x06B6, 0x41,
  0x06B7, 0x00,
  0x06B8, 0x50,
  0x06B9, 0x00,
  0x06BA, 0x51,
  0x06BB, 0x00,
  0x06BC, 0x00,
  0x06BD, 0x00,
  0x06BE, 0x07,
  0x06BF, 0x7B,
  0x06C0, 0x48,
  0x06C1, 0x01,
  0x06C2, 0xB4,
  0x06C3, 0x03,
  0x06C4, 0xBD,
  0x06C5, 0x01,
  0x06C6, 0x90,
  0x06C7, 0x01,
  0x06C8, 0x00,
  0x06C9, 0x00,
  0x06CA, 0x00,
  0x06CB, 0xD5,
  0x06CC, 0x48,
  0x06CD, 0x01,
  0x06CE, 0xB4,
  0x06CF, 0x03,
  0x06D0, 0xBD,
  0x06D1, 0x01,
  0x06D2, 0x90,
  0x06D3, 0x01,
  0x06D4, 0x00,
  0x06D5, 0x00,
  0x06D6, 0x00,
  0x06D7, 0xC1,
  0x06D8, 0x46,
  0x06D9, 0x74,
  0x06DA, 0xB4,
  0x06DB, 0x30,
  0x06DC, 0x78,
  0x06DD, 0x25,
  0x06DE, 0x1E,
  0x06DF, 0x64,
  0x06E0, 0x42,
  0x06E1, 0xAB,
  0x06E2, 0x1C,
  0x06E3, 0x64,
  0x06E4, 0x46,
  0x06E5, 0x1D,
  0x06E6, 0xD2,
  0x06E7, 0x00,
  0x06E8, 0x00,
  0x06E9, 0x5B,
  0x06EA, 0x5D,
  0x06EB, 0x63,
  0x06EC, 0xBC,
  0x06ED, 0x30,
  0x06EE, 0x18,
  0x06EF, 0xE3,
  0x06F0, 0x00,
  0x06F1, 0x00,
  0x06F2, 0x47,
  0x06F3, 0x18,
  0x06F4, 0x2A,
  0x06F5, 0x05,
  0x06F6, 0xB5,
  0x06F7, 0xF0,
  0x06F8, 0x20,
  0x06F9, 0x00,
  0x06FA, 0xDD,
  0x06FB, 0x02,
  0x06FC, 0xBD,
  0x06FD, 0xF0,
  0x06FE, 0x43,
  0x06FF, 0xC0,
  0x0700, 0x23,
  0x0701, 0x00,
  0x0702, 0xB6,
  0x0703, 0x72,
  0x0704, 0xE0,
  0x0705, 0x0B,
  0x0706, 0x46,
  0x0707, 0x1C,
  0x0708, 0x4E,
  0x0709, 0x09,
  0x070A, 0x00,
  0x070B, 0x9D,
  0x070C, 0x19,
  0x070D, 0xAE,
  0x070E, 0x59,
  0x070F, 0x47,
  0x0710, 0x59,
  0x0711, 0x4D,
  0x0712, 0x60,
  0x0713, 0x77,
  0x0714, 0x1D,
  0x0715, 0x5E,
  0x0716, 0x62,
  0x0717, 0xF5,
  0x0718, 0x40,
  0x0719, 0xB5,
  0x071A, 0x25,
  0x071B, 0x01,
  0x071C, 0x1C,
  0x071D, 0x5B,
  0x071E, 0x43,
  0x071F, 0x2C,
  0x0720, 0xDB,
  0x0721, 0xF1,
  0x0722, 0x42,
  0x0723, 0x93,
  0x0724, 0x07,
  0x0725, 0x80,
  0x0726, 0x20,
  0x0727, 0x01,
  0x0728, 0xB6,
  0x0729, 0x62,
  0x072A, 0x62,
  0x072B, 0x44,
  0x072C, 0xBD,
  0x072D, 0xF0,
  0x072E, 0x20,
  0x072F, 0x00,
  0x0730, 0x40,
  0x0731, 0x00,
  0x0732, 0x00,
  0x0733, 0x40,
  0x0734, 0x28,
  0x0735, 0x02,
  0x0736, 0xB5,
  0x0737, 0x1F,
  0x0738, 0x28,
  0x0739, 0x01,
  0x073A, 0xD0,
  0x073B, 0x24,
  0x073C, 0x48,
  0x073D, 0x16,
  0x073E, 0xD0,
  0x073F, 0x27,
  0x0740, 0x48,
  0x0741, 0x15,
  0x0742, 0x90,
  0x0743, 0x00,
  0x0744, 0x4B,
  0x0745, 0x15,
  0x0746, 0x30,
  0x0747, 0x08,
  0x0748, 0x68,
  0x0749, 0x1B,
  0x074A, 0x90,
  0x074B, 0x01,
  0x074C, 0xA9,
  0x074D, 0x02,
  0x074E, 0x22,
  0x074F, 0x02,
  0x0750, 0x47,
  0x0751, 0x98,
  0x0752, 0x46,
  0x0753, 0x68,
  0x0754, 0x99,
  0x0755, 0x03,
  0x0756, 0x98,
  0x0757, 0x02,
  0x0758, 0x15,
  0x0759, 0x00,
  0x075A, 0x01,
  0x075B, 0x00,
  0x075C, 0x43,
  0x075D, 0xC0,
  0x075E, 0xD5,
  0x075F, 0x01,
  0x0760, 0x4A,
  0x0761, 0x0F,
  0x0762, 0x1C,
  0x0763, 0x40,
  0x0764, 0x68,
  0x0765, 0x93,
  0x0766, 0x05,
  0x0767, 0x00,
  0x0768, 0x18,
  0x0769, 0x18,
  0x076A, 0x0D,
  0x076B, 0x00,
  0x076C, 0x01,
  0x076D, 0x08,
  0x076E, 0x60,
  0x076F, 0x90,
  0x0770, 0xD5,
  0x0771, 0x01,
  0x0772, 0x15,
  0x0773, 0x00,
  0x0774, 0x1C,
  0x0775, 0x40,
  0x0776, 0x43,
  0x0777, 0xC0,
  0x0778, 0x68,
  0x0779, 0xD1,
  0x077A, 0x05,
  0x077B, 0x00,
  0x077C, 0x18,
  0x077D, 0x08,
  0x077E, 0x0D,
  0x077F, 0x00,
  0x0780, 0xBD,
  0x0781, 0x1F,
  0x0782, 0x60,
  0x0783, 0xD0,
  0x0784, 0x90,
  0x0785, 0x00,
  0x0786, 0x48,
  0x0787, 0x07,
  0x0788, 0x30,
  0x0789, 0x08,
  0x078A, 0x48,
  0x078B, 0x06,
  0x078C, 0x48,
  0x078D, 0x06,
  0x078E, 0xE7,
  0x078F, 0xDB,
  0x0790, 0x48,
  0x0791, 0x05,
  0x0792, 0x90,
  0x0793, 0x00,
  0x0794, 0xE7,
  0x0795, 0xD6,
  0x0796, 0x30,
  0x0797, 0x08,
  0x0798, 0x00,
  0x0799, 0x00,
  0x079A, 0x41,
  0x079B, 0x04,
  0x079C, 0x10,
  0x079D, 0x00,
  0x079E, 0x0E,
  0x079F, 0xE4,
  0x07A0, 0x10,
  0x07A1, 0x00,
  0x07A2, 0x0F,
  0x07A3, 0x10,
  0x07A4, 0x00,
  0x07A5, 0x00,
  0x07A6, 0x43,
  0x07A7, 0x40,
  0x07A8, 0x00,
  0x07A9, 0x00,
  0x07AA, 0x51,
  0x07AB, 0x04,
  0x07AC, 0x4E,
  0x07AD, 0x18,
  0x07AE, 0xB5,
  0x07AF, 0xF3,
  0x07B0, 0x68,
  0x07B1, 0xF4,
  0x07B2, 0xB0,
  0x07B3, 0x81,
  0x07B4, 0x2C,
  0x07B5, 0x00,
  0x07B6, 0x68,
  0x07B7, 0xB7,
  0x07B8, 0x12,
  0x07B9, 0xA1,
  0x07BA, 0xD0,
  0x07BB, 0x26,
  0x07BC, 0xF7,
  0x07BD, 0xFF,
  0x07BE, 0x46,
  0x07BF, 0x38,
  0x07C0, 0x46,
  0x07C1, 0x05,
  0x07C2, 0xFF,
  0x07C3, 0x7F,
  0x07C4, 0x28,
  0x07C5, 0x02,
  0x07C6, 0x98,
  0x07C7, 0x01,
  0x07C8, 0xD0,
  0x07C9, 0x16,
  0x07CA, 0x48,
  0x07CB, 0x12,
  0x07CC, 0x7C,
  0x07CD, 0x80,
  0x07CE, 0x30,
  0x07CF, 0x80,
  0x07D0, 0x09,
  0x07D1, 0x00,
  0x07D2, 0x49,
  0x07D3, 0x0F,
  0x07D4, 0x31,
  0x07D5, 0x54,
  0x07D6, 0x00,
  0x07D7, 0x40,
  0x07D8, 0x46,
  0x07D9, 0x20,
  0x07DA, 0x5A,
  0x07DB, 0x09,
  0x07DC, 0x46,
  0x07DD, 0x39,
  0x07DE, 0x43,
  0x07DF, 0x48,
  0x07E0, 0xFF,
  0x07E1, 0x74,
  0x07E2, 0xF7,
  0x07E3, 0xFF,
  0x07E4, 0x60,
  0x07E5, 0x08,
  0x07E6, 0x99,
  0x07E7, 0x02,
  0x07E8, 0x88,
  0x07E9, 0x01,
  0x07EA, 0x69,
  0x07EB, 0x30,
  0x07EC, 0x42,
  0x07ED, 0x85,
  0x07EE, 0x88,
  0x07EF, 0x40,
  0x07F0, 0x42,
  0x07F1, 0x8D,
  0x07F2, 0xD9,
  0x07F3, 0x07,
  0x07F4, 0x20,
  0x07F5, 0x00,
  0x07F6, 0xD2,
  0x07F7, 0x05,
  0x07F8, 0x30,
  0x07F9, 0xFF,
  0x07FA, 0xBD,
  0x07FB, 0xFE,
  0x07FC, 0x7B,
  0x07FD, 0x00,
  0x07FE, 0x30,
  0x07FF, 0x01,
  0x0800, 0x48,
  0x0801, 0x05,
  0x0802, 0xE7,
  0x0803, 0xE6,
  0x0804, 0xD3,
  0x0805, 0x01,
  0x0806, 0x42,
  0x0807, 0x85,
  0x0808, 0xBD,
  0x0809, 0xFE,
  0x080A, 0x20,
  0x080B, 0x02,
  0x080C, 0xBD,
  0x080D, 0xFE,
  0x080E, 0x20,
  0x080F, 0x01,
  0x0810, 0x10,
  0x0811, 0x00,
  0x0812, 0x0F,
  0x0813, 0x10,
  0x0814, 0x10,
  0x0815, 0x00,
  0x0816, 0x19,
  0x0817, 0x40,
  0x0818, 0x00,
  0x0819, 0x00,
  0x081A, 0x0C,
  0x081B, 0xA6,
  0x081C, 0x48,
  0x081D, 0x10,
  0x081E, 0xB5,
  0x081F, 0x70,
  0x0820, 0x4E,
  0x0821, 0x0F,
  0x0822, 0x69,
  0x0823, 0xC5,
  0x0824, 0x36,
  0x0825, 0x98,
  0x0826, 0x4B,
  0x0827, 0x0E,
  0x0828, 0x3B,
  0x0829, 0x18,
  0x082A, 0x1C,
  0x082B, 0x71,
  0x082C, 0x1D,
  0x082D, 0x1A,
  0x082E, 0x1C,
  0x082F, 0xCC,
  0x0830, 0x1D,
  0x0831, 0x10,
  0x0832, 0x78,
  0x0833, 0x36,
  0x0834, 0xD0,
  0x0835, 0x11,
  0x0836, 0x2E,
  0x0837, 0x00,
  0x0838, 0x78,
  0x0839, 0x09,
  0x083A, 0x68,
  0x083B, 0x24,
  0x083C, 0x06,
  0x083D, 0x64,
  0x083E, 0x19,
  0x083F, 0x64,
  0x0840, 0x29,
  0x0841, 0x00,
  0x0842, 0x0E,
  0x0843, 0x64,
  0x0844, 0x2C,
  0x0845, 0x10,
  0x0846, 0xD0,
  0x0847, 0x0A,
  0x0848, 0x49,
  0x0849, 0x06,
  0x084A, 0xD1,
  0x084B, 0x08,
  0x084C, 0x2C,
  0x084D, 0x00,
  0x084E, 0x7F,
  0x084F, 0xCC,
  0x0850, 0x68,
  0x0851, 0x00,
  0x0852, 0xD0,
  0x0853, 0x04,
  0x0854, 0x60,
  0x0855, 0x10,
  0x0856, 0x60,
  0x0857, 0x18,
  0x0858, 0x77,
  0x0859, 0xC8,
  0x085A, 0x20,
  0x085B, 0x00,
  0x085C, 0x00,
  0x085D, 0x00,
  0x085E, 0xBD,
  0x085F, 0x70,
  0x0860, 0x10,
  0x0861, 0x00,
  0x0862, 0x17,
  0x0863, 0x40,
  0x0864, 0x10,
  0x0865, 0x00,
  0x0866, 0x1A,
  0x0867, 0x60,
  0x0868, 0x49,
  0x0869, 0x03,
  0x086A, 0x48,
  0x086B, 0x04,
  0x086C, 0x49,
  0x086D, 0x04,
  0x086E, 0x60,
  0x086F, 0x41,
  0x0870, 0x49,
  0x0871, 0x04,
  0x0872, 0x60,
  0x0873, 0x81,
  0x0874, 0x47,
  0x0875, 0x70,
  0x0876, 0x62,
  0x0877, 0x41,
  0x0878, 0x10,
  0x0879, 0x00,
  0x087A, 0x08,
  0x087B, 0x89,
  0x087C, 0x10,
  0x087D, 0x00,
  0x087E, 0x1B,
  0x087F, 0x48,
  0x0880, 0x10,
  0x0881, 0x00,
  0x0882, 0x08,
  0x0883, 0xA5,
  0x0884, 0x10,
  0x0885, 0x00,
  0x0886, 0x08,
  0x0887, 0x1D,
  0x0888, 0x48,
  0x0889, 0x04,
  0x088A, 0xB5,
  0x088B, 0x10,
  0x088C, 0x28,
  0x088D, 0x01,
  0x088E, 0x78,
  0x088F, 0x00,
  0x0890, 0x48,
  0x0891, 0x03,
  0x0892, 0xD1,
  0x0893, 0x03,
  0x0894, 0xF0,
  0x0895, 0x00,
  0x0896, 0x78,
  0x0897, 0x00,
  0x0898, 0xBD,
  0x0899, 0x10,
  0x089A, 0xFA,
  0x089B, 0x6B,
  0x089C, 0x10,
  0x089D, 0x00,
  0x089E, 0x0F,
  0x089F, 0x8C,
  0x08A0, 0x10,
  0x08A1, 0x00,
  0x08A2, 0x0F,
  0x08A3, 0x13,
  0x08A4, 0x78,
  0x08A5, 0x01,
  0x08A6, 0x48,
  0x08A7, 0x06,
  0x08A8, 0x29,
  0x08A9, 0x01,
  0x08AA, 0x20,
  0x08AB, 0x00,
  0x08AC, 0x49,
  0x08AD, 0x05,
  0x08AE, 0xD0,
  0x08AF, 0x06,
  0x08B0, 0x29,
  0x08B1, 0x01,
  0x08B2, 0x78,
  0x08B3, 0x09,
  0x08B4, 0x49,
  0x08B5, 0x04,
  0x08B6, 0xD1,
  0x08B7, 0x01,
  0x08B8, 0x47,
  0x08B9, 0x70,
  0x08BA, 0x70,
  0x08BB, 0x08,
  0x08BC, 0xE7,
  0x08BD, 0xFB,
  0x08BE, 0x49,
  0x08BF, 0x03,
  0x08C0, 0x10,
  0x08C1, 0x00,
  0x08C2, 0x0F,
  0x08C3, 0x8C,
  0x08C4, 0x10,
  0x08C5, 0x00,
  0x08C6, 0x0F,
  0x08C7, 0x11,
  0x08C8, 0x10,
  0x08C9, 0x00,
  0x08CA, 0x0F,
  0x08CB, 0x12,
  0x08CC, 0x10,
  0x08CD, 0x00,
  0x08CE, 0x0F,
  0x08CF, 0x10,
  0x08D0, 0xB2,
  0x08D1, 0xC1,
  0x08D2, 0xB5,
  0x08D3, 0x10,
  0x08D4, 0xF7,
  0x08D5, 0xFF,
  0x08D6, 0x20,
  0x08D7, 0x01,
  0x08D8, 0xBD,
  0x08D9, 0x10,
  0x08DA, 0xFC,
  0x08DB, 0x45,
  0x08DC, 0xB2,
  0x08DD, 0xC1,
  0x08DE, 0xB5,
  0x08DF, 0x10,
  0x08E0, 0xD1,
  0x08E1, 0x01,
  0x08E2, 0x29,
  0x08E3, 0x03,
  0x08E4, 0xF9,
  0x08E5, 0x20,
  0x08E6, 0xF0,
  0x08E7, 0x00,
  0x08E8, 0x00,
  0x08E9, 0x00,
  0x08EA, 0xBD,
  0x08EB, 0x10,
  0x08EC, 0x48,
  0x08ED, 0x15,
  0x08EE, 0xB5,
  0x08EF, 0x7C,
  0x08F0, 0x89,
  0x08F1, 0x00,
  0x08F2, 0x4C,
  0x08F3, 0x13,
  0x08F4, 0xFD,
  0x08F5, 0xEC,
  0x08F6, 0xF7,
  0x08F7, 0xFF,
  0x08F8, 0x14,
  0x08F9, 0x05,
  0x08FA, 0x04,
  0x08FB, 0x00,
  0x08FC, 0x47,
  0x08FD, 0xA0,
  0x08FE, 0x20,
  0x08FF, 0x2C,
  0x0900, 0x14,
  0x0901, 0x00,
  0x0902, 0x04,
  0x0903, 0x00,
  0x0904, 0x49,
  0x0905, 0x10,
  0x0906, 0x43,
  0x0907, 0x68,
  0x0908, 0x42,
  0x0909, 0x88,
  0x090A, 0x13,
  0x090B, 0x40,
  0x090C, 0x43,
  0x090D, 0xC9,
  0x090E, 0xDC,
  0x090F, 0x02,
  0x0910, 0xDA,
  0x0911, 0x00,
  0x0912, 0x42,
  0x0913, 0x88,
  0x0914, 0xB2,
  0x0915, 0x84,
  0x0916, 0x46,
  0x0917, 0x08,
  0x0918, 0x4B,
  0x0919, 0x0D,
  0x091A, 0x48,
  0x091B, 0x0C,
  0x091C, 0x68,
  0x091D, 0x1B,
  0x091E, 0x90,
  0x091F, 0x01,
  0x0920, 0x46,
  0x0921, 0x69,
  0x0922, 0x22,
  0x0923, 0x01,
  0x0924, 0x47,
  0x0925, 0x98,
  0x0926, 0xA8,
  0x0927, 0x01,
  0x0928, 0x4B,
  0x0929, 0x0A,
  0x092A, 0x98,
  0x092B, 0x00,
  0x092C, 0x04,
  0x092D, 0x00,
  0x092E, 0x0C,
  0x092F, 0x00,
  0x0930, 0x90,
  0x0931, 0x00,
  0x0932, 0x43,
  0x0933, 0x20,
  0x0934, 0x22,
  0x0935, 0x01,
  0x0936, 0x68,
  0x0937, 0x1B,
  0x0938, 0xA8,
  0x0939, 0x01,
  0x093A, 0x46,
  0x093B, 0x69,
  0x093C, 0xBD,
  0x093D, 0x7C,
  0x093E, 0x47,
  0x093F, 0x98,
  0x0940, 0x00,
  0x0941, 0x00,
  0x0942, 0x07,
  0x0943, 0x7B,
  0x0944, 0x10,
  0x0945, 0x00,
  0x0946, 0x1A,
  0x0947, 0x80,
  0x0948, 0x00,
  0x0949, 0x00,
  0x094A, 0x7F,
  0x094B, 0xFF,
  0x094C, 0x00,
  0x094D, 0x00,
  0x094E, 0x61,
  0x094F, 0x74,
  0x0950, 0x10,
  0x0951, 0x00,
  0x0952, 0x0E,
  0x0953, 0xE4,
  0x0954, 0x10,
  0x0955, 0x00,
  0x0956, 0x0E,
  0x0957, 0xE0,
  0x0958, 0x48,
  0x0959, 0x0F,
  0x095A, 0xB5,
  0x095B, 0x70,
  0x095C, 0x89,
  0x095D, 0x40,
  0x095E, 0x4C,
  0x095F, 0x0D,
  0x0960, 0xFD,
  0x0961, 0xB6,
  0x0962, 0xF7,
  0x0963, 0xFF,
  0x0964, 0x15,
  0x0965, 0x85,
  0x0966, 0x05,
  0x0967, 0x80,
  0x0968, 0x47,
  0x0969, 0xA0,
  0x096A, 0x20,
  0x096B, 0x2A,
  0x096C, 0x38,
  0x096D, 0xFF,
  0x096E, 0x38,
  0x096F, 0xFF,
  0x0970, 0x05,
  0x0971, 0x80,
  0x0972, 0x38,
  0x0973, 0x02,
  0x0974, 0x49,
  0x0975, 0x09,
  0x0976, 0x15,
  0x0977, 0x80,
  0x0978, 0x42,
  0x0979, 0x88,
  0x097A, 0x19,
  0x097B, 0x40,
  0x097C, 0x43,
  0x097D, 0xC9,
  0x097E, 0xDC,
  0x097F, 0x02,
  0x0980, 0xDA,
  0x0981, 0x00,
  0x0982, 0x42,
  0x0983, 0x88,
  0x0984, 0x30,
  0x0985, 0xFF,
  0x0986, 0x46,
  0x0987, 0x08,
  0x0988, 0x49,
  0x0989, 0x05,
  0x098A, 0x30,
  0x098B, 0xFF,
  0x098C, 0x60,
  0x098D, 0x88,
  0x098E, 0x30,
  0x098F, 0x02,
  0x0990, 0x00,
  0x0991, 0x00,
  0x0992, 0xBD,
  0x0993, 0x70,
  0x0994, 0x00,
  0x0995, 0x00,
  0x0996, 0x07,
  0x0997, 0x7B,
  0x0998, 0x10,
  0x0999, 0x00,
  0x099A, 0x1A,
  0x099B, 0x80,
  0x099C, 0x00,
  0x099D, 0x00,
  0x099E, 0x01,
  0x099F, 0xFF,
  0x09A0, 0xC8,
  0x09A1, 0x00,
  0x09A2, 0x00,
  0x09A3, 0x40,
  0x09A4, 0x49,
  0x09A5, 0x0F,
  0x09A6, 0xB5,
  0x09A7, 0x00,
  0x09A8, 0x20,
  0x09A9, 0x28,
  0x09AA, 0xB0,
  0x09AB, 0x85,
  0x09AC, 0x21,
  0x09AD, 0x01,
  0x09AE, 0x47,
  0x09AF, 0x88,
  0x09B0, 0x68,
  0x09B1, 0x09,
  0x09B2, 0x07,
  0x09B3, 0x89,
  0x09B4, 0x21,
  0x09B5, 0x00,
  0x09B6, 0x07,
  0x09B7, 0x8A,
  0x09B8, 0xD0,
  0x09B9, 0x0D,
  0x09BA, 0x2A,
  0x09BB, 0x00,
  0x09BC, 0x92,
  0x09BD, 0x03,
  0x09BE, 0x4A,
  0x09BF, 0x0A,
  0x09C0, 0x4B,
  0x09C1, 0x0B,
  0x09C2, 0x4A,
  0x09C3, 0x0A,
  0x09C4, 0x91,
  0x09C5, 0x02,
  0x09C6, 0x92,
  0x09C7, 0x04,
  0x09C8, 0x68,
  0x09C9, 0x1B,
  0x09CA, 0x90,
  0x09CB, 0x01,
  0x09CC, 0xA9,
  0x09CD, 0x01,
  0x09CE, 0x22,
  0x09CF, 0x02,
  0x09D0, 0x47,
  0x09D1, 0x98,
  0x09D2, 0xA8,
  0x09D3, 0x03,
  0x09D4, 0xBD,
  0x09D5, 0x00,
  0x09D6, 0xB0,
  0x09D7, 0x05,
  0x09D8, 0x62,
  0x09D9, 0x50,
  0x09DA, 0x4A,
  0x09DB, 0x06,
  0x09DC, 0x62,
  0x09DD, 0xC1,
  0x09DE, 0x48,
  0x09DF, 0x06,
  0x09E0, 0x00,
  0x09E1, 0x00,
  0x09E2, 0xE7,
  0x09E3, 0xF8,
  0x09E4, 0x00,
  0x09E5, 0x00,
  0x09E6, 0x07,
  0x09E7, 0x7B,
  0x09E8, 0x00,
  0x09E9, 0x00,
  0x09EA, 0x61,
  0x09EB, 0x64,
  0x09EC, 0x00,
  0x09ED, 0x00,
  0x09EE, 0x43,
  0x09EF, 0x6C,
  0x09F0, 0x10,
  0x09F1, 0x00,
  0x09F2, 0x0E,
  0x09F3, 0xE0,
  0x09F4, 0x50,
  0x09F5, 0x00,
  0x09F6, 0x61,
  0x09F7, 0x40,
  0x09F8, 0x50,
  0x09F9, 0x00,
  0x09FA, 0x43,
  0x09FB, 0x40,
  0x09FC, 0x20,
  0x09FD, 0x00,
  0x09FE, 0xB5,
  0x09FF, 0x00,
  0x0A00, 0x07,
  0x0A01, 0x89,
  0x0A02, 0x21,
  0x0A03, 0x01,
  0x0A04, 0xB0,
  0x0A05, 0x85,
  0x0A06, 0x68,
  0x0A07, 0x09,
  0x0A08, 0xD0,
  0x0A09, 0x0D,
  0x0A0A, 0x07,
  0x0A0B, 0x89,
  0x0A0C, 0x91,
  0x0A0D, 0x03,
  0x0A0E, 0x49,
  0x0A0F, 0x09,
  0x0A10, 0x4B,
  0x0A11, 0x0A,
  0x0A12, 0x49,
  0x0A13, 0x09,
  0x0A14, 0x91,
  0x0A15, 0x04,
  0x0A16, 0x90,
  0x0A17, 0x01,
  0x0A18, 0x68,
  0x0A19, 0x1B,
  0x0A1A, 0x90,
  0x0A1B, 0x02,
  0x0A1C, 0xA9,
  0x0A1D, 0x01,
  0x0A1E, 0x22,
  0x0A1F, 0x02,
  0x0A20, 0x47,
  0x0A21, 0x98,
  0x0A22, 0xA8,
  0x0A23, 0x03,
  0x0A24, 0xBD,
  0x0A25, 0x00,
  0x0A26, 0xB0,
  0x0A27, 0x05,
  0x0A28, 0x62,
  0x0A29, 0x48,
  0x0A2A, 0x49,
  0x0A2B, 0x05,
  0x0A2C, 0x62,
  0x0A2D, 0xC8,
  0x0A2E, 0x49,
  0x0A2F, 0x05,
  0x0A30, 0x00,
  0x0A31, 0x00,
  0x0A32, 0xE7,
  0x0A33, 0xF8,
  0x0A34, 0x00,
  0x0A35, 0x00,
  0x0A36, 0x61,
  0x0A37, 0x64,
  0x0A38, 0x00,
  0x0A39, 0x00,
  0x0A3A, 0x43,
  0x0A3B, 0x6C,
  0x0A3C, 0x10,
  0x0A3D, 0x00,
  0x0A3E, 0x0E,
  0x0A3F, 0xE0,
  0x0A40, 0x50,
  0x0A41, 0x00,
  0x0A42, 0x61,
  0x0A43, 0x40,
  0x0A44, 0x50,
  0x0A45, 0x00,
  0x0A46, 0x43,
  0x0A47, 0x40,
  0x0A48, 0x8B,
  0x0A49, 0x00,
  0x0A4A, 0x48,
  0x0A4B, 0x04,
  0x0A4C, 0x49,
  0x0A4D, 0x04,
  0x0A4E, 0x4A,
  0x0A4F, 0x05,
  0x0A50, 0x76,
  0x0A51, 0x13,
  0x0A52, 0x0A,
  0x0A53, 0x03,
  0x0A54, 0x76,
  0x0A55, 0x50,
  0x0A56, 0x75,
  0x0A57, 0x0B,
  0x0A58, 0x47,
  0x0A59, 0x70,
  0x0A5A, 0x75,
  0x0A5B, 0x48,
  0x0A5C, 0x10,
  0x0A5D, 0x00,
  0x0A5E, 0x1A,
  0x0A5F, 0x80,
  0x0A60, 0x10,
  0x0A61, 0x00,
  0x0A62, 0x1D,
  0x0A63, 0xC0,
  0x0A64, 0x10,
  0x0A65, 0x00,
  0x0A66, 0x1C,
  0x0A67, 0xD0,
  0x0A68, 0x8B,
  0x0A69, 0x40,
  0x0A6A, 0x48,
  0x0A6B, 0x04,
  0x0A6C, 0x49,
  0x0A6D, 0x04,
  0x0A6E, 0x4A,
  0x0A6F, 0x05,
  0x0A70, 0x76,
  0x0A71, 0x93,
  0x0A72, 0x0A,
  0x0A73, 0x03,
  0x0A74, 0x76,
  0x0A75, 0xD0,
  0x0A76, 0x76,
  0x0A77, 0x8B,
  0x0A78, 0x47,
  0x0A79, 0x70,
  0x0A7A, 0x76,
  0x0A7B, 0xC8,
  0x0A7C, 0x10,
  0x0A7D, 0x00,
  0x0A7E, 0x1A,
  0x0A7F, 0x80,
  0x0A80, 0x10,
  0x0A81, 0x00,
  0x0A82, 0x1D,
  0x0A83, 0xC0,
  0x0A84, 0x10,
  0x0A85, 0x00,
  0x0A86, 0x1C,
  0x0A87, 0xD0,
  0x0A88, 0x07,
  0x0A89, 0xC0,
  0x0A8A, 0xB5,
  0x0A8B, 0x1C,
  0x0A8C, 0xD0,
  0x0A8D, 0x01,
  0x0A8E, 0x0F,
  0x0A8F, 0xC0,
  0x0A90, 0xE0,
  0x0A91, 0x00,
  0x0A92, 0x48,
  0x0A93, 0x08,
  0x0A94, 0x4B,
  0x0A95, 0x09,
  0x0A96, 0x48,
  0x0A97, 0x08,
  0x0A98, 0x68,
  0x0A99, 0x1B,
  0x0A9A, 0x90,
  0x0A9B, 0x00,
  0x0A9C, 0xA9,
  0x0A9D, 0x01,
  0x0A9E, 0x22,
  0x0A9F, 0x01,
  0x0AA0, 0x47,
  0x0AA1, 0x98,
  0x0AA2, 0x46,
  0x0AA3, 0x68,
  0x0AA4, 0x49,
  0x0AA5, 0x06,
  0x0AA6, 0x98,
  0x0AA7, 0x01,
  0x0AA8, 0x77,
  0x0AA9, 0x0A,
  0x0AAA, 0x0E,
  0x0AAB, 0x02,
  0x0AAC, 0x77,
  0x0AAD, 0x48,
  0x0AAE, 0x0C,
  0x0AAF, 0x00,
  0x0AB0, 0x00,
  0x0AB1, 0x00,
  0x0AB2, 0xBD,
  0x0AB3, 0x1C,
  0x0AB4, 0x00,
  0x0AB5, 0x00,
  0x0AB6, 0x50,
  0x0AB7, 0x54,
  0x0AB8, 0x00,
  0x0AB9, 0x00,
  0x0ABA, 0x40,
  0x0ABB, 0x54,
  0x0ABC, 0x10,
  0x0ABD, 0x00,
  0x0ABE, 0x0E,
  0x0ABF, 0xE4,
  0x0AC0, 0x10,
  0x0AC1, 0x00,
  0x0AC2, 0x1A,
  0x0AC3, 0x80,
  0x0AC4, 0xB2,
  0x0AC5, 0xC1,
  0x0AC6, 0xB5,
  0x0AC7, 0x10,
  0x0AC8, 0xD0,
  0x0AC9, 0x01,
  0x0ACA, 0x29,
  0x0ACB, 0x01,
  0x0ACC, 0xD1,
  0x0ACD, 0x01,
  0x0ACE, 0x29,
  0x0ACF, 0x02,
  0x0AD0, 0xF8,
  0x0AD1, 0x2A,
  0x0AD2, 0xF0,
  0x0AD3, 0x00,
  0x0AD4, 0xB5,
  0x0AD5, 0x10,
  0x0AD6, 0xBD,
  0x0AD7, 0x10,
  0x0AD8, 0x20,
  0x0AD9, 0x00,
  0x0ADA, 0xB2,
  0x0ADB, 0xC1,
  0x0ADC, 0xFB,
  0x0ADD, 0x42,
  0x0ADE, 0xF7,
  0x0ADF, 0xFF,
  0x0AE0, 0x00,
  0x0AE1, 0x00,
  0x0AE2, 0xBD,
  0x0AE3, 0x10,
  0x0AE4, 0x07,
  0x0AE5, 0x82,
  0x0AE6, 0xB5,
  0x0AE7, 0x10,
  0x0AE8, 0x0F,
  0x0AE9, 0x92,
  0x0AEA, 0x49,
  0x0AEB, 0x0D,
  0x0AEC, 0x07,
  0x0AED, 0x82,
  0x0AEE, 0x73,
  0x0AEF, 0x0A,
  0x0AF0, 0x4B,
  0x0AF1, 0x0C,
  0x0AF2, 0x0F,
  0x0AF3, 0xD2,
  0x0AF4, 0x58,
  0x0AF5, 0x9C,
  0x0AF6, 0x00,
  0x0AF7, 0xD2,
  0x0AF8, 0x68,
  0x0AF9, 0x53,
  0x0AFA, 0x18,
  0x0AFB, 0xD2,
  0x0AFC, 0xD3,
  0x0AFD, 0x02,
  0x0AFE, 0x42,
  0x0AFF, 0x9C,
  0x0B00, 0x23,
  0x0B01, 0x01,
  0x0B02, 0x24,
  0x0B03, 0x00,
  0x0B04, 0x24,
  0x0B05, 0x01,
  0x0B06, 0xE0,
  0x0B07, 0x01,
  0x0B08, 0x07,
  0x0B09, 0xC0,
  0x0B0A, 0x23,
  0x0B0B, 0x00,
  0x0B0C, 0x00,
  0x0B0D, 0xA0,
  0x0B0E, 0xD0,
  0x0B0F, 0x01,
  0x0B10, 0x00,
  0x0B11, 0x98,
  0x0B12, 0xE0,
  0x0B13, 0x00,
  0x0B14, 0x0A,
  0x0B15, 0x02,
  0x0B16, 0x58,
  0x0B17, 0x10,
  0x0B18, 0x73,
  0x0B19, 0xC8,
  0x0B1A, 0x73,
  0x0B1B, 0x8A,
  0x0B1C, 0x00,
  0x0B1D, 0x00,
  0x0B1E, 0xBD,
  0x0B1F, 0x10,
  0x0B20, 0x10,
  0x0B21, 0x00,
  0x0B22, 0x1A,
  0x0B23, 0x00,
  0x0B24, 0x10,
  0x0B25, 0x00,
  0x0B26, 0x0F,
  0x0B27, 0x98,
  0x0B28, 0xB2,
  0x0B29, 0xC0,
  0x0B2A, 0xB5,
  0x0B2B, 0xF0,
  0x0B2C, 0x28,
  0x0B2D, 0x01,
  0x0B2E, 0xB0,
  0x0B2F, 0x85,
  0x0B30, 0x28,
  0x0B31, 0x02,
  0x0B32, 0xD0,
  0x0B33, 0x42,
  0x0B34, 0x28,
  0x0B35, 0x03,
  0x0B36, 0xD0,
  0x0B37, 0x42,
  0x0B38, 0x24,
  0x0B39, 0x02,
  0x0B3A, 0xD1,
  0x0B3B, 0x7E,
  0x0B3C, 0x4A,
  0x0B3D, 0x7F,
  0x0B3E, 0x4E,
  0x0B3F, 0x80,
  0x0B40, 0x36,
  0x0B41, 0xFF,
  0x0B42, 0x46,
  0x0B43, 0x35,
  0x0B44, 0x35,
  0x0B45, 0x80,
  0x0B46, 0x20,
  0x0B47, 0x00,
  0x0B48, 0x70,
  0x0B49, 0xD4,
  0x0B4A, 0x36,
  0x0B4B, 0x01,
  0x0B4C, 0xD0,
  0x0B4D, 0x37,
  0x0B4E, 0x2C,
  0x0B4F, 0x02,
  0x0B50, 0x21,
  0x0B51, 0x01,
  0x0B52, 0x74,
  0x0B53, 0xE8,
  0x0B54, 0x20,
  0x0B55, 0x00,
  0x0B56, 0x70,
  0x0B57, 0x51,
  0x0B58, 0x60,
  0x0B59, 0xD0,
  0x0B5A, 0x60,
  0x0B5B, 0x90,
  0x0B5C, 0x60,
  0x0B5D, 0x50,
  0x0B5E, 0x20,
  0x0B5F, 0x04,
  0x0B60, 0xD0,
  0x0B61, 0x2F,
  0x0B62, 0x2C,
  0x0B63, 0x02,
  0x0B64, 0x7C,
  0x0B65, 0xA9,
  0x0B66, 0x7C,
  0x0B67, 0xA8,
  0x0B68, 0x07,
  0x0B69, 0x09,
  0x0B6A, 0x09,
  0x0B6B, 0x00,
  0x0B6C, 0x23,
  0x0B6D, 0x0A,
  0x0B6E, 0x0F,
  0x0B6F, 0x09,
  0x0B70, 0x23,
  0x0B71, 0x40,
  0x0B72, 0x43,
  0x0B73, 0x59,
  0x0B74, 0x4B,
  0x0B75, 0x71,
  0x0B76, 0x93,
  0x0B77, 0x04,
  0x0B78, 0x33,
  0x0B79, 0x14,
  0x0B7A, 0x00,
  0x0B7B, 0x80,
  0x0B7C, 0x48,
  0x0B7D, 0x6F,
  0x0B7E, 0x18,
  0x0B7F, 0xC3,
  0x0B80, 0x61,
  0x0B81, 0x03,
  0x0B82, 0x31,
  0x0B83, 0x6E,
  0x0B84, 0x48,
  0x0B85, 0x6F,
  0x0B86, 0x27,
  0x0B87, 0x05,
  0x0B88, 0xFD,
  0x0B89, 0xA0,
  0x0B8A, 0xF7,
  0x0B8B, 0xFF,
  0x0B8C, 0x90,
  0x0B8D, 0x03,
  0x0B8E, 0x01,
  0x0B8F, 0x00,
  0x0B90, 0xF0,
  0x0B91, 0x00,
  0x0B92, 0x46,
  0x0B93, 0x20,
  0x0B94, 0x49,
  0x0B95, 0x69,
  0x0B96, 0xF9,
  0x0B97, 0x41,
  0x0B98, 0x70,
  0x0B99, 0x08,
  0x0B9A, 0x20,
  0x0B9B, 0x01,
  0x0B9C, 0x20,
  0x0B9D, 0x04,
  0x0B9E, 0x49,
  0x0B9F, 0x6A,
  0x0BA0, 0x48,
  0x0BA1, 0x6A,
  0x0BA2, 0x60,
  0x0BA3, 0x08,
  0x0BA4, 0x22,
  0x0BA5, 0x0C,
  0x0BA6, 0x68,
  0x0BA7, 0x81,
  0x0BA8, 0x60,
  0x0BA9, 0x81,
  0x0BAA, 0x43,
  0x0BAB, 0x91,
  0x0BAC, 0x98,
  0x0BAD, 0x03,
  0x0BAE, 0x49,
  0x0BAF, 0x68,
  0x0BB0, 0x1E,
  0x0BB1, 0x40,
  0x0BB2, 0x68,
  0x0BB3, 0x09,
  0x0BB4, 0xE0,
  0x0BB5, 0x0C,
  0x0BB6, 0x47,
  0x0BB7, 0x88,
  0x0BB8, 0xE7,
  0x0BB9, 0xBF,
  0x0BBA, 0x24,
  0x0BBB, 0x00,
  0x0BBC, 0xE7,
  0x0BBD, 0xBD,
  0x0BBE, 0x24,
  0x0BBF, 0x01,
  0x0BC0, 0xE7,
  0x0BC1, 0xC6,
  0x0BC2, 0x73,
  0x0BC3, 0x70,
  0x0BC4, 0x7B,
  0x0BC5, 0x31,
  0x0BC6, 0x7B,
  0x0BC7, 0x30,
  0x0BC8, 0xE7,
  0x0BC9, 0xCE,
  0x0BCA, 0x09,
  0x0BCB, 0x00,
  0x0BCC, 0xF7,
  0x0BCD, 0xFF,
  0x0BCE, 0x46,
  0x0BCF, 0x20,
  0x0BD0, 0x48,
  0x0BD1, 0x5A,
  0x0BD2, 0xFD,
  0x0BD3, 0xB1,
  0x0BD4, 0x28,
  0x0BD5, 0x01,
  0x0BD6, 0x78,
  0x0BD7, 0x00,
  0x0BD8, 0x48,
  0x0BD9, 0x5E,
  0x0BDA, 0xD0,
  0x0BDB, 0xF8,
  0x0BDC, 0x47,
  0x0BDD, 0x80,
  0x0BDE, 0x68,
  0x0BDF, 0x00,
  0x0BE0, 0x68,
  0x0BE1, 0x90,
  0x0BE2, 0x4A,
  0x0BE3, 0x5A,
  0x0BE4, 0x43,
  0x0BE5, 0x88,
  0x0BE6, 0x21,
  0x0BE7, 0x0C,
  0x0BE8, 0x60,
  0x0BE9, 0x90,
  0x0BEA, 0x1D,
  0x0BEB, 0x00,
  0x0BEC, 0x20,
  0x0BED, 0x04,
  0x0BEE, 0x49,
  0x0BEF, 0x56,
  0x0BF0, 0x60,
  0x0BF1, 0x08,
  0x0BF2, 0x31,
  0x0BF3, 0x80,
  0x0BF4, 0xF0,
  0x0BF5, 0x00,
  0x0BF6, 0x46,
  0x0BF7, 0x20,
  0x0BF8, 0xA9,
  0x0BF9, 0x02,
  0x0BFA, 0xF9,
  0x0BFB, 0x41,
  0x0BFC, 0xF7,
  0x0BFD, 0xFF,
  0x0BFE, 0x46,
  0x0BFF, 0x20,
  0x0C00, 0x28,
  0x0C01, 0x01,
  0x0C02, 0xFD,
  0x0C03, 0xD5,
  0x0C04, 0x28,
  0x0C05, 0x02,
  0x0C06, 0xD0,
  0x0C07, 0x1F,
  0x0C08, 0x2F,
  0x0C09, 0x00,
  0x0C0A, 0xD0,
  0x0C0B, 0x7A,
  0x0C0C, 0x2C,
  0x0C0D, 0x02,
  0x0C0E, 0xD0,
  0x0C0F, 0x7F,
  0x0C10, 0x48,
  0x0C11, 0x51,
  0x0C12, 0xD0,
  0x0C13, 0x77,
  0x0C14, 0x90,
  0x0C15, 0x00,
  0x0C16, 0x4B,
  0x0C17, 0x51,
  0x0C18, 0x22,
  0x0C19, 0x01,
  0x0C1A, 0x68,
  0x0C1B, 0x1B,
  0x0C1C, 0x46,
  0x0C1D, 0x68,
  0x0C1E, 0xA9,
  0x0C1F, 0x01,
  0x0C20, 0x2C,
  0x0C21, 0x02,
  0x0C22, 0x47,
  0x0C23, 0x98,
  0x0C24, 0x98,
  0x0C25, 0x01,
  0x0C26, 0xD0,
  0x0C27, 0x7B,
  0x0C28, 0xD0,
  0x0C29, 0x7F,
  0x0C2A, 0x2C,
  0x0C2B, 0x01,
  0x0C2C, 0x0E,
  0x0C2D, 0x09,
  0x0C2E, 0x04,
  0x0C2F, 0x01,
  0x0C30, 0x74,
  0x0C31, 0x68,
  0x0C32, 0x74,
  0x0C33, 0x29,
  0x0C34, 0xE0,
  0x0C35, 0x00,
  0x0C36, 0x20,
  0x0C37, 0x01,
  0x0C38, 0x74,
  0x0C39, 0xE8,
  0x0C3A, 0xE0,
  0x0C3B, 0x03,
  0x0C3C, 0x20,
  0x0C3D, 0x00,
  0x0C3E, 0x49,
  0x0C3F, 0x3F,
  0x0C40, 0xB0,
  0x0C41, 0x05,
  0x0C42, 0x70,
  0x0C43, 0x48,
  0x0C44, 0x2F,
  0x0C45, 0x00,
  0x0C46, 0xBD,
  0x0C47, 0xF0,
  0x0C48, 0x1E,
  0x0C49, 0x7F,
  0x0C4A, 0xD0,
  0x0C4B, 0x61,
  0x0C4C, 0xD0,
  0x0C4D, 0x23,
  0x0C4E, 0x2C,
  0x0C4F, 0x02,
  0x0C50, 0xD0,
  0x0C51, 0x23,
  0x0C52, 0x2C,
  0x0C53, 0x01,
  0x0C54, 0x90,
  0x0C55, 0x00,
  0x0C56, 0x48,
  0x0C57, 0x42,
  0x0C58, 0x4B,
  0x0C59, 0x42,
  0x0C5A, 0x20,
  0x0C5B, 0x00,
  0x0C5C, 0x68,
  0x0C5D, 0x1B,
  0x0C5E, 0x90,
  0x0C5F, 0x01,
  0x0C60, 0xA9,
  0x0C61, 0x01,
  0x0C62, 0x22,
  0x0C63, 0x01,
  0x0C64, 0x47,
  0x0C65, 0x98,
  0x0C66, 0x46,
  0x0C67, 0x68,
  0x0C68, 0xD0,
  0x0C69, 0x19,
  0x0C6A, 0x2C,
  0x0C6B, 0x02,
  0x0C6C, 0x4B,
  0x0C6D, 0x3B,
  0x0C6E, 0x48,
  0x0C6F, 0x3A,
  0x0C70, 0x68,
  0x0C71, 0x1B,
  0x0C72, 0x90,
  0x0C73, 0x00,
  0x0C74, 0xA9,
  0x0C75, 0x01,
  0x0C76, 0x22,
  0x0C77, 0x01,
  0x0C78, 0x47,
  0x0C79, 0x98,
  0x0C7A, 0x46,
  0x0C7B, 0x68,
  0x0C7C, 0x2C,
  0x0C7D, 0x01,
  0x0C7E, 0x98,
  0x0C7F, 0x01,
  0x0C80, 0xB2,
  0x0C81, 0x81,
  0x0C82, 0xD0,
  0x0C83, 0x11,
  0x0C84, 0x43,
  0x0C85, 0x51,
  0x0C86, 0x9A,
  0x0C87, 0x02,
  0x0C88, 0x2C,
  0x0C89, 0x02,
  0x0C8A, 0x09,
  0x0C8B, 0xC9,
  0x0C8C, 0x4A,
  0x0C8D, 0x32,
  0x0C8E, 0xD0,
  0x0C8F, 0x0D,
  0x0C90, 0x2C,
  0x0C91, 0x01,
  0x0C92, 0x92,
  0x0C93, 0x00,
  0x0C94, 0xE0,
  0x0C95, 0x0B,
  0x0C96, 0xD0,
  0x0C97, 0x38,
  0x0C98, 0xE7,
  0x0C99, 0xDC,
  0x0C9A, 0x48,
  0x0C9B, 0x33,
  0x0C9C, 0xE7,
  0x0C9D, 0xDA,
  0x0C9E, 0x48,
  0x0C9F, 0x33,
  0x0CA0, 0x30,
  0x0CA1, 0x94,
  0x0CA2, 0x48,
  0x0CA3, 0x2D,
  0x0CA4, 0x0C,
  0x0CA5, 0x01,
  0x0CA6, 0xE7,
  0x0CA7, 0xE3,
  0x0CA8, 0x4A,
  0x0CA9, 0x2B,
  0x0CAA, 0xE7,
  0x0CAB, 0xEC,
  0x0CAC, 0x92,
  0x0CAD, 0x00,
  0x0CAE, 0x32,
  0x0CAF, 0x94,
  0x0CB0, 0xB2,
  0x0CB1, 0x89,
  0x0CB2, 0x0C,
  0x0CB3, 0x00,
  0x0CB4, 0x43,
  0x0CB5, 0x01,
  0x0CB6, 0x04,
  0x0CB7, 0x00,
  0x0CB8, 0x91,
  0x0CB9, 0x01,
  0x0CBA, 0x4B,
  0x0CBB, 0x2A,
  0x0CBC, 0x22,
  0x0CBD, 0x01,
  0x0CBE, 0x68,
  0x0CBF, 0x1B,
  0x0CC0, 0x46,
  0x0CC1, 0x68,
  0x0CC2, 0xA9,
  0x0CC3, 0x01,
  0x0CC4, 0x49,
  0x0CC5, 0x1D,
  0x0CC6, 0x47,
  0x0CC7, 0x98,
  0x0CC8, 0x60,
  0x0CC9, 0xC8,
  0x0CCA, 0x20,
  0x0CCB, 0x00,
  0x0CCC, 0x20,
  0x0CCD, 0x01,
  0x0CCE, 0x60,
  0x0CCF, 0x88,
  0x0CD0, 0x49,
  0x0CD1, 0x1D,
  0x0CD2, 0x70,
  0x0CD3, 0x88,
  0x0CD4, 0x60,
  0x0CD5, 0x08,
  0x0CD6, 0x20,
  0x0CD7, 0x04,
  0x0CD8, 0x98,
  0x0CD9, 0x04,
  0x0CDA, 0x21,
  0x0CDB, 0xD3,
  0x0CDC, 0x43,
  0x0CDD, 0x48,
  0x0CDE, 0x00,
  0x0CDF, 0xC9,
  0x0CE0, 0x1E,
  0x0CE1, 0x40,
  0x0CE2, 0x49,
  0x0CE3, 0x1B,
  0x0CE4, 0x47,
  0x0CE5, 0x88,
  0x0CE6, 0x68,
  0x0CE7, 0x09,
  0x0CE8, 0x78,
  0x0CE9, 0x81,
  0x0CEA, 0x48,
  0x0CEB, 0x14,
  0x0CEC, 0xD0,
  0x0CED, 0xFC,
  0x0CEE, 0x29,
  0x0CEF, 0x01,
  0x0CF0, 0x68,
  0x0CF1, 0x00,
  0x0CF2, 0x48,
  0x0CF3, 0x18,
  0x0CF4, 0x49,
  0x0CF5, 0x14,
  0x0CF6, 0x47,
  0x0CF7, 0x80,
  0x0CF8, 0x31,
  0x0CF9, 0x80,
  0x0CFA, 0x20,
  0x0CFB, 0x04,
  0x0CFC, 0xE7,
  0x0CFD, 0x47,
  0x0CFE, 0x60,
  0x0CFF, 0x08,
  0x0D00, 0xE0,
  0x0D01, 0x09,
  0x0D02, 0xE0,
  0x0D03, 0x05,
  0x0D04, 0xE0,
  0x0D05, 0x11,
  0x0D06, 0xE0,
  0x0D07, 0x0B,
  0x0D08, 0xB2,
  0x0D09, 0x80,
  0x0D0A, 0x04,
  0x0D0B, 0x09,
  0x0D0C, 0x20,
  0x0D0D, 0xFF,
  0x0D0E, 0xE7,
  0x0D0F, 0xD3,
  0x0D10, 0xD1,
  0x0D11, 0x90,
  0x0D12, 0x2C,
  0x0D13, 0x02,
  0x0D14, 0xE7,
  0x0D15, 0x91,
  0x0D16, 0x73,
  0x0D17, 0x70,
  0x0D18, 0x30,
  0x0D19, 0x94,
  0x0D1A, 0x48,
  0x0D1B, 0x0F,
  0x0D1C, 0x98,
  0x0D1D, 0x01,
  0x0D1E, 0xE7,
  0x0D1F, 0x7A,
  0x0D20, 0x0E,
  0x0D21, 0x09,
  0x0D22, 0x04,
  0x0D23, 0x01,
  0x0D24, 0x72,
  0x0D25, 0xF0,
  0x0D26, 0x72,
  0x0D27, 0xB1,
  0x0D28, 0xE7,
  0x0D29, 0xF3,
  0x0D2A, 0x20,
  0x0D2B, 0x03,
  0x0D2C, 0x74,
  0x0D2D, 0x29,
  0x0D2E, 0x0E,
  0x0D2F, 0x01,
  0x0D30, 0x0E,
  0x0D31, 0x00,
  0x0D32, 0x02,
  0x0D33, 0x00,
  0x0D34, 0x20,
  0x0D35, 0x02,
  0x0D36, 0x74,
  0x0D37, 0x68,
  0x0D38, 0x00,
  0x0D39, 0x00,
  0x0D3A, 0xE7,
  0x0D3B, 0x7D,
  0x0D3C, 0x10,
  0x0D3D, 0x00,
  0x0D3E, 0x0F,
  0x0D3F, 0x10,
  0x0D40, 0x10,
  0x0D41, 0x00,
  0x0D42, 0x19,
  0x0D43, 0x40,
  0x0D44, 0x01,
  0x0D45, 0x9B,
  0x0D46, 0xFC,
  0x0D47, 0xC0,
  0x0D48, 0xE0,
  0x0D49, 0x00,
  0x0D4A, 0xE1,
  0x0D4B, 0x00,
  0x0D4C, 0xC6,
  0x0D4D, 0x00,
  0x0D4E, 0x00,
  0x0D4F, 0x40,
  0x0D50, 0x10,
  0x0D51, 0x00,
  0x0D52, 0x0F,
  0x0D53, 0x04,
  0x0D54, 0x10,
  0x0D55, 0x00,
  0x0D56, 0x0F,
  0x0D57, 0x08,
  0x0D58, 0x00,
  0x0D59, 0x00,
  0x0D5A, 0x60,
  0x0D5B, 0xD0,
  0x0D5C, 0x10,
  0x0D5D, 0x00,
  0x0D5E, 0x0E,
  0x0D5F, 0xE4,
  0x0D60, 0x00,
  0x0D61, 0x00,
  0x0D62, 0x41,
  0x0D63, 0x08,
  0x0D64, 0x10,
  0x0D65, 0x00,
  0x0D66, 0x0E,
  0x0D67, 0xE0,
  0x0D68, 0x00,
  0x0D69, 0x00,
  0x0D6A, 0x43,
  0x0D6B, 0x44,
  0x0D6C, 0x00,
  0x0D6D, 0x00,
  0x0D6E, 0x51,
  0x0D6F, 0x08,
  0x0D70, 0x4A,
  0x0D71, 0x23,
  0x0D72, 0xB5,
  0x0D73, 0x1C,
  0x0D74, 0x88,
  0x0D75, 0x51,
  0x0D76, 0x24,
  0x0D77, 0xFF,
  0x0D78, 0x42,
  0x0D79, 0xA1,
  0x0D7A, 0x34,
  0x0D7B, 0x01,
  0x0D7C, 0x49,
  0x0D7D, 0x21,
  0x0D7E, 0xD1,
  0x0D7F, 0x01,
  0x0D80, 0x68,
  0x0D81, 0x93,
  0x0D82, 0xE0,
  0x0D83, 0x02,
  0x0D84, 0x5A,
  0x0D85, 0x59,
  0x0D86, 0x00,
  0x0D87, 0x49,
  0x0D88, 0x2A,
  0x0D89, 0x02,
  0x0D8A, 0x88,
  0x0D8B, 0x92,
  0x0D8C, 0x2A,
  0x0D8D, 0x03,
  0x0D8E, 0xD0,
  0x0D8F, 0x01,
  0x0D90, 0x42,
  0x0D91, 0x49,
  0x0D92, 0xD1,
  0x0D93, 0x00,
  0x0D94, 0x04,
  0x0D95, 0x09,
  0x0D96, 0x4A,
  0x0D97, 0x1C,
  0x0D98, 0x12,
  0x0D99, 0x89,
  0x0D9A, 0x68,
  0x0D9B, 0x12,
  0x0D9C, 0x91,
  0x0D9D, 0x01,
  0x0D9E, 0x43,
  0x0D9F, 0x51,
  0x0DA0, 0xD0,
  0x0DA1, 0x16,
  0x0DA2, 0x28,
  0x0DA3, 0x02,
  0x0DA4, 0xD0,
  0x0DA5, 0x16,
  0x0DA6, 0x28,
  0x0DA7, 0x01,
  0x0DA8, 0x90,
  0x0DA9, 0x00,
  0x0DAA, 0x48,
  0x0DAB, 0x18,
  0x0DAC, 0xA9,
  0x0DAD, 0x01,
  0x0DAE, 0x22,
  0x0DAF, 0x01,
  0x0DB0, 0xF7,
  0x0DB1, 0xFF,
  0x0DB2, 0x46,
  0x0DB3, 0x68,
  0x0DB4, 0x4B,
  0x0DB5, 0x12,
  0x0DB6, 0xFC,
  0x0DB7, 0x9F,
  0x0DB8, 0x1C,
  0x0DB9, 0x9A,
  0x0DBA, 0x1C,
  0x0DBB, 0x9B,
  0x0DBC, 0x88,
  0x0DBD, 0x18,
  0x0DBE, 0x88,
  0x0DBF, 0x11,
  0x0DC0, 0xD0,
  0x0DC1, 0x0A,
  0x0DC2, 0x29,
  0x0DC3, 0x00,
  0x0DC4, 0xD0,
  0x0DC5, 0x08,
  0x0DC6, 0x29,
  0x0DC7, 0x02,
  0x0DC8, 0xD0,
  0x0DC9, 0x0E,
  0x0DCA, 0x29,
  0x0DCB, 0x01,
  0x0DCC, 0xD0,
  0x0DCD, 0x0C,
  0x0DCE, 0x29,
  0x0DCF, 0x03,
  0x0DD0, 0x48,
  0x0DD1, 0x0F,
  0x0DD2, 0xE0,
  0x0DD3, 0x0E,
  0x0DD4, 0x48,
  0x0DD5, 0x0F,
  0x0DD6, 0xE7,
  0x0DD7, 0xE9,
  0x0DD8, 0x42,
  0x0DD9, 0xA0,
  0x0DDA, 0xE7,
  0x0DDB, 0xE7,
  0x0DDC, 0x1C,
  0x0DDD, 0x49,
  0x0DDE, 0xD1,
  0x0DDF, 0x03,
  0x0DE0, 0x0F,
  0x0DE1, 0x89,
  0x0DE2, 0x07,
  0x0DE3, 0x89,
  0x0DE4, 0x1D,
  0x0DE5, 0x00,
  0x0DE6, 0xE0,
  0x0DE7, 0x03,
  0x0DE8, 0x28,
  0x0DE9, 0x00,
  0x0DEA, 0xE0,
  0x0DEB, 0x02,
  0x0DEC, 0x1F,
  0x0DED, 0x00,
  0x0DEE, 0xD0,
  0x0DEF, 0x03,
  0x0DF0, 0x80,
  0x0DF1, 0x11,
  0x0DF2, 0x80,
  0x0DF3, 0x18,
  0x0DF4, 0x1C,
  0x0DF5, 0x49,
  0x0DF6, 0xBD,
  0x0DF7, 0x1C,
  0x0DF8, 0x0F,
  0x0DF9, 0x89,
  0x0DFA, 0x07,
  0x0DFB, 0x89,
  0x0DFC, 0xE7,
  0x0DFD, 0xF7,
  0x0DFE, 0x20,
  0x0DFF, 0x04,
  0x0E00, 0x10,
  0x0E01, 0x00,
  0x0E02, 0x0F,
  0x0E03, 0x8C,
  0x0E04, 0x00,
  0x0E05, 0x00,
  0x0E06, 0x7F,
  0x0E07, 0xFF,
  0x0E08, 0x10,
  0x0E09, 0x00,
  0x0E0A, 0x0F,
  0x0E0B, 0x14,
  0x0E0C, 0x00,
  0x0E0D, 0x00,
  0x0E0E, 0x41,
  0x0E0F, 0x08,
  0x0E10, 0x00,
  0x0E11, 0x00,
  0x0E12, 0x43,
  0x0E13, 0x44,
  0x0E14, 0x00,
  0x0E15, 0x00,
  0x0E16, 0x51,
  0x0E17, 0x08,
  0x0E18, 0x49,
  0x0E19, 0x13,
  0x0E1A, 0xB5,
  0x0E1B, 0x10,
  0x0E1C, 0x70,
  0x0E1D, 0x0A,
  0x0E1E, 0x22,
  0x0E1F, 0x01,
  0x0E20, 0x28,
  0x0E21, 0x02,
  0x0E22, 0x22,
  0x0E23, 0x00,
  0x0E24, 0x48,
  0x0E25, 0x11,
  0x0E26, 0x80,
  0x0E27, 0x4A,
  0x0E28, 0xD0,
  0x0E29, 0x18,
  0x0E2A, 0x80,
  0x0E2B, 0x8A,
  0x0E2C, 0x7C,
  0x0E2D, 0x80,
  0x0E2E, 0x30,
  0x0E2F, 0x80,
  0x0E30, 0x07,
  0x0E31, 0x01,
  0x0E32, 0x4A,
  0x0E33, 0x0F,
  0x0E34, 0x20,
  0x0E35, 0x0A,
  0x0E36, 0x0F,
  0x0E37, 0x09,
  0x0E38, 0x31,
  0x0E39, 0x6E,
  0x0E3A, 0x43,
  0x0E3B, 0x41,
  0x0E3C, 0x60,
  0x0E3D, 0x10,
  0x0E3E, 0x20,
  0x0E3F, 0x02,
  0x0E40, 0x06,
  0x0E41, 0x40,
  0x0E42, 0x20,
  0x0E43, 0x63,
  0x0E44, 0x23,
  0x0E45, 0x0C,
  0x0E46, 0x6A,
  0x0E47, 0x82,
  0x0E48, 0x62,
  0x0E49, 0x82,
  0x0E4A, 0x43,
  0x0E4B, 0x9A,
  0x0E4C, 0xF7,
  0x0E4D, 0xFF,
  0x0E4E, 0x48,
  0x0E4F, 0x09,
  0x0E50, 0x49,
  0x0E51, 0x09,
  0x0E52, 0xFC,
  0x0E53, 0x3D,
  0x0E54, 0x68,
  0x0E55, 0x09,
  0x0E56, 0x0A,
  0x0E57, 0x00,
  0x0E58, 0x47,
  0x0E59, 0x88,
  0x0E5A, 0x1E,
  0x0E5B, 0x40,
  0x0E5C, 0x30,
  0x0E5D, 0xFF,
  0x0E5E, 0xBD,
  0x0E5F, 0x10,
  0x0E60, 0x7B,
  0x0E61, 0x00,
  0x0E62, 0x30,
  0x0E63, 0x01,
  0x0E64, 0x00,
  0x0E65, 0x00,
  0x0E66, 0xE7,
  0x0E67, 0xE4,
  0x0E68, 0x10,
  0x0E69, 0x00,
  0x0E6A, 0x0F,
  0x0E6B, 0x8C,
  0x0E6C, 0x10,
  0x0E6D, 0x00,
  0x0E6E, 0x19,
  0x0E6F, 0x40,
  0x0E70, 0xE0,
  0x0E71, 0x00,
  0x0E72, 0xE1,
  0x0E73, 0x00,
  0x0E74, 0x01,
  0x0E75, 0x9B,
  0x0E76, 0xFC,
  0x0E77, 0xC0,
  0x0E78, 0x10,
  0x0E79, 0x00,
  0x0E7A, 0x0E,
  0x0E7B, 0xFC,
  0x0E7C, 0x46,
  0x0E7D, 0x04,
  0x0E7E, 0xB5,
  0x0E7F, 0x1C,
  0x0E80, 0x68,
  0x0E81, 0x00,
  0x0E82, 0x48,
  0x0E83, 0x10,
  0x0E84, 0x20,
  0x0E85, 0x63,
  0x0E86, 0x47,
  0x0E87, 0x80,
  0x0E88, 0x6A,
  0x0E89, 0x81,
  0x0E8A, 0x06,
  0x0E8B, 0x40,
  0x0E8C, 0x43,
  0x0E8D, 0x91,
  0x0E8E, 0x22,
  0x0E8F, 0x0C,
  0x0E90, 0x62,
  0x0E91, 0x81,
  0x0E92, 0x1D,
  0x0E93, 0x09,
  0x0E94, 0x20,
  0x0E95, 0x02,
  0x0E96, 0x49,
  0x0E97, 0x0C,
  0x0E98, 0x2C,
  0x0E99, 0x02,
  0x0E9A, 0x60,
  0x0E9B, 0x08,
  0x0E9C, 0x2C,
  0x0E9D, 0x01,
  0x0E9E, 0xD0,
  0x0E9F, 0x0E,
  0x0EA0, 0x48,
  0x0EA1, 0x0A,
  0x0EA2, 0xD0,
  0x0EA3, 0x0E,
  0x0EA4, 0x4B,
  0x0EA5, 0x0A,
  0x0EA6, 0x24,
  0x0EA7, 0x00,
  0x0EA8, 0x90,
  0x0EA9, 0x00,
  0x0EAA, 0x94,
  0x0EAB, 0x01,
  0x0EAC, 0x22,
  0x0EAD, 0x01,
  0x0EAE, 0x68,
  0x0EAF, 0x1B,
  0x0EB0, 0x46,
  0x0EB1, 0x68,
  0x0EB2, 0xA9,
  0x0EB3, 0x01,
  0x0EB4, 0x48,
  0x0EB5, 0x07,
  0x0EB6, 0x47,
  0x0EB7, 0x98,
  0x0EB8, 0xBD,
  0x0EB9, 0x1C,
  0x0EBA, 0x70,
  0x0EBB, 0x04,
  0x0EBC, 0xE7,
  0x0EBD, 0xF1,
  0x0EBE, 0x48,
  0x0EBF, 0x06,
  0x0EC0, 0xE7,
  0x0EC1, 0xEF,
  0x0EC2, 0x48,
  0x0EC3, 0x06,
  0x0EC4, 0x10,
  0x0EC5, 0x00,
  0x0EC6, 0x0F,
  0x0EC7, 0x00,
  0x0EC8, 0xE0,
  0x0EC9, 0x00,
  0x0ECA, 0xE1,
  0x0ECB, 0x80,
  0x0ECC, 0x00,
  0x0ECD, 0x00,
  0x0ECE, 0x41,
  0x0ECF, 0x08,
  0x0ED0, 0x10,
  0x0ED1, 0x00,
  0x0ED2, 0x0E,
  0x0ED3, 0xE0,
  0x0ED4, 0x10,
  0x0ED5, 0x00,
  0x0ED6, 0x0F,
  0x0ED7, 0x8C,
  0x0ED8, 0x00,
  0x0ED9, 0x00,
  0x0EDA, 0x43,
  0x0EDB, 0x44,
  0x0EDC, 0x00,
  0x0EDD, 0x00,
  0x0EDE, 0x51,
  0x0EDF, 0x08,
  0x0EE0, 0x10,
  0x0EE1, 0x00,
  0x0EE2, 0x06,
  0x0EE3, 0x87,
  0x0EE4, 0x10,
  0x0EE5, 0x00,
  0x0EE6, 0x06,
  0x0EE7, 0x87,
  0x0EE8, 0x10,
  0x0EE9, 0x00,
  0x0EEA, 0x06,
  0x0EEB, 0x87,
  0x0EEC, 0x10,
  0x0EED, 0x00,
  0x0EEE, 0x06,
  0x0EEF, 0x87,
  0x0EF0, 0x10,
  0x0EF1, 0x00,
  0x0EF2, 0x06,
  0x0EF3, 0x87,
  0x0EF4, 0x10,
  0x0EF5, 0x00,
  0x0EF6, 0x06,
  0x0EF7, 0x87,
  0x0EF8, 0x10,
  0x0EF9, 0x00,
  0x0EFA, 0x06,
  0x0EFB, 0x87,
  0x0EFC, 0x10,
  0x0EFD, 0x00,
  0x0EFE, 0x06,
  0x0EFF, 0x87,
  0x0F00, 0x10,
  0x0F01, 0x00,
  0x0F02, 0x06,
  0x0F03, 0x87,
  0x0F04, 0x10,
  0x0F05, 0x00,
  0x0F06, 0x06,
  0x0F07, 0x87,
  0x0F08, 0x10,
  0x0F09, 0x00,
  0x0F0A, 0x06,
  0x0F0B, 0x87,
  0x0F0C, 0x00,
  0x0F0D, 0x00,
  0x0F0E, 0x00,
  0x0F0F, 0x00,
  0x0F10, 0x00,
  0x0F11, 0x00,
  0x0F12, 0x00,
  0x0F13, 0x00,
  0x0F14, 0x00,
  0x0F15, 0x00,
  0x0F16, 0x00,
  0x0F17, 0x00,
  0x0F18, 0x00,
  0x0F19, 0x00,
  0x0F1A, 0x00,
  0x0F1B, 0x00,
  0x0F1C, 0x00,
  0x0F1D, 0x00,
  0x0F1E, 0x00,
  0x0F1F, 0x00,
  0x0F20, 0x00,
  0x0F21, 0x00,
  0x0F22, 0x00,
  0x0F23, 0x00,
  0x0F24, 0x01,
  0x0F25, 0xCA,
  0x0F26, 0x02,
  0x0F27, 0x3F,
  0x0F28, 0x02,
  0x0F29, 0x02,
  0x0F2A, 0x02,
  0x0F2B, 0x86,
  0x0F2C, 0x02,
  0x0F2D, 0x40,
  0x0F2E, 0x02,
  0x0F2F, 0xD4,
  0x0F30, 0x02,
  0x0F31, 0x87,
  0x0F32, 0x03,
  0x0F33, 0x2D,
  0x0F34, 0x02,
  0x0F35, 0xD5,
  0x0F36, 0x03,
  0x0F37, 0x90,
  0x0F38, 0x03,
  0x0F39, 0x2E,
  0x0F3A, 0x04,
  0x0F3B, 0x00,
  0x0F3C, 0x03,
  0x0F3D, 0x91,
  0x0F3E, 0x04,
  0x0F3F, 0x7C,
  0x0F40, 0x04,
  0x0F41, 0x01,
  0x0F42, 0x05,
  0x0F43, 0x09,
  0x0F44, 0x04,
  0x0F45, 0x7D,
  0x0F46, 0x05,
  0x0F47, 0xA6,
  0x0F48, 0x05,
  0x0F49, 0x0A,
  0x0F4A, 0x06,
  0x0F4B, 0x56,
  0x0F4C, 0x05,
  0x0F4D, 0xA7,
  0x0F4E, 0x07,
  0x0F4F, 0x1C,
  0x0F50, 0x06,
  0x0F51, 0x57,
  0x0F52, 0x07,
  0x0F53, 0xFB,
  0x0F54, 0x07,
  0x0F55, 0x1D,
  0x0F56, 0x08,
  0x0F57, 0xF4,
  0x0F58, 0x07,
  0x0F59, 0xFC,
  0x0F5A, 0x0A,
  0x0F5B, 0x0C,
  0x0F5C, 0x08,
  0x0F5D, 0xF5,
  0x0F5E, 0x0B,
  0x0F5F, 0x46,
  0x0F60, 0x0A,
  0x0F61, 0x0D,
  0x0F62, 0x0C,
  0x0F63, 0xA6,
  0x0F64, 0x00,
  0x0F65, 0x47,
  0x0F66, 0x00,
  0x0F67, 0x40,
  0x0F68, 0x00,
  0x0F69, 0x5A,
  0x0F6A, 0x00,
  0x0F6B, 0x50,
  0x0F6C, 0x00,
  0x0F6D, 0x72,
  0x0F6E, 0x00,
  0x0F6F, 0x65,
  0x0F70, 0x00,
  0x0F71, 0x8F,
  0x0F72, 0x00,
  0x0F73, 0x80,
  0x0F74, 0x00,
  0x0F75, 0xB4,
  0x0F76, 0x00,
  0x0F77, 0xA1,
  0x0F78, 0x00,
  0x0F79, 0xE3,
  0x0F7A, 0x00,
  0x0F7B, 0xCA,
  0x0F7C, 0x01,
  0x0F7D, 0x1E,
  0x0F7E, 0x00,
  0x0F7F, 0xFF,
  0x0F80, 0x01,
  0x0F81, 0x68,
  0x0F82, 0x01,
  0x0F83, 0x41,
  0x0F84, 0x00,
  0x0F85, 0x00,
  0x0F86, 0x00,
  0x0F87, 0x00,
  0x0F88, 0x00,
  0x0F89, 0x00,
  0x0F8A, 0x00,
  0x0F8B, 0x00,
  0x0F8C, 0x00,
  0x0F8D, 0x00,
  0x0F8E, 0x00,
  0x0F8F, 0x00,
  0x0F90, 0x00,
  0x0F91, 0x00,
  0x0F92, 0x00,
  0x0F93, 0x00,
  0x0F94, 0x00,
  0x0F95, 0x00,
  0x0F96, 0x5E,
  0x0F97, 0x00,
  0xffff, 0xff,
};

const T_I2CRegWriteData SensorFocusWnd::i2cdat_dp2_[] = {
  0x1C00, 0x00,
  0x1C01, 0x00,
  0x1C02, 0x77,
  0x1C03, 0x7E,
  0x1C04, 0x02,
  0x1C05, 0x0F,
  0x1C06, 0x24,
  0x1C07, 0x1B,
  0x1C08, 0x00,
  0x1C09, 0x01,
  0x1C0A, 0x00,
  0x1C0B, 0x00,
  0x1C0C, 0x00,
  0x1C0D, 0x22,
  0x1C0E, 0x04,
  0x1C0F, 0x0B,
  0x1C10, 0x80,
  0x1C11, 0x45,
  0x1C12, 0x00,
  0x1C13, 0x00,
  0x1C14, 0x00,
  0x1C15, 0x00,
  0x1C16, 0x02,
  0x1C17, 0x14,
  0x1C18, 0x00,
  0x1C19, 0x01,
  0x1C1A, 0x00,
  0x1C1B, 0x01,
  0x1C1C, 0x08,
  0x1C1D, 0x00,
  0x1C1E, 0x00,
  0x1C1F, 0x02,
  0x1C20, 0x10,
  0x1C21, 0x10,
  0x1C22, 0x10,
  0x1C23, 0x03,
  0x1C24, 0x0C,
  0x1C25, 0x00,
  0x1C26, 0x0C,
  0x1C27, 0x00,
  0x1C28, 0x04,
  0x1C29, 0x00,
  0x1C2A, 0x04,
  0x1C2B, 0x00,
  0x1C2C, 0x0C,
  0x1C2D, 0x30,
  0x1C2E, 0x01,
  0x1C2F, 0x11,
  0x1C30, 0x00,
  0x1C31, 0xF0,
  0x1C32, 0x22,
  0x1C33, 0x71,
  0x1C34, 0x02,
  0x1C35, 0x27,
  0x1C36, 0x00,
  0x1C37, 0xF0,
  0x1C38, 0x83,
  0x1C39, 0xC0,
  0x1C3A, 0x05,
  0x1C3B, 0x00,
  0x1C3C, 0x02,
  0x1C3D, 0x32,
  0x1C3E, 0x00,
  0x1C3F, 0x02,
  0x1C40, 0x00,
  0x1C41, 0xC0,
  0x1C42, 0x00,
  0x1C43, 0x00,
  0x1C44, 0x00,
  0x1C45, 0x20,
  0x1C46, 0x00,
  0x1C47, 0x00,
  0x1C48, 0xD0,
  0x1C49, 0x00,
  0x1C4A, 0xD0,
  0x1C4B, 0x00,
  0x1C4C, 0xD0,
  0x1C4D, 0x00,
  0x1C4E, 0xD0,
  0x1C4F, 0x00,
  0x1C50, 0x50,
  0x1C51, 0x50,
  0x1C52, 0x50,
  0x1C53, 0x50,
  0x1C54, 0x3F,
  0x1C55, 0x3F,
  0x1C56, 0x07,
  0x1C57, 0x80,
  0x1C58, 0x04,
  0x1C59, 0x00,
  0x1C5A, 0x04,
  0x1C5B, 0x00,
  0x1C5C, 0x00,
  0x1C5D, 0x00,
  0x1C5E, 0x00,
  0x1C5F, 0x00,
  0x1C60, 0x20,
  0x1C61, 0x00,
  0x1C62, 0x20,
  0x1C63, 0x00,
  0x1C64, 0x04,
  0x1C65, 0x00,
  0x1C66, 0x04,
  0x1C67, 0x00,
  0x1C68, 0x00,
  0x1C69, 0x04,
  0x1C6A, 0x00,
  0x1C6B, 0x04,
  0x1C6C, 0x00,
  0x1C6D, 0x00,
  0x1C6E, 0x00,
  0x1C6F, 0x04,
  0x1C70, 0x00,
  0x1C71, 0x02,
  0x1C72, 0x02,
  0x1C73, 0x00,
  0x1C74, 0x83,
  0x1C75, 0x91,
  0x1C76, 0x3C,
  0x1C77, 0x6F,
  0x1C78, 0x83,
  0x1C79, 0x9A,
  0x1C7A, 0x3C,
  0x1C7B, 0x66,
  0x1C7C, 0x3F,
  0x1C7D, 0xFE,
  0x1C7E, 0x3F,
  0x1C7F, 0xFC,
  0x1C80, 0x3F,
  0x1C81, 0xFF,
  0x1C82, 0x3F,
  0x1C83, 0xFF,
  0x1C84, 0x40,
  0x1C85, 0x00,
  0x1C86, 0x00,
  0x1C87, 0x00,
  0x1C88, 0x00,
  0x1C89, 0x00,
  0x1C8A, 0x00,
  0x1C8B, 0x00,
  0x1C8C, 0x3E,
  0x1C8D, 0xFC,
  0x1C8E, 0xC1,
  0x1C8F, 0x0A,
  0x1C90, 0x7F,
  0x1C91, 0xF8,
  0x1C92, 0x7F,
  0x1C93, 0xF8,
  0x1C94, 0x1A,
  0x1C95, 0x5E,
  0x1C96, 0x01,
  0x1C97, 0xA5,
  0x1C98, 0x05,
  0x1C99, 0xDE,
  0x1C9A, 0x01,
  0x1C9B, 0x7A,
  0x1C9C, 0x01,
  0x1C9D, 0x0E,
  0x1C9E, 0x28,
  0x1C9F, 0x00,
  0x1CA0, 0x00,
  0x1CA1, 0x02,
  0x1CA2, 0x04,
  0x1CA3, 0x65,
  0x1CA4, 0x01,
  0x1CA5, 0x51,
  0x1CA6, 0x01,
  0x1CA7, 0x00,
  0x1CA8, 0x3F,
  0x1CA9, 0xF2,
  0x1CAA, 0x3F,
  0x1CAB, 0xF2,
  0x1CAC, 0x3F,
  0x1CAD, 0xE0,
  0x1CAE, 0x7F,
  0x1CAF, 0xFF,
  0x1CB0, 0x05,
  0x1CB1, 0xDE,
  0x1CB2, 0x04,
  0x1CB3, 0x65,
  0x1CB4, 0x01,
  0x1CB5, 0x00,
  0x1CB6, 0x11,
  0x1CB7, 0x9A,
  0x1CB8, 0x50,
  0x1CB9, 0x46,
  0x1CBA, 0xB0,
  0x1CBB, 0x9E,
  0x1CBC, 0x7F,
  0x1CBD, 0x74,
  0x1CBE, 0x13,
  0x1CBF, 0x26,
  0x1CC0, 0xDD,
  0x1CC1, 0x45,
  0x1CC2, 0x10,
  0x1CC3, 0x1C,
  0x1CC4, 0x7C,
  0x1CC5, 0xD5,
  0x1CC6, 0x85,
  0x1CC7, 0xD0,
  0x1CC8, 0x21,
  0x1CC9, 0x87,
  0x1CCA, 0xBE,
  0x1CCB, 0x1E,
  0x1CCC, 0x20,
  0x1CCD, 0x65,
  0x1CCE, 0x7E,
  0x1CCF, 0xDA,
  0x1CD0, 0x82,
  0x1CD1, 0x38,
  0x1CD2, 0x50,
  0x1CD3, 0x00,
  0x1CD4, 0x30,
  0x1CD5, 0x00,
  0x1CD6, 0x30,
  0x1CD7, 0x00,
  0x1CD8, 0x30,
  0x1CD9, 0x00,
  0x1CDA, 0x30,
  0x1CDB, 0x00,
  0x1CDC, 0x00,
  0x1CDD, 0x00,
  0x1CDE, 0x00,
  0x1CDF, 0x00,
  0x1CE0, 0x00,
  0x1CE1, 0x00,
  0x1CE2, 0x00,
  0x1CE3, 0x00,
  0x1CE4, 0x00,
  0x1CE5, 0x00,
  0x1CE6, 0x00,
  0x1CE7, 0x00,
  0x1CE8, 0x10,
  0x1CE9, 0x00,
  0x1CEA, 0x10,
  0x1CEB, 0x00,
  0x1CEC, 0x01,
  0x1CED, 0x18,
  0x1CEE, 0x10,
  0x1CEF, 0x00,
  0x1CF0, 0x00,
  0x1CF1, 0x30,
  0x1CF2, 0x00,
  0x1CF3, 0x40,
  0x1CF4, 0x00,
  0x1CF5, 0x05,
  0x1CF6, 0x00,
  0x1CF7, 0x02,
  0x1CF8, 0x00,
  0x1CF9, 0x00,
  0x1CFA, 0x07,
  0x1CFB, 0xC0,
  0x1CFC, 0x00,
  0x1CFD, 0x01,
  0x1CFE, 0x00,
  0x1CFF, 0x00,
  0x1D00, 0xB4,
  0x1D01, 0x70,
  0x1D02, 0x12,
  0x1D03, 0x00,
  0x1D04, 0x07,
  0x1D05, 0x80,
  0x1D06, 0x01,
  0x1D07, 0xFF,
  0x1D08, 0x00,
  0x1D09, 0x00,
  0x1D0A, 0x18,
  0x1D0B, 0x00,
  0x1D0C, 0x18,
  0x1D0D, 0x00,
  0x1D0E, 0x50,
  0x1D0F, 0x00,
  0x1D10, 0x40,
  0x1D11, 0x00,
  0x1D12, 0xC0,
  0x1D13, 0x00,
  0x1D14, 0x3C,
  0x1D15, 0x00,
  0x1D16, 0x40,
  0x1D17, 0x00,
  0x1D18, 0x00,
  0x1D19, 0x00,
  0x1D1A, 0x40,
  0x1D1B, 0x00,
  0x1D1C, 0x06,
  0x1D1D, 0x78,
  0x1D1E, 0x06,
  0x1D1F, 0x78,
  0x1D20, 0x00,
  0x1D21, 0x00,
  0x1D22, 0x33,
  0x1D23, 0x18,
  0x1D24, 0x00,
  0x1D25, 0x00,
  0x1D26, 0xF9,
  0x1D27, 0x88,
  0x1D28, 0xF9,
  0x1D29, 0x88,
  0x1D2A, 0x00,
  0x1D2B, 0x00,
  0x1D2C, 0x33,
  0x1D2D, 0x18,
  0x1D2E, 0x00,
  0x1D2F, 0x00,
  0x1D30, 0x7F,
  0x1D31, 0xFF,
  0x1D32, 0x00,
  0x1D33, 0x00,
  0x1D34, 0x00,
  0x1D35, 0x00,
  0x1D36, 0xC0,
  0x1D37, 0x00,
  0x1D38, 0x00,
  0x1D39, 0x00,
  0x1D3A, 0x00,
  0x1D3B, 0x00,
  0x1D3C, 0x00,
  0x1D3D, 0x00,
  0x1D3E, 0x00,
  0x1D3F, 0x00,
  0x1D40, 0x00,
  0x1D41, 0x00,
  0x1D42, 0x00,
  0x1D43, 0x00,
  0x1D44, 0x00,
  0x1D45, 0x00,
  0x1D46, 0x00,
  0x1D47, 0x00,
  0x1D48, 0x00,
  0x1D49, 0x00,
  0x1D4A, 0x00,
  0x1D4B, 0x00,
  0x1D4C, 0x00,
  0x1D4D, 0x00,
  0x1D4E, 0x00,
  0x1D4F, 0x00,
  0x1D50, 0x00,
  0x1D51, 0x00,
  0x1D52, 0x00,
  0x1D53, 0x00,
  0x1D54, 0x00,
  0x1D55, 0x00,
  0x1D56, 0x00,
  0x1D57, 0x00,
  0x1D58, 0x00,
  0x1D59, 0x00,
  0x1D5A, 0x00,
  0x1D5B, 0x00,
  0x1D5C, 0x00,
  0x1D5D, 0x00,
  0x1D5E, 0x00,
  0x1D5F, 0x00,
  0x1D60, 0x00,
  0x1D61, 0x00,
  0x1D62, 0x00,
  0x1D63, 0x00,
  0x1D64, 0x03,
  0x1D65, 0x6D,
  0x1D66, 0x01,
  0x1D67, 0x23,
  0x1D68, 0x06,
  0x1D69, 0x54,
  0x1D6A, 0x05,
  0x1D6B, 0x42,
  0x1D6C, 0x24,
  0x1D6D, 0x01,
  0x1D6E, 0x1E,
  0x1D6F, 0x12,
  0x1D70, 0x01,
  0x1D71, 0x01,
  0x1D72, 0x18,
  0x1D73, 0x14,
  0x1D74, 0x01,
  0x1D75, 0x01,
  0x1D76, 0x18,
  0x1D77, 0x14,
  0x1D78, 0x0F,
  0x1D79, 0x0F,
  0x1D7A, 0x18,
  0x1D7B, 0x10,
  0x1D7C, 0x0F,
  0x1D7D, 0x0F,
  0x1D7E, 0x18,
  0x1D7F, 0x10,
  0x1D80, 0x00,
  0x1D81, 0x01,
  0x1D82, 0x00,
  0x1D83, 0x05,
  0x1D84, 0x00,
  0x1D85, 0x01,
  0x1D86, 0x00,
  0x1D87, 0x05,
  0x1D88, 0x00,
  0x1D89, 0x01,
  0x1D8A, 0x00,
  0x1D8B, 0x02,
  0x1D8C, 0x00,
  0x1D8D, 0x01,
  0x1D8E, 0x00,
  0x1D8F, 0x05,
  0x1D90, 0x00,
  0x1D91, 0x00,
  0x1D92, 0x01,
  0x1D93, 0x84,
  0x1D94, 0xE0,
  0x1D95, 0xE0,
  0x1D96, 0x00,
  0x1D97, 0x80,
  0x1D98, 0x00,
  0x1D99, 0xC3,
  0x1D9A, 0xC3,
  0x1D9B, 0x00,
  0x1D9C, 0x00,
  0x1D9D, 0x00,
  0x1D9E, 0x00,
  0x1D9F, 0x00,
  0x1DA0, 0x00,
  0x1DA1, 0x00,
  0x1DA2, 0x00,
  0x1DA3, 0x00,
  0x1DA4, 0x00,
  0x1DA5, 0x00,
  0x1DA6, 0x00,
  0x1DA7, 0x00,
  0x1DA8, 0x00,
  0x1DA9, 0x00,
  0x1DAA, 0x00,
  0x1DAB, 0x00,
  0x1DAC, 0x00,
  0x1DAD, 0x10,
  0x1DAE, 0x00,
  0x1DAF, 0x00,
  0x1DB0, 0x00,
  0x1DB1, 0x00,
  0x1DB2, 0x00,
  0x1DB3, 0x00,
  0x1DB4, 0x7F,
  0x1DB5, 0xF4,
  0x1DB6, 0x41,
  0x1DB7, 0xEB,
  0x1DB8, 0x00,
  0x1DB9, 0x00,
  0x1DBA, 0x00,
  0x1DBB, 0x00,
  0x1DBC, 0x00,
  0x1DBD, 0x00,
  0x1DBE, 0x00,
  0x1DBF, 0x00,
  0xffff, 0xff,
};

const T_I2CRegWriteData SensorFocusWnd::i2cdat_calb_[] = {
  0x1DC0, 0x00,
  0x1DC1, 0x20,
  0x1DC2, 0x00,
  0x1DC3, 0x20,
  0x1DC4, 0x02,
  0x1DC5, 0x00,
  0x1DC6, 0x02,
  0x1DC7, 0x00,
  0x1DC8, 0x00,
  0x1DC9, 0x00,
  0x1DCA, 0x00,
  0x1DCB, 0x00,
  0x1DCC, 0x00,
  0x1DCD, 0x10,
  0x1DCE, 0x00,
  0x1DCF, 0x10,
  0x1DD0, 0x08,
  0x1DD1, 0x00,
  0x1DD2, 0x08,
  0x1DD3, 0x00,
  0x1DD4, 0x00,
  0x1DD5, 0x80,
  0x1DD6, 0x08,
  0x1DD7, 0x00,
  0x1DD8, 0x08,
  0x1DD9, 0x00,
  0x1DDA, 0x29,
  0x1DDB, 0x24,
  0x1DDC, 0xD6,
  0x1DDD, 0xDC,
  0x1DDE, 0x20,
  0x1DDF, 0x00,
  0x1DE0, 0x20,
  0x1DE1, 0x00,
  0x1DE2, 0x00,
  0x1DE3, 0x00,
  0x1DE4, 0x00,
  0x1DE5, 0x00,
  0x1DE6, 0x00,
  0x1DE7, 0x50,
  0x1DE8, 0x02,
  0x1DE9, 0x90,
  0x1DEA, 0x00,
  0x1DEB, 0x10,
  0x1DEC, 0x00,
  0x1DED, 0x00,
  0x1DEE, 0xE0,
  0x1DEF, 0x00,
  0xffff, 0xff,
};

const T_I2CRegWriteData SensorFocusWnd::i2cdat_gpo_[] = {
  0x4288, 0x00,
  0x42B0, 0x00,
  0x4BDD, 0x15,
  0x42A8, 0xFF,
  0x42AA, 0xFD,
  0x4270, 0x00,
  0xffff, 0xff,
};

const T_I2CRegWriteData SensorFocusWnd::i2cdat_pwmfrequency_[] = {
  0x60D0, 0x00,
  0x60D4, 0x0A,
  0x60D1, 0x00,
  0x60D5, 0x00,
  0x60D2, 0x5F,
  0x60D3, 0x00,
  0x60D4, 0x0A,
  0x60D0, 0x01,
  0xffff, 0xff,
};

const T_I2CRegWriteData SensorFocusWnd::i2cdat_SineTest_dp1_[] = {
  0x0000, 0x24,
  0x0001, 0x00,
  0x0002, 0xB5,
  0x0003, 0x70,
  0x0004, 0x28,
  0x0005, 0x00,
  0x0006, 0x46,
  0x0007, 0x22,
  0x0008, 0x4A,
  0x0009, 0x24,
  0x000A, 0xDA,
  0x000B, 0x10,
  0x000C, 0xD0,
  0x000D, 0x04,
  0x000E, 0x18,
  0x000F, 0x80,
  0x0010, 0xD0,
  0x0011, 0x04,
  0x0012, 0x28,
  0x0013, 0x01,
  0x0014, 0xD1,
  0x0015, 0x3E,
  0x0016, 0x28,
  0x0017, 0x04,
  0x0018, 0x4A,
  0x0019, 0x21,
  0x001A, 0xE0,
  0x001B, 0x03,
  0x001C, 0x4A,
  0x001D, 0x21,
  0x001E, 0xE0,
  0x001F, 0x02,
  0x0020, 0x4A,
  0x0021, 0x21,
  0x0022, 0xE0,
  0x0023, 0x00,
  0x0024, 0x47,
  0x0025, 0x90,
  0x0026, 0x46,
  0x0027, 0x08,
  0x0028, 0xE0,
  0x0029, 0x34,
  0x002A, 0x46,
  0x002B, 0x04,
  0x002C, 0x35,
  0x002D, 0x32,
  0x002E, 0x25,
  0x002F, 0xFF,
  0x0030, 0x42,
  0x0031, 0xA8,
  0x0032, 0x1B,
  0x0033, 0x43,
  0x0034, 0xDC,
  0x0035, 0x12,
  0x0036, 0xD0,
  0x0037, 0x29,
  0x0038, 0xD0,
  0x0039, 0x26,
  0x003A, 0x28,
  0x003B, 0x3D,
  0x003C, 0x28,
  0x003D, 0x3A,
  0x003E, 0xDC,
  0x003F, 0x06,
  0x0040, 0x28,
  0x0041, 0x3B,
  0x0042, 0xD0,
  0x0043, 0x23,
  0x0044, 0x28,
  0x0045, 0x3C,
  0x0046, 0xD0,
  0x0047, 0x21,
  0x0048, 0xE0,
  0x0049, 0x1E,
  0x004A, 0xD1,
  0x004B, 0x25,
  0x004C, 0xD0,
  0x004D, 0x15,
  0x004E, 0x28,
  0x004F, 0x3E,
  0x0050, 0xD0,
  0x0051, 0x17,
  0x0052, 0x28,
  0x0053, 0x3F,
  0x0054, 0x3A,
  0x0055, 0xFF,
  0x0056, 0x46,
  0x0057, 0x02,
  0x0058, 0xD1,
  0x0059, 0x1C,
  0x005A, 0x3A,
  0x005B, 0x31,
  0x005C, 0x2B,
  0x005D, 0x05,
  0x005E, 0xE0,
  0x005F, 0x14,
  0x0060, 0xDC,
  0x0061, 0x06,
  0x0062, 0xD0,
  0x0063, 0x13,
  0x0064, 0xD0,
  0x0065, 0x10,
  0x0066, 0x2B,
  0x0067, 0x01,
  0x0068, 0xD0,
  0x0069, 0x0E,
  0x006A, 0x2B,
  0x006B, 0x03,
  0x006C, 0xD1,
  0x006D, 0x12,
  0x006E, 0x2B,
  0x006F, 0x04,
  0x0070, 0x2B,
  0x0071, 0x06,
  0x0072, 0xE0,
  0x0073, 0x0B,
  0x0074, 0x2B,
  0x0075, 0x1D,
  0x0076, 0xD0,
  0x0077, 0x09,
  0x0078, 0xE0,
  0x0079, 0x01,
  0x007A, 0xD1,
  0x007B, 0x0D,
  0x007C, 0xE0,
  0x007D, 0x04,
  0x007E, 0x4A,
  0x007F, 0x0B,
  0x0080, 0xE0,
  0x0081, 0x02,
  0x0082, 0x4A,
  0x0083, 0x0B,
  0x0084, 0xE0,
  0x0085, 0x00,
  0x0086, 0x4A,
  0x0087, 0x0B,
  0x0088, 0x4B,
  0x0089, 0x0C,
  0x008A, 0x4A,
  0x008B, 0x0B,
  0x008C, 0x2A,
  0x008D, 0x00,
  0x008E, 0x54,
  0x008F, 0x19,
  0x0090, 0xB2,
  0x0091, 0xC8,
  0x0092, 0xD0,
  0x0093, 0x01,
  0x0094, 0x46,
  0x0095, 0x20,
  0x0096, 0x47,
  0x0097, 0x90,
  0x0098, 0x00,
  0x0099, 0x00,
  0x009A, 0xBD,
  0x009B, 0x70,
  0x009C, 0x7F,
  0x009D, 0xFF,
  0x009E, 0xFF,
  0x009F, 0xD6,
  0x00A0, 0x10,
  0x00A1, 0x00,
  0x00A2, 0x01,
  0x00A3, 0xE5,
  0x00A4, 0x10,
  0x00A5, 0x00,
  0x00A6, 0x01,
  0x00A7, 0x97,
  0x00A8, 0x10,
  0x00A9, 0x00,
  0x00AA, 0x00,
  0x00AB, 0xC1,
  0x00AC, 0x10,
  0x00AD, 0x00,
  0x00AE, 0x08,
  0x00AF, 0x01,
  0x00B0, 0x10,
  0x00B1, 0x00,
  0x00B2, 0x07,
  0x00B3, 0x55,
  0x00B4, 0x10,
  0x00B5, 0x00,
  0x00B6, 0x08,
  0x00B7, 0x0D,
  0x00B8, 0x10,
  0x00B9, 0x00,
  0x00BA, 0x06,
  0x00BB, 0x21,
  0x00BC, 0x10,
  0x00BD, 0x00,
  0x00BE, 0x19,
  0x00BF, 0x40,
  0x00C0, 0x4D,
  0x00C1, 0x64,
  0x00C2, 0xB5,
  0x00C3, 0xF8,
  0x00C4, 0x4E,
  0x00C5, 0x63,
  0x00C6, 0x46,
  0x00C7, 0x04,
  0x00C8, 0xF0,
  0x00C9, 0x00,
  0x00CA, 0x35,
  0x00CB, 0xB4,
  0x00CC, 0x4A,
  0x00CD, 0x62,
  0x00CE, 0xFA,
  0x00CF, 0x15,
  0x00D0, 0xD0,
  0x00D1, 0x19,
  0x00D2, 0x2C,
  0x00D3, 0x00,
  0x00D4, 0x60,
  0x00D5, 0x10,
  0x00D6, 0x6B,
  0x00D7, 0x60,
  0x00D8, 0x60,
  0x00D9, 0x50,
  0x00DA, 0x6B,
  0x00DB, 0xA0,
  0x00DC, 0x60,
  0x00DD, 0x90,
  0x00DE, 0x69,
  0x00DF, 0x60,
  0x00E0, 0x60,
  0x00E1, 0xD0,
  0x00E2, 0x68,
  0x00E3, 0xA0,
  0x00E4, 0x61,
  0x00E5, 0x10,
  0x00E6, 0x68,
  0x00E7, 0xE0,
  0x00E8, 0x61,
  0x00E9, 0x50,
  0x00EA, 0x69,
  0x00EB, 0x20,
  0x00EC, 0x61,
  0x00ED, 0x90,
  0x00EE, 0x6B,
  0x00EF, 0x20,
  0x00F0, 0x61,
  0x00F1, 0xD0,
  0x00F2, 0x69,
  0x00F3, 0xE0,
  0x00F4, 0x62,
  0x00F5, 0x10,
  0x00F6, 0x6A,
  0x00F7, 0xA0,
  0x00F8, 0x62,
  0x00F9, 0x50,
  0x00FA, 0x6A,
  0x00FB, 0x20,
  0x00FC, 0x62,
  0x00FD, 0x90,
  0x00FE, 0x6A,
  0x00FF, 0xE0,
  0x0100, 0x62,
  0x0101, 0xD0,
  0x0102, 0x6C,
  0x0103, 0xE0,
  0x0104, 0x63,
  0x0105, 0x10,
  0x0106, 0x6D,
  0x0107, 0x60,
  0x0108, 0x78,
  0x0109, 0x20,
  0x010A, 0x4C,
  0x010B, 0x54,
  0x010C, 0x21,
  0x010D, 0x02,
  0x010E, 0x4F,
  0x010F, 0x52,
  0x0110, 0x20,
  0x0111, 0x01,
  0x0112, 0x69,
  0x0113, 0x7A,
  0x0114, 0x21,
  0x0115, 0x19,
  0x0116, 0x47,
  0x0117, 0x90,
  0x0118, 0x63,
  0x0119, 0xC8,
  0x011A, 0x06,
  0x011B, 0xC9,
  0x011C, 0x20,
  0x011D, 0x00,
  0x011E, 0x49,
  0x011F, 0x50,
  0x0120, 0x49,
  0x0121, 0x4F,
  0x0122, 0x61,
  0x0123, 0x08,
  0x0124, 0x39,
  0x0125, 0xC0,
  0x0126, 0x48,
  0x0127, 0x4F,
  0x0128, 0x78,
  0x0129, 0x24,
  0x012A, 0x61,
  0x012B, 0x88,
  0x012C, 0x69,
  0x012D, 0x7A,
  0x012E, 0x21,
  0x012F, 0x42,
  0x0130, 0x47,
  0x0131, 0x90,
  0x0132, 0x20,
  0x0133, 0x06,
  0x0134, 0x06,
  0x0135, 0x00,
  0x0136, 0x07,
  0x0137, 0x07,
  0x0138, 0x0F,
  0x0139, 0x00,
  0x013A, 0x0F,
  0x013B, 0x3F,
  0x013C, 0xD1,
  0x013D, 0x28,
  0x013E, 0x28,
  0x013F, 0x02,
  0x0140, 0xD0,
  0x0141, 0x04,
  0x0142, 0x2C,
  0x0143, 0x13,
  0x0144, 0xD0,
  0x0145, 0x06,
  0x0146, 0x2C,
  0x0147, 0x41,
  0x0148, 0xD0,
  0x0149, 0x09,
  0x014A, 0x2C,
  0x014B, 0x42,
  0x014C, 0x4D,
  0x014D, 0x46,
  0x014E, 0xE0,
  0x014F, 0x0C,
  0x0150, 0x35,
  0x0151, 0xB4,
  0x0152, 0x4E,
  0x0153, 0x45,
  0x0154, 0x4E,
  0x0155, 0x44,
  0x0156, 0xE0,
  0x0157, 0x08,
  0x0158, 0x36,
  0x0159, 0x08,
  0x015A, 0x4D,
  0x015B, 0x43,
  0x015C, 0xE0,
  0x015D, 0x03,
  0x015E, 0x35,
  0x015F, 0xBC,
  0x0160, 0x4D,
  0x0161, 0x41,
  0x0162, 0x4E,
  0x0163, 0x3C,
  0x0164, 0x3D,
  0x0165, 0x84,
  0x0166, 0x36,
  0x0167, 0x14,
  0x0168, 0x21,
  0x0169, 0x43,
  0x016A, 0x4C,
  0x016B, 0x3B,
  0x016C, 0x20,
  0x016D, 0x06,
  0x016E, 0x69,
  0x016F, 0x62,
  0x0170, 0x69,
  0x0171, 0xA1,
  0x0172, 0x47,
  0x0173, 0x90,
  0x0174, 0x20,
  0x0175, 0x0B,
  0x0176, 0x47,
  0x0177, 0x88,
  0x0178, 0x20,
  0x0179, 0x13,
  0x017A, 0x47,
  0x017B, 0xB0,
  0x017C, 0x20,
  0x017D, 0x0C,
  0x017E, 0x47,
  0x017F, 0xA8,
  0x0180, 0x20,
  0x0181, 0x80,
  0x0182, 0x47,
  0x0183, 0xB0,
  0x0184, 0x2F,
  0x0185, 0x01,
  0x0186, 0x47,
  0x0187, 0xA8,
  0x0188, 0x20,
  0x0189, 0x1F,
  0x018A, 0xD1,
  0x018B, 0x03,
  0x018C, 0x20,
  0x018D, 0x08,
  0x018E, 0x47,
  0x018F, 0xB0,
  0x0190, 0x20,
  0x0191, 0x00,
  0x0192, 0x47,
  0x0193, 0xA8,
  0x0194, 0xB5,
  0x0195, 0x00,
  0x0196, 0xBD,
  0x0197, 0xF8,
  0x0198, 0xB0,
  0x0199, 0x87,
  0x019A, 0x4B,
  0x019B, 0x2F,
  0x019C, 0x28,
  0x019D, 0x00,
  0x019E, 0x6B,
  0x019F, 0x59,
  0x01A0, 0x28,
  0x01A1, 0x04,
  0x01A2, 0xD0,
  0x01A3, 0x01,
  0x01A4, 0x29,
  0x01A5, 0x01,
  0x01A6, 0xD1,
  0x01A7, 0x1B,
  0x01A8, 0x20,
  0x01A9, 0x00,
  0x01AA, 0xD1,
  0x01AB, 0x19,
  0x01AC, 0x68,
  0x01AD, 0x09,
  0x01AE, 0x07,
  0x01AF, 0x89,
  0x01B0, 0xD0,
  0x01B1, 0x0E,
  0x01B2, 0x07,
  0x01B3, 0x89,
  0x01B4, 0x91,
  0x01B5, 0x04,
  0x01B6, 0x49,
  0x01B7, 0x2D,
  0x01B8, 0x91,
  0x01B9, 0x05,
  0x01BA, 0x49,
  0x01BB, 0x2D,
  0x01BC, 0x49,
  0x01BD, 0x2D,
  0x01BE, 0x90,
  0x01BF, 0x01,
  0x01C0, 0x91,
  0x01C1, 0x06,
  0x01C2, 0x90,
  0x01C3, 0x02,
  0x01C4, 0x68,
  0x01C5, 0x1B,
  0x01C6, 0x90,
  0x01C7, 0x03,
  0x01C8, 0xA9,
  0x01C9, 0x01,
  0x01CA, 0x22,
  0x01CB, 0x01,
  0x01CC, 0x47,
  0x01CD, 0x98,
  0x01CE, 0xA8,
  0x01CF, 0x04,
  0x01D0, 0x49,
  0x01D1, 0x23,
  0x01D2, 0xE0,
  0x01D3, 0x05,
  0x01D4, 0x49,
  0x01D5, 0x28,
  0x01D6, 0x61,
  0x01D7, 0x08,
  0x01D8, 0x49,
  0x01D9, 0x28,
  0x01DA, 0x63,
  0x01DB, 0x88,
  0x01DC, 0x20,
  0x01DD, 0x00,
  0x01DE, 0x63,
  0x01DF, 0x88,
  0x01E0, 0xBD,
  0x01E1, 0x00,
  0x01E2, 0xB0,
  0x01E3, 0x07,
  0x01E4, 0x49,
  0x01E5, 0x1D,
  0x01E6, 0xB5,
  0x01E7, 0x70,
  0x01E8, 0x31,
  0x01E9, 0x20,
  0x01EA, 0x4C,
  0x01EB, 0x25,
  0x01EC, 0xB0,
  0x01ED, 0x88,
  0x01EE, 0x78,
  0x01EF, 0x09,
  0x01F0, 0x0F,
  0x01F1, 0x49,
  0x01F2, 0x07,
  0x01F3, 0x49,
  0x01F4, 0x28,
  0x01F5, 0x01,
  0x01F6, 0x4D,
  0x01F7, 0x18,
  0x01F8, 0xD1,
  0x01F9, 0x26,
  0x01FA, 0x63,
  0x01FB, 0x69,
  0x01FC, 0xD0,
  0x01FD, 0x01,
  0x01FE, 0x29,
  0x01FF, 0x00,
  0x0200, 0xD1,
  0x0201, 0x22,
  0x0202, 0x29,
  0x0203, 0x04,
  0x0204, 0x47,
  0x0205, 0xA0,
  0x0206, 0x20,
  0x0207, 0x0E,
  0x0208, 0x20,
  0x0209, 0x0C,
  0x020A, 0x46,
  0x020B, 0x06,
  0x020C, 0x04,
  0x020D, 0x00,
  0x020E, 0x47,
  0x020F, 0xA0,
  0x0210, 0x21,
  0x0211, 0x01,
  0x0212, 0x43,
  0x0213, 0x30,
  0x0214, 0x68,
  0x0215, 0x09,
  0x0216, 0x07,
  0x0217, 0x89,
  0x0218, 0x21,
  0x0219, 0x00,
  0x021A, 0x07,
  0x021B, 0x8A,
  0x021C, 0xD0,
  0x021D, 0x0E,
  0x021E, 0x2A,
  0x021F, 0x00,
  0x0220, 0x92,
  0x0221, 0x04,
  0x0222, 0x4A,
  0x0223, 0x12,
  0x0224, 0x92,
  0x0225, 0x05,
  0x0226, 0x4A,
  0x0227, 0x12,
  0x0228, 0x92,
  0x0229, 0x06,
  0x022A, 0x4A,
  0x022B, 0x12,
  0x022C, 0x91,
  0x022D, 0x03,
  0x022E, 0x91,
  0x022F, 0x02,
  0x0230, 0x68,
  0x0231, 0x2B,
  0x0232, 0x90,
  0x0233, 0x01,
  0x0234, 0xA9,
  0x0235, 0x01,
  0x0236, 0x22,
  0x0237, 0x03,
  0x0238, 0x47,
  0x0239, 0x98,
  0x023A, 0xA8,
  0x023B, 0x04,
  0x023C, 0x4A,
  0x023D, 0x08,
  0x023E, 0xE0,
  0x023F, 0x05,
  0x0240, 0x48,
  0x0241, 0x0D,
  0x0242, 0x61,
  0x0243, 0x10,
  0x0244, 0x48,
  0x0245, 0x0D,
  0x0246, 0x63,
  0x0247, 0x81,
  0x0248, 0x20,
  0x0249, 0x00,
  0x024A, 0x63,
  0x024B, 0x81,
  0x024C, 0xBD,
  0x024D, 0x70,
  0x024E, 0xB0,
  0x024F, 0x08,
  0x0250, 0x00,
  0x0251, 0x00,
  0x0252, 0x47,
  0x0253, 0x70,
  0x0254, 0x00,
  0x0255, 0x00,
  0x0256, 0x29,
  0x0257, 0xD9,
  0x0258, 0x10,
  0x0259, 0x00,
  0x025A, 0x0E,
  0x025B, 0x40,
  0x025C, 0x10,
  0x025D, 0x00,
  0x025E, 0x19,
  0x025F, 0x40,
  0x0260, 0x50,
  0x0261, 0x00,
  0x0262, 0x60,
  0x0263, 0xC0,
  0x0264, 0x30,
  0x0265, 0x48,
  0x0266, 0x30,
  0x0267, 0x48,
  0x0268, 0x00,
  0x0269, 0x00,
  0x026A, 0x2B,
  0x026B, 0x25,
  0x026C, 0x00,
  0x026D, 0x00,
  0x026E, 0x60,
  0x026F, 0xD0,
  0x0270, 0x00,
  0x0271, 0x00,
  0x0272, 0x41,
  0x0273, 0x38,
  0x0274, 0x00,
  0x0275, 0x00,
  0x0276, 0x51,
  0x0277, 0x38,
  0x0278, 0x50,
  0x0279, 0x00,
  0x027A, 0x41,
  0x027B, 0x00,
  0x027C, 0x50,
  0x027D, 0x00,
  0x027E, 0x51,
  0x027F, 0x00,
  0x0280, 0x00,
  0x0281, 0x00,
  0x0282, 0x07,
  0x0283, 0x7B,
  0x0284, 0x49,
  0x0285, 0x1A,
  0x0286, 0xB4,
  0x0287, 0x30,
  0x0288, 0x2A,
  0x0289, 0x01,
  0x028A, 0x78,
  0x028B, 0x0A,
  0x028C, 0x22,
  0x028D, 0x19,
  0x028E, 0xD0,
  0x028F, 0x17,
  0x0290, 0x68,
  0x0291, 0x13,
  0x0292, 0x06,
  0x0293, 0xD2,
  0x0294, 0x03,
  0x0295, 0x24,
  0x0296, 0x24,
  0x0297, 0x01,
  0x0298, 0x60,
  0x0299, 0x13,
  0x029A, 0x43,
  0x029B, 0xA3,
  0x029C, 0x70,
  0x029D, 0x0A,
  0x029E, 0x22,
  0x029F, 0x01,
  0x02A0, 0x21,
  0x02A1, 0x04,
  0x02A2, 0x4A,
  0x02A3, 0x14,
  0x02A4, 0x49,
  0x02A5, 0x14,
  0x02A6, 0x60,
  0x02A7, 0x11,
  0x02A8, 0x08,
  0x02A9, 0x52,
  0x02AA, 0x68,
  0x02AB, 0x8A,
  0x02AC, 0x60,
  0x02AD, 0x8A,
  0x02AE, 0x00,
  0x02AF, 0x52,
  0x02B0, 0x00,
  0x02B1, 0x80,
  0x02B2, 0x49,
  0x02B3, 0x12,
  0x02B4, 0x49,
  0x02B5, 0x12,
  0x02B6, 0x58,
  0x02B7, 0x08,
  0x02B8, 0xBC,
  0x02B9, 0x30,
  0x02BA, 0x68,
  0x02BB, 0x09,
  0x02BC, 0xBC,
  0x02BD, 0x30,
  0x02BE, 0x47,
  0x02BF, 0x08,
  0x02C0, 0xB5,
  0x02C1, 0x10,
  0x02C2, 0x47,
  0x02C3, 0x70,
  0x02C4, 0x68,
  0x02C5, 0x00,
  0x02C6, 0x48,
  0x02C7, 0x0F,
  0x02C8, 0x48,
  0x02C9, 0x0B,
  0x02CA, 0x47,
  0x02CB, 0x80,
  0x02CC, 0x22,
  0x02CD, 0x01,
  0x02CE, 0x68,
  0x02CF, 0x81,
  0x02D0, 0x60,
  0x02D1, 0x81,
  0x02D2, 0x43,
  0x02D3, 0x11,
  0x02D4, 0x20,
  0x02D5, 0x04,
  0x02D6, 0x49,
  0x02D7, 0x07,
  0x02D8, 0x60,
  0x02D9, 0x08,
  0x02DA, 0x31,
  0x02DB, 0x80,
  0x02DC, 0x20,
  0x02DD, 0x00,
  0x02DE, 0x49,
  0x02DF, 0x04,
  0x02E0, 0x20,
  0x02E1, 0x19,
  0x02E2, 0x70,
  0x02E3, 0x08,
  0x02E4, 0x68,
  0x02E5, 0x01,
  0x02E6, 0x06,
  0x02E7, 0xC0,
  0x02E8, 0x43,
  0x02E9, 0x11,
  0x02EA, 0x03,
  0x02EB, 0x12,
  0x02EC, 0xBD,
  0x02ED, 0x10,
  0x02EE, 0x60,
  0x02EF, 0x01,
  0x02F0, 0x10,
  0x02F1, 0x00,
  0x02F2, 0x0E,
  0x02F3, 0xAC,
  0x02F4, 0xE0,
  0x02F5, 0x00,
  0x02F6, 0xE1,
  0x02F7, 0x00,
  0x02F8, 0xC6,
  0x02F9, 0x00,
  0x02FA, 0x00,
  0x02FB, 0x40,
  0x02FC, 0x10,
  0x02FD, 0x00,
  0x02FE, 0x0E,
  0x02FF, 0x18,
  0x0300, 0x10,
  0x0301, 0x00,
  0x0302, 0x0E,
  0x0303, 0x64,
  0x0304, 0x10,
  0x0305, 0x00,
  0x0306, 0x0E,
  0x0307, 0x68,
  0x0308, 0x47,
  0x0309, 0x70,
  0x030A, 0xBA,
  0x030B, 0x40,
  0x030C, 0x88,
  0x030D, 0x0C,
  0x030E, 0xB5,
  0x030F, 0x30,
  0x0310, 0x2C,
  0x0311, 0x00,
  0x0312, 0x88,
  0x0313, 0x03,
  0x0314, 0x2C,
  0x0315, 0x02,
  0x0316, 0xD0,
  0x0317, 0x06,
  0x0318, 0x2C,
  0x0319, 0x01,
  0x031A, 0xD0,
  0x031B, 0x04,
  0x031C, 0x2C,
  0x031D, 0x03,
  0x031E, 0xD0,
  0x031F, 0x0B,
  0x0320, 0xE0,
  0x0321, 0x0B,
  0x0322, 0xD0,
  0x0323, 0x09,
  0x0324, 0x3D,
  0x0325, 0xF9,
  0x0326, 0x1F,
  0x0327, 0xDD,
  0x0328, 0x1C,
  0x0329, 0x64,
  0x032A, 0xD1,
  0x032B, 0x03,
  0x032C, 0x0F,
  0x032D, 0xA4,
  0x032E, 0x07,
  0x032F, 0xA4,
  0x0330, 0x18,
  0x0331, 0x9B,
  0x0332, 0xE0,
  0x0333, 0x03,
  0x0334, 0x2B,
  0x0335, 0x00,
  0x0336, 0xE0,
  0x0337, 0x02,
  0x0338, 0x1A,
  0x0339, 0x9B,
  0x033A, 0xD0,
  0x033B, 0x03,
  0x033C, 0x80,
  0x033D, 0x0C,
  0x033E, 0x80,
  0x033F, 0x03,
  0x0340, 0x1C,
  0x0341, 0x64,
  0x0342, 0xBD,
  0x0343, 0x30,
  0x0344, 0x0F,
  0x0345, 0xA4,
  0x0346, 0x07,
  0x0347, 0xA4,
  0x0348, 0xE7,
  0x0349, 0xF7,
  0x034A, 0x46,
  0x034B, 0x13,
  0x034C, 0x05,
  0x034D, 0x00,
  0x034E, 0xB5,
  0x034F, 0x30,
  0x0350, 0x4A,
  0x0351, 0x0A,
  0x0352, 0x11,
  0x0353, 0x04,
  0x0354, 0x4B,
  0x0355, 0x08,
  0x0356, 0x05,
  0x0357, 0x08,
  0x0358, 0x06,
  0x0359, 0x10,
  0x035A, 0x11,
  0x035B, 0x05,
  0x035C, 0x29,
  0x035D, 0x00,
  0x035E, 0x6A,
  0x035F, 0x41,
  0x0360, 0x49,
  0x0361, 0x07,
  0x0362, 0xD1,
  0x0363, 0xFC,
  0x0364, 0x60,
  0x0365, 0x8A,
  0x0366, 0x60,
  0x0367, 0x4B,
  0x0368, 0x63,
  0x0369, 0x0D,
  0x036A, 0x62,
  0x036B, 0xCC,
  0x036C, 0x62,
  0x036D, 0x41,
  0x036E, 0x21,
  0x036F, 0x60,
  0x0370, 0x29,
  0x0371, 0x00,
  0x0372, 0x6A,
  0x0373, 0x41,
  0x0374, 0xBD,
  0x0375, 0x30,
  0x0376, 0xD1,
  0x0377, 0xFC,
  0x0378, 0x00,
  0x0379, 0x00,
  0x037A, 0x42,
  0x037B, 0x40,
  0x037C, 0x00,
  0x037D, 0x00,
  0x037E, 0x52,
  0x037F, 0x40,
  0x0380, 0x40,
  0x0381, 0x00,
  0x0382, 0x00,
  0x0383, 0x40,
  0x0384, 0x48,
  0x0385, 0x01,
  0x0386, 0xB4,
  0x0387, 0x03,
  0x0388, 0xBD,
  0x0389, 0x01,
  0x038A, 0x90,
  0x038B, 0x01,
  0x038C, 0x00,
  0x038D, 0x00,
  0x038E, 0x00,
  0x038F, 0xC1,
  0x0390, 0x46,
  0x0391, 0x0B,
  0x0392, 0xB5,
  0x0393, 0xF0,
  0x0394, 0x21,
  0x0395, 0x00,
  0x0396, 0xB0,
  0x0397, 0x85,
  0x0398, 0x46,
  0x0399, 0x6C,
  0x039A, 0x25,
  0x039B, 0x03,
  0x039C, 0xE0,
  0x039D, 0x05,
  0x039E, 0x03,
  0x039F, 0x6D,
  0x03A0, 0x00,
  0x03A1, 0xB6,
  0x03A2, 0x18,
  0x03A3, 0x5E,
  0x03A4, 0x00,
  0x03A5, 0x8F,
  0x03A6, 0x43,
  0x03A7, 0x2E,
  0x03A8, 0x1C,
  0x03A9, 0x49,
  0x03AA, 0x51,
  0x03AB, 0xE6,
  0x03AC, 0xD3,
  0x03AD, 0xF7,
  0x03AE, 0x42,
  0x03AF, 0x91,
  0x03B0, 0x46,
  0x03B1, 0x01,
  0x03B2, 0x4B,
  0x03B3, 0x03,
  0x03B4, 0x46,
  0x03B5, 0x20,
  0x03B6, 0x68,
  0x03B7, 0x1B,
  0x03B8, 0xB0,
  0x03B9, 0x05,
  0x03BA, 0x47,
  0x03BB, 0x98,
  0x03BC, 0x00,
  0x03BD, 0x00,
  0x03BE, 0xBD,
  0x03BF, 0xF0,
  0x03C0, 0x10,
  0x03C1, 0x00,
  0x03C2, 0x0E,
  0x03C3, 0x40,
  0x03C4, 0x4F,
  0x03C5, 0x25,
  0x03C6, 0xB5,
  0x03C7, 0xFE,
  0x03C8, 0x69,
  0x03C9, 0xBE,
  0x03CA, 0x69,
  0x03CB, 0x7D,
  0x03CC, 0xD1,
  0x03CD, 0x06,
  0x03CE, 0x2D,
  0x03CF, 0x02,
  0x03D0, 0x22,
  0x03D1, 0x02,
  0x03D2, 0x4B,
  0x03D3, 0x23,
  0x03D4, 0x68,
  0x03D5, 0x1B,
  0x03D6, 0x46,
  0x03D7, 0x38,
  0x03D8, 0x30,
  0x03D9, 0x24,
  0x03DA, 0x46,
  0x03DB, 0x69,
  0x03DC, 0x68,
  0x03DD, 0xF8,
  0x03DE, 0x47,
  0x03DF, 0x98,
  0x03E0, 0x0F,
  0x03E1, 0xE4,
  0x03E2, 0x07,
  0x03E3, 0xF4,
  0x03E4, 0x2D,
  0x03E5, 0x02,
  0x03E6, 0x90,
  0x03E7, 0x02,
  0x03E8, 0x49,
  0x03E9, 0x1C,
  0x03EA, 0xD0,
  0x03EB, 0x0D,
  0x03EC, 0x31,
  0x03ED, 0x1C,
  0x03EE, 0x00,
  0x03EF, 0xA0,
  0x03F0, 0x2D,
  0x03F1, 0x01,
  0x03F2, 0x58,
  0x03F3, 0x08,
  0x03F4, 0xD0,
  0x03F5, 0x0A,
  0x03F6, 0x68,
  0x03F7, 0x07,
  0x03F8, 0xD0,
  0x03F9, 0x16,
  0x03FA, 0x2D,
  0x03FB, 0x02,
  0x03FC, 0x0D,
  0x03FD, 0x00,
  0x03FE, 0x05,
  0x03FF, 0x38,
  0x0400, 0xD0,
  0x0401, 0x14,
  0x0402, 0x2E,
  0x0403, 0x02,
  0x0404, 0x00,
  0x0405, 0xA0,
  0x0406, 0xE0,
  0x0407, 0x15,
  0x0408, 0x58,
  0x0409, 0x0F,
  0x040A, 0x46,
  0x040B, 0x69,
  0x040C, 0x4A,
  0x040D, 0x15,
  0x040E, 0xE0,
  0x040F, 0x0D,
  0x0410, 0x68,
  0x0411, 0x12,
  0x0412, 0x21,
  0x0413, 0x13,
  0x0414, 0x47,
  0x0415, 0x90,
  0x0416, 0x20,
  0x0417, 0x00,
  0x0418, 0xD0,
  0x0419, 0x02,
  0x041A, 0x28,
  0x041B, 0x02,
  0x041C, 0xD0,
  0x041D, 0x02,
  0x041E, 0x2C,
  0x041F, 0x01,
  0x0420, 0x2C,
  0x0421, 0x01,
  0x0422, 0xE0,
  0x0423, 0x03,
  0x0424, 0xB2,
  0x0425, 0xB8,
  0x0426, 0xD0,
  0x0427, 0x01,
  0x0428, 0x0C,
  0x0429, 0x38,
  0x042A, 0xE7,
  0x042B, 0xEA,
  0x042C, 0x2C,
  0x042D, 0x01,
  0x042E, 0xE7,
  0x042F, 0xE8,
  0x0430, 0x21,
  0x0431, 0x00,
  0x0432, 0xD0,
  0x0433, 0x11,
  0x0434, 0x0A,
  0x0435, 0x02,
  0x0436, 0x9B,
  0x0437, 0x02,
  0x0438, 0x4B,
  0x0439, 0x0B,
  0x043A, 0x18,
  0x043B, 0x59,
  0x043C, 0x54,
  0x043D, 0x5A,
  0x043E, 0x1C,
  0x043F, 0x64,
  0x0440, 0x70,
  0x0441, 0x48,
  0x0442, 0x18,
  0x0443, 0xC9,
  0x0444, 0x69,
  0x0445, 0x00,
  0x0446, 0x48,
  0x0447, 0x05,
  0x0448, 0xDB,
  0x0449, 0xCC,
  0x044A, 0x42,
  0x044B, 0x84,
  0x044C, 0x68,
  0x044D, 0xC8,
  0x044E, 0x49,
  0x044F, 0x03,
  0x0450, 0x60,
  0x0451, 0xC8,
  0x0452, 0x1C,
  0x0453, 0x80,
  0x0454, 0x21,
  0x0455, 0x01,
  0x0456, 0xBD,
  0x0457, 0xFE,
  0x0458, 0xE7,
  0x0459, 0xEB,
  0x045A, 0x02,
  0x045B, 0x89,
  0x045C, 0x10,
  0x045D, 0x00,
  0x045E, 0x0E,
  0x045F, 0x80,
  0x0460, 0x10,
  0x0461, 0x00,
  0x0462, 0x0E,
  0x0463, 0x44,
  0x0464, 0x10,
  0x0465, 0x00,
  0x0466, 0x0E,
  0x0467, 0x54,
  0x0468, 0x10,
  0x0469, 0x00,
  0x046A, 0x0F,
  0x046B, 0x00,
  0x046C, 0x06,
  0x046D, 0x92,
  0x046E, 0x22,
  0x046F, 0x31,
  0x0470, 0x23,
  0x0471, 0x01,
  0x0472, 0x6A,
  0x0473, 0x51,
  0x0474, 0x0D,
  0x0475, 0x09,
  0x0476, 0x05,
  0x0477, 0x09,
  0x0478, 0x42,
  0x0479, 0x99,
  0x047A, 0x02,
  0x047B, 0x5B,
  0x047C, 0x18,
  0x047D, 0x08,
  0x047E, 0xD3,
  0x047F, 0x07,
  0x0480, 0x18,
  0x0481, 0x40,
  0x0482, 0x49,
  0x0483, 0x03,
  0x0484, 0x78,
  0x0485, 0x00,
  0x0486, 0x1A,
  0x0487, 0xC0,
  0x0488, 0x20,
  0x0489, 0x00,
  0x048A, 0x62,
  0x048B, 0xD0,
  0x048C, 0x47,
  0x048D, 0x70,
  0x048E, 0x63,
  0x048F, 0x50,
  0x0490, 0x10,
  0x0491, 0x00,
  0x0492, 0x0F,
  0x0493, 0x00,
  0x0494, 0x6B,
  0x0495, 0x08,
  0x0496, 0x49,
  0x0497, 0x0C,
  0x0498, 0x01,
  0x0499, 0x00,
  0x049A, 0x6B,
  0x049B, 0x49,
  0x049C, 0xD5,
  0x049D, 0x01,
  0x049E, 0x15,
  0x049F, 0x00,
  0x04A0, 0x1C,
  0x04A1, 0x40,
  0x04A2, 0x43,
  0x04A3, 0xC0,
  0x04A4, 0x05,
  0x04A5, 0x00,
  0x04A6, 0x4A,
  0x04A7, 0x09,
  0x04A8, 0x0D,
  0x04A9, 0x00,
  0x04AA, 0x68,
  0x04AB, 0x93,
  0x04AC, 0x60,
  0x04AD, 0x90,
  0x04AE, 0x18,
  0x04AF, 0x18,
  0x04B0, 0x15,
  0x04B1, 0x00,
  0x04B2, 0x01,
  0x04B3, 0x08,
  0x04B4, 0x43,
  0x04B5, 0xC0,
  0x04B6, 0xD5,
  0x04B7, 0x01,
  0x04B8, 0x05,
  0x04B9, 0x00,
  0x04BA, 0x1C,
  0x04BB, 0x40,
  0x04BC, 0x0D,
  0x04BD, 0x00,
  0x04BE, 0x68,
  0x04BF, 0xD1,
  0x04C0, 0x60,
  0x04C1, 0xD0,
  0x04C2, 0x18,
  0x04C3, 0x08,
  0x04C4, 0x00,
  0x04C5, 0x00,
  0x04C6, 0x47,
  0x04C7, 0x70,
  0x04C8, 0x40,
  0x04C9, 0x00,
  0x04CA, 0x00,
  0x04CB, 0x80,
  0x04CC, 0x10,
  0x04CD, 0x00,
  0x04CE, 0x0E,
  0x04CF, 0xD0,
  0x04D0, 0x20,
  0x04D1, 0x02,
  0x04D2, 0xB5,
  0x04D3, 0x10,
  0x04D4, 0xFF,
  0x04D5, 0xCA,
  0x04D6, 0xF7,
  0x04D7, 0xFF,
  0x04D8, 0xB5,
  0x04D9, 0x10,
  0x04DA, 0xBD,
  0x04DB, 0x10,
  0x04DC, 0xF7,
  0x04DD, 0xFF,
  0x04DE, 0x20,
  0x04DF, 0x01,
  0x04E0, 0xBD,
  0x04E1, 0x10,
  0x04E2, 0xFF,
  0x04E3, 0xC5,
  0x04E4, 0x20,
  0x04E5, 0x01,
  0x04E6, 0xB5,
  0x04E7, 0x10,
  0x04E8, 0xFF,
  0x04E9, 0xC0,
  0x04EA, 0xF7,
  0x04EB, 0xFF,
  0x04EC, 0xB5,
  0x04ED, 0x10,
  0x04EE, 0xBD,
  0x04EF, 0x10,
  0x04F0, 0xF7,
  0x04F1, 0xFF,
  0x04F2, 0x20,
  0x04F3, 0x00,
  0x04F4, 0xBD,
  0x04F5, 0x10,
  0x04F6, 0xFF,
  0x04F7, 0xBB,
  0x04F8, 0x49,
  0x04F9, 0x06,
  0x04FA, 0x48,
  0x04FB, 0x07,
  0x04FC, 0x49,
  0x04FD, 0x07,
  0x04FE, 0x60,
  0x04FF, 0x41,
  0x0500, 0x49,
  0x0501, 0x07,
  0x0502, 0x60,
  0x0503, 0x81,
  0x0504, 0x49,
  0x0505, 0x07,
  0x0506, 0x61,
  0x0507, 0x01,
  0x0508, 0x49,
  0x0509, 0x07,
  0x050A, 0x60,
  0x050B, 0xC1,
  0x050C, 0x49,
  0x050D, 0x07,
  0x050E, 0x61,
  0x050F, 0x41,
  0x0510, 0x47,
  0x0511, 0x70,
  0x0512, 0x61,
  0x0513, 0x81,
  0x0514, 0x10,
  0x0515, 0x00,
  0x0516, 0x05,
  0x0517, 0x31,
  0x0518, 0x10,
  0x0519, 0x00,
  0x051A, 0x1B,
  0x051B, 0x48,
  0x051C, 0x10,
  0x051D, 0x00,
  0x051E, 0x05,
  0x051F, 0x9D,
  0x0520, 0x10,
  0x0521, 0x00,
  0x0522, 0x04,
  0x0523, 0xEF,
  0x0524, 0x10,
  0x0525, 0x00,
  0x0526, 0x04,
  0x0527, 0xD1,
  0x0528, 0x10,
  0x0529, 0x00,
  0x052A, 0x04,
  0x052B, 0xDB,
  0x052C, 0x10,
  0x052D, 0x00,
  0x052E, 0x04,
  0x052F, 0xE5,
  0x0530, 0x48,
  0x0531, 0x13,
  0x0532, 0xB5,
  0x0533, 0x10,
  0x0534, 0x28,
  0x0535, 0x01,
  0x0536, 0x78,
  0x0537, 0x00,
  0x0538, 0x4C,
  0x0539, 0x12,
  0x053A, 0xD1,
  0x053B, 0x0B,
  0x053C, 0x68,
  0x053D, 0x61,
  0x053E, 0x68,
  0x053F, 0x20,
  0x0540, 0xD2,
  0x0541, 0x06,
  0x0542, 0x42,
  0x0543, 0x88,
  0x0544, 0xFB,
  0x0545, 0x9C,
  0x0546, 0xF0,
  0x0547, 0x00,
  0x0548, 0xFF,
  0x0549, 0xA4,
  0x054A, 0xF7,
  0x054B, 0xFF,
  0x054C, 0x1C,
  0x054D, 0x40,
  0x054E, 0x68,
  0x054F, 0x20,
  0x0550, 0x4C,
  0x0551, 0x0D,
  0x0552, 0x60,
  0x0553, 0x20,
  0x0554, 0x28,
  0x0555, 0x01,
  0x0556, 0x78,
  0x0557, 0x20,
  0x0558, 0xF7,
  0x0559, 0xFF,
  0x055A, 0xD1,
  0x055B, 0x10,
  0x055C, 0x49,
  0x055D, 0x0B,
  0x055E, 0xFF,
  0x055F, 0x33,
  0x0560, 0x42,
  0x0561, 0x88,
  0x0562, 0x68,
  0x0563, 0x09,
  0x0564, 0x20,
  0x0565, 0x00,
  0x0566, 0xD1,
  0x0567, 0x0A,
  0x0568, 0x48,
  0x0569, 0x09,
  0x056A, 0x70,
  0x056B, 0x20,
  0x056C, 0x47,
  0x056D, 0x80,
  0x056E, 0x68,
  0x056F, 0x00,
  0x0570, 0x4A,
  0x0571, 0x08,
  0x0572, 0x21,
  0x0573, 0x01,
  0x0574, 0x60,
  0x0575, 0x10,
  0x0576, 0x20,
  0x0577, 0x02,
  0x0578, 0x73,
  0x0579, 0x81,
  0x057A, 0x48,
  0x057B, 0x07,
  0x057C, 0x00,
  0x057D, 0x00,
  0x057E, 0xBD,
  0x057F, 0x10,
  0x0580, 0x10,
  0x0581, 0x00,
  0x0582, 0x0E,
  0x0583, 0xE0,
  0x0584, 0x10,
  0x0585, 0x00,
  0x0586, 0x0E,
  0x0587, 0x78,
  0x0588, 0x10,
  0x0589, 0x00,
  0x058A, 0x0E,
  0x058B, 0x80,
  0x058C, 0x10,
  0x058D, 0x00,
  0x058E, 0x0E,
  0x058F, 0x88,
  0x0590, 0x10,
  0x0591, 0x00,
  0x0592, 0x0E,
  0x0593, 0x60,
  0x0594, 0xE0,
  0x0595, 0x00,
  0x0596, 0xE1,
  0x0597, 0x80,
  0x0598, 0x10,
  0x0599, 0x00,
  0x059A, 0x1A,
  0x059B, 0x80,
  0x059C, 0x48,
  0x059D, 0x03,
  0x059E, 0xB5,
  0x059F, 0x10,
  0x05A0, 0x28,
  0x05A1, 0x01,
  0x05A2, 0x78,
  0x05A3, 0x00,
  0x05A4, 0xF0,
  0x05A5, 0x00,
  0x05A6, 0xD1,
  0x05A7, 0x01,
  0x05A8, 0xBD,
  0x05A9, 0x10,
  0x05AA, 0xFB,
  0x05AB, 0x19,
  0x05AC, 0x10,
  0x05AD, 0x00,
  0x05AE, 0x0E,
  0x05AF, 0xAC,
  0x05B0, 0x08,
  0x05B1, 0x9B,
  0x05B2, 0xB5,
  0x05B3, 0x7C,
  0x05B4, 0x4C,
  0x05B5, 0x17,
  0x05B6, 0x25,
  0x05B7, 0x00,
  0x05B8, 0xD0,
  0x05B9, 0x23,
  0x05BA, 0x2B,
  0x05BB, 0x01,
  0x05BC, 0xD0,
  0x05BD, 0x24,
  0x05BE, 0x2B,
  0x05BF, 0x02,
  0x05C0, 0x36,
  0x05C1, 0x01,
  0x05C2, 0x26,
  0x05C3, 0xFF,
  0x05C4, 0x26,
  0x05C5, 0x01,
  0x05C6, 0x80,
  0x05C7, 0x66,
  0x05C8, 0x80,
  0x05C9, 0xA5,
  0x05CA, 0x80,
  0x05CB, 0xE6,
  0x05CC, 0x61,
  0x05CD, 0x61,
  0x05CE, 0x81,
  0x05CF, 0x25,
  0x05D0, 0x62,
  0x05D1, 0x22,
  0x05D2, 0x61,
  0x05D3, 0xA2,
  0x05D4, 0x61,
  0x05D5, 0xE1,
  0x05D6, 0x46,
  0x05D7, 0x15,
  0x05D8, 0xD0,
  0x05D9, 0x19,
  0x05DA, 0x2B,
  0x05DB, 0x01,
  0x05DC, 0xD0,
  0x05DD, 0x17,
  0x05DE, 0x2B,
  0x05DF, 0x02,
  0x05E0, 0x02,
  0x05E1, 0x9B,
  0x05E2, 0x23,
  0x05E3, 0x01,
  0x05E4, 0x43,
  0x05E5, 0x43,
  0x05E6, 0x61,
  0x05E7, 0x20,
  0x05E8, 0x4A,
  0x05E9, 0x0B,
  0x05EA, 0x11,
  0x05EB, 0xD8,
  0x05EC, 0x68,
  0x05ED, 0x12,
  0x05EE, 0x18,
  0x05EF, 0x44,
  0x05F0, 0x46,
  0x05F1, 0x20,
  0x05F2, 0x46,
  0x05F3, 0x29,
  0x05F4, 0x49,
  0x05F5, 0x09,
  0x05F6, 0x47,
  0x05F7, 0x90,
  0x05F8, 0x94,
  0x05F9, 0x00,
  0x05FA, 0x95,
  0x05FB, 0x01,
  0x05FC, 0x46,
  0x05FD, 0x68,
  0x05FE, 0x68,
  0x05FF, 0x09,
  0x0600, 0xBD,
  0x0601, 0x7C,
  0x0602, 0x47,
  0x0603, 0x88,
  0x0604, 0x80,
  0x0605, 0xE5,
  0x0606, 0x80,
  0x0607, 0x65,
  0x0608, 0x80,
  0x0609, 0x65,
  0x060A, 0xE7,
  0x060B, 0xDF,
  0x060C, 0xE7,
  0x060D, 0xDB,
  0x060E, 0x26,
  0x060F, 0x02,
  0x0610, 0xE7,
  0x0611, 0xE7,
  0x0612, 0x23,
  0x0613, 0x00,
  0x0614, 0x10,
  0x0615, 0x00,
  0x0616, 0x0E,
  0x0617, 0xAC,
  0x0618, 0x10,
  0x0619, 0x00,
  0x061A, 0x0E,
  0x061B, 0x6C,
  0x061C, 0x10,
  0x061D, 0x00,
  0x061E, 0x0E,
  0x061F, 0x70,
  0x0620, 0x46,
  0x0621, 0x05,
  0x0622, 0xB5,
  0x0623, 0xF0,
  0x0624, 0xB0,
  0x0625, 0x8B,
  0x0626, 0x48,
  0x0627, 0x40,
  0x0628, 0x48,
  0x0629, 0x40,
  0x062A, 0x68,
  0x062B, 0x07,
  0x062C, 0x48,
  0x062D, 0x40,
  0x062E, 0x68,
  0x062F, 0x03,
  0x0630, 0x28,
  0x0631, 0x03,
  0x0632, 0x78,
  0x0633, 0x00,
  0x0634, 0x4C,
  0x0635, 0x3F,
  0x0636, 0xD1,
  0x0637, 0x64,
  0x0638, 0x28,
  0x0639, 0x00,
  0x063A, 0x68,
  0x063B, 0x60,
  0x063C, 0x48,
  0x063D, 0x3E,
  0x063E, 0xD1,
  0x063F, 0x16,
  0x0640, 0x1D,
  0x0641, 0x00,
  0x0642, 0x90,
  0x0643, 0x00,
  0x0644, 0x48,
  0x0645, 0x3D,
  0x0646, 0x90,
  0x0647, 0x01,
  0x0648, 0x1D,
  0x0649, 0x00,
  0x064A, 0x90,
  0x064B, 0x02,
  0x064C, 0x22,
  0x064D, 0x04,
  0x064E, 0x90,
  0x064F, 0x03,
  0x0650, 0x46,
  0x0651, 0x68,
  0x0652, 0xA9,
  0x0653, 0x04,
  0x0654, 0x48,
  0x0655, 0x3A,
  0x0656, 0x47,
  0x0657, 0x98,
  0x0658, 0x60,
  0x0659, 0x01,
  0x065A, 0x99,
  0x065B, 0x04,
  0x065C, 0x60,
  0x065D, 0x41,
  0x065E, 0x99,
  0x065F, 0x05,
  0x0660, 0x60,
  0x0661, 0x81,
  0x0662, 0x99,
  0x0663, 0x06,
  0x0664, 0x60,
  0x0665, 0xC1,
  0x0666, 0x99,
  0x0667, 0x07,
  0x0668, 0x60,
  0x0669, 0x60,
  0x066A, 0x20,
  0x066B, 0x01,
  0x066C, 0x07,
  0x066D, 0xA8,
  0x066E, 0x4C,
  0x066F, 0x35,
  0x0670, 0x7C,
  0x0671, 0x66,
  0x0672, 0xD0,
  0x0673, 0x48,
  0x0674, 0x90,
  0x0675, 0x0A,
  0x0676, 0x7C,
  0x0677, 0xA0,
  0x0678, 0xF7,
  0x0679, 0xFF,
  0x067A, 0x8A,
  0x067B, 0xA0,
  0x067C, 0x04,
  0x067D, 0x00,
  0x067E, 0xFE,
  0x067F, 0x45,
  0x0680, 0x90,
  0x0681, 0x09,
  0x0682, 0x14,
  0x0683, 0x00,
  0x0684, 0xF7,
  0x0685, 0xFF,
  0x0686, 0x8A,
  0x0687, 0xE0,
  0x0688, 0x04,
  0x0689, 0x00,
  0x068A, 0xFE,
  0x068B, 0x3F,
  0x068C, 0x90,
  0x068D, 0x08,
  0x068E, 0x14,
  0x068F, 0x00,
  0x0690, 0xD8,
  0x0691, 0x34,
  0x0692, 0x2E,
  0x0693, 0x0A,
  0x0694, 0xD0,
  0x0695, 0x32,
  0x0696, 0x2E,
  0x0697, 0x00,
  0x0698, 0x1E,
  0x0699, 0x76,
  0x069A, 0x98,
  0x069B, 0x0A,
  0x069C, 0xD8,
  0x069D, 0x2E,
  0x069E, 0x28,
  0x069F, 0x80,
  0x06A0, 0x99,
  0x06A1, 0x09,
  0x06A2, 0x20,
  0x06A3, 0x01,
  0x06A4, 0x18,
  0x06A5, 0x0A,
  0x06A6, 0x02,
  0x06A7, 0x80,
  0x06A8, 0x42,
  0x06A9, 0x8A,
  0x06AA, 0x49,
  0x06AB, 0x27,
  0x06AC, 0x9A,
  0x06AD, 0x08,
  0x06AE, 0xD2,
  0x06AF, 0x27,
  0x06B0, 0x42,
  0x06B1, 0x88,
  0x06B2, 0x18,
  0x06B3, 0x10,
  0x06B4, 0x06,
  0x06B5, 0x28,
  0x06B6, 0xD2,
  0x06B7, 0x23,
  0x06B8, 0x48,
  0x06B9, 0x1F,
  0x06BA, 0xD5,
  0x06BB, 0x11,
  0x06BC, 0x1D,
  0x06BD, 0x00,
  0x06BE, 0x90,
  0x06BF, 0x00,
  0x06C0, 0x48,
  0x06C1, 0x1E,
  0x06C2, 0x90,
  0x06C3, 0x01,
  0x06C4, 0x1D,
  0x06C5, 0x00,
  0x06C6, 0x90,
  0x06C7, 0x02,
  0x06C8, 0x48,
  0x06C9, 0x20,
  0x06CA, 0x90,
  0x06CB, 0x03,
  0x06CC, 0x02,
  0x06CD, 0xC1,
  0x06CE, 0x90,
  0x06CF, 0x04,
  0x06D0, 0x91,
  0x06D1, 0x07,
  0x06D2, 0x91,
  0x06D3, 0x05,
  0x06D4, 0x22,
  0x06D5, 0x04,
  0x06D6, 0x90,
  0x06D7, 0x06,
  0x06D8, 0x46,
  0x06D9, 0x68,
  0x06DA, 0xA9,
  0x06DB, 0x04,
  0x06DC, 0x46,
  0x06DD, 0x2B,
  0x06DE, 0x47,
  0x06DF, 0xB8,
  0x06E0, 0x40,
  0x06E1, 0x03,
  0x06E2, 0x20,
  0x06E3, 0x0C,
  0x06E4, 0x99,
  0x06E5, 0x09,
  0x06E6, 0x9A,
  0x06E7, 0x08,
  0x06E8, 0xF7,
  0x06E9, 0xFF,
  0x06EA, 0x98,
  0x06EB, 0x0A,
  0x06EC, 0x49,
  0x06ED, 0x18,
  0x06EE, 0xFF,
  0x06EF, 0x61,
  0x06F0, 0x68,
  0x06F1, 0x09,
  0x06F2, 0x20,
  0x06F3, 0xFF,
  0x06F4, 0x47,
  0x06F5, 0x88,
  0x06F6, 0x30,
  0x06F7, 0x2D,
  0x06F8, 0xF7,
  0x06F9, 0xFF,
  0x06FA, 0x46,
  0x06FB, 0x30,
  0x06FC, 0x74,
  0x06FD, 0x25,
  0x06FE, 0xFD,
  0x06FF, 0xC3,
  0x0700, 0xBD,
  0x0701, 0xF0,
  0x0702, 0xB0,
  0x0703, 0x0B,
  0x0704, 0xFD,
  0x0705, 0xDD,
  0x0706, 0xF7,
  0x0707, 0xFF,
  0x0708, 0x06,
  0x0709, 0x00,
  0x070A, 0x7C,
  0x070B, 0x20,
  0x070C, 0x48,
  0x070D, 0x0A,
  0x070E, 0xD5,
  0x070F, 0xF7,
  0x0710, 0x1D,
  0x0711, 0x00,
  0x0712, 0x90,
  0x0713, 0x00,
  0x0714, 0x48,
  0x0715, 0x09,
  0x0716, 0x90,
  0x0717, 0x01,
  0x0718, 0x1D,
  0x0719, 0x00,
  0x071A, 0x90,
  0x071B, 0x02,
  0x071C, 0x22,
  0x071D, 0x04,
  0x071E, 0x90,
  0x071F, 0x03,
  0x0720, 0x46,
  0x0721, 0x68,
  0x0722, 0x49,
  0x0723, 0x07,
  0x0724, 0xE7,
  0x0725, 0xEA,
  0x0726, 0x47,
  0x0727, 0xB8,
  0x0728, 0x10,
  0x0729, 0x00,
  0x072A, 0x0E,
  0x072B, 0x40,
  0x072C, 0x10,
  0x072D, 0x00,
  0x072E, 0x0E,
  0x072F, 0x44,
  0x0730, 0x10,
  0x0731, 0x00,
  0x0732, 0x19,
  0x0733, 0x60,
  0x0734, 0x10,
  0x0735, 0x00,
  0x0736, 0x0E,
  0x0737, 0x80,
  0x0738, 0x00,
  0x0739, 0x00,
  0x073A, 0x42,
  0x073B, 0x4C,
  0x073C, 0x00,
  0x073D, 0x00,
  0x073E, 0x52,
  0x073F, 0x4C,
  0x0740, 0x10,
  0x0741, 0x00,
  0x0742, 0x0E,
  0x0743, 0xF0,
  0x0744, 0x10,
  0x0745, 0x00,
  0x0746, 0x1A,
  0x0747, 0x60,
  0x0748, 0x00,
  0x0749, 0x00,
  0x074A, 0x08,
  0x074B, 0x01,
  0x074C, 0x07,
  0x074D, 0xFF,
  0x074E, 0x00,
  0x074F, 0x00,
  0x0750, 0x10,
  0x0751, 0x00,
  0x0752, 0x0E,
  0x0753, 0x58,
  0x0754, 0x46,
  0x0755, 0x04,
  0x0756, 0xB5,
  0x0757, 0xF8,
  0x0758, 0x40,
  0x0759, 0x08,
  0x075A, 0x21,
  0x075B, 0xF0,
  0x075C, 0x09,
  0x075D, 0x01,
  0x075E, 0x4D,
  0x075F, 0x1D,
  0x0760, 0x61,
  0x0761, 0x69,
  0x0762, 0x4E,
  0x0763, 0x1D,
  0x0764, 0x29,
  0x0765, 0x01,
  0x0766, 0xD0,
  0x0767, 0x20,
  0x0768, 0x29,
  0x0769, 0x02,
  0x076A, 0xD0,
  0x076B, 0x23,
  0x076C, 0x46,
  0x076D, 0x28,
  0x076E, 0xD1,
  0x076F, 0x1B,
  0x0770, 0x30,
  0x0771, 0x24,
  0x0772, 0x49,
  0x0773, 0x1A,
  0x0774, 0x49,
  0x0775, 0x1A,
  0x0776, 0x60,
  0x0777, 0x01,
  0x0778, 0x07,
  0x0779, 0x20,
  0x077A, 0x60,
  0x077B, 0x41,
  0x077C, 0x22,
  0x077D, 0x02,
  0x077E, 0x0F,
  0x077F, 0x00,
  0x0780, 0x61,
  0x0781, 0xA8,
  0x0782, 0x21,
  0x0783, 0x01,
  0x0784, 0xD0,
  0x0785, 0x20,
  0x0786, 0x28,
  0x0787, 0x02,
  0x0788, 0x02,
  0x0789, 0xC8,
  0x078A, 0xD2,
  0x078B, 0x0D,
  0x078C, 0x60,
  0x078D, 0xA8,
  0x078E, 0x61,
  0x078F, 0x29,
  0x0790, 0x20,
  0x0791, 0x00,
  0x0792, 0x4B,
  0x0793, 0x14,
  0x0794, 0x60,
  0x0795, 0xE8,
  0x0796, 0x73,
  0x0797, 0x98,
  0x0798, 0x48,
  0x0799, 0x13,
  0x079A, 0x70,
  0x079B, 0x29,
  0x079C, 0x49,
  0x079D, 0x14,
  0x079E, 0x60,
  0x079F, 0x02,
  0x07A0, 0x68,
  0x07A1, 0x09,
  0x07A2, 0x48,
  0x07A3, 0x12,
  0x07A4, 0xBD,
  0x07A5, 0xF8,
  0x07A6, 0x47,
  0x07A7, 0x88,
  0x07A8, 0x60,
  0x07A9, 0x30,
  0x07AA, 0x48,
  0x07AB, 0x12,
  0x07AC, 0x60,
  0x07AD, 0x70,
  0x07AE, 0x1D,
  0x07AF, 0x00,
  0x07B0, 0x4F,
  0x07B1, 0x11,
  0x07B2, 0xE7,
  0x07B3, 0xE3,
  0x07B4, 0x60,
  0x07B5, 0x37,
  0x07B6, 0x4A,
  0x07B7, 0x11,
  0x07B8, 0x68,
  0x07B9, 0x12,
  0x07BA, 0x21,
  0x07BB, 0x13,
  0x07BC, 0x47,
  0x07BD, 0x90,
  0x07BE, 0x20,
  0x07BF, 0x00,
  0x07C0, 0xD0,
  0x07C1, 0x00,
  0x07C2, 0x28,
  0x07C3, 0x02,
  0x07C4, 0x60,
  0x07C5, 0x77,
  0x07C6, 0x1D,
  0x07C7, 0x3F,
  0x07C8, 0x20,
  0x07C9, 0x01,
  0x07CA, 0xE7,
  0x07CB, 0xD7,
  0x07CC, 0x61,
  0x07CD, 0x2A,
  0x07CE, 0x02,
  0x07CF, 0x80,
  0x07D0, 0xE7,
  0x07D1, 0xDD,
  0x07D2, 0x60,
  0x07D3, 0xA8,
  0x07D4, 0x10,
  0x07D5, 0x00,
  0x07D6, 0x0E,
  0x07D7, 0x80,
  0x07D8, 0x10,
  0x07D9, 0x00,
  0x07DA, 0x0E,
  0x07DB, 0x9C,
  0x07DC, 0x00,
  0x07DD, 0x00,
  0x07DE, 0x40,
  0x07DF, 0x54,
  0x07E0, 0x00,
  0x07E1, 0x00,
  0x07E2, 0x50,
  0x07E3, 0x54,
  0x07E4, 0x10,
  0x07E5, 0x00,
  0x07E6, 0x1A,
  0x07E7, 0x80,
  0x07E8, 0xE0,
  0x07E9, 0x00,
  0x07EA, 0xE1,
  0x07EB, 0x00,
  0x07EC, 0x00,
  0x07ED, 0x00,
  0x07EE, 0x06,
  0x07EF, 0x97,
  0x07F0, 0x10,
  0x07F1, 0x00,
  0x07F2, 0x0E,
  0x07F3, 0x5C,
  0x07F4, 0xC8,
  0x07F5, 0x00,
  0x07F6, 0x00,
  0x07F7, 0x60,
  0x07F8, 0x10,
  0x07F9, 0x00,
  0x07FA, 0x17,
  0x07FB, 0x60,
  0x07FC, 0x10,
  0x07FD, 0x00,
  0x07FE, 0x0E,
  0x07FF, 0x54,
  0x0800, 0xB2,
  0x0801, 0xC0,
  0x0802, 0xB5,
  0x0803, 0x10,
  0x0804, 0xF9,
  0x0805, 0x7E,
  0x0806, 0xF0,
  0x0807, 0x00,
  0x0808, 0x00,
  0x0809, 0x00,
  0x080A, 0xBD,
  0x080B, 0x10,
  0x080C, 0x49,
  0x080D, 0xB0,
  0x080E, 0xB5,
  0x080F, 0xF0,
  0x0810, 0x68,
  0x0811, 0x0C,
  0x0812, 0xB0,
  0x0813, 0x89,
  0x0814, 0xD8,
  0x0815, 0x7D,
  0x0816, 0x28,
  0x0817, 0x01,
  0x0818, 0x4D,
  0x0819, 0xAE,
  0x081A, 0x26,
  0x081B, 0x03,
  0x081C, 0x28,
  0x081D, 0x00,
  0x081E, 0x03,
  0x081F, 0x76,
  0x0820, 0x20,
  0x0821, 0x00,
  0x0822, 0xD0,
  0x0823, 0x5E,
  0x0824, 0x90,
  0x0825, 0x03,
  0x0826, 0x90,
  0x0827, 0x04,
  0x0828, 0x90,
  0x0829, 0x01,
  0x082A, 0x90,
  0x082B, 0x02,
  0x082C, 0x21,
  0x082D, 0x01,
  0x082E, 0x22,
  0x082F, 0x04,
  0x0830, 0xF7,
  0x0831, 0xFF,
  0x0832, 0xA8,
  0x0833, 0x01,
  0x0834, 0x22,
  0x0835, 0x04,
  0x0836, 0xFD,
  0x0837, 0xAD,
  0x0838, 0xA8,
  0x0839, 0x01,
  0x083A, 0x21,
  0x083B, 0x05,
  0x083C, 0xFD,
  0x083D, 0xA8,
  0x083E, 0xF7,
  0x083F, 0xFF,
  0x0840, 0x21,
  0x0841, 0x09,
  0x0842, 0x22,
  0x0843, 0x02,
  0x0844, 0xF7,
  0x0845, 0xFF,
  0x0846, 0xA8,
  0x0847, 0x01,
  0x0848, 0x22,
  0x0849, 0x04,
  0x084A, 0xFD,
  0x084B, 0xA3,
  0x084C, 0xA8,
  0x084D, 0x01,
  0x084E, 0x21,
  0x084F, 0x0C,
  0x0850, 0xFD,
  0x0851, 0x9E,
  0x0852, 0xF7,
  0x0853, 0xFF,
  0x0854, 0x21,
  0x0855, 0x11,
  0x0856, 0x22,
  0x0857, 0x04,
  0x0858, 0xF7,
  0x0859, 0xFF,
  0x085A, 0xA8,
  0x085B, 0x01,
  0x085C, 0x22,
  0x085D, 0x02,
  0x085E, 0xFD,
  0x085F, 0x99,
  0x0860, 0xA8,
  0x0861, 0x01,
  0x0862, 0x21,
  0x0863, 0x16,
  0x0864, 0xFD,
  0x0865, 0x94,
  0x0866, 0xF7,
  0x0867, 0xFF,
  0x0868, 0x21,
  0x0869, 0x36,
  0x086A, 0x22,
  0x086B, 0x02,
  0x086C, 0xF7,
  0x086D, 0xFF,
  0x086E, 0xA8,
  0x086F, 0x01,
  0x0870, 0x48,
  0x0871, 0x99,
  0x0872, 0xFD,
  0x0873, 0x8F,
  0x0874, 0x90,
  0x0875, 0x02,
  0x0876, 0x90,
  0x0877, 0x03,
  0x0878, 0x48,
  0x0879, 0x98,
  0x087A, 0x90,
  0x087B, 0x01,
  0x087C, 0x48,
  0x087D, 0x98,
  0x087E, 0x90,
  0x087F, 0x04,
  0x0880, 0x48,
  0x0881, 0x97,
  0x0882, 0x90,
  0x0883, 0x06,
  0x0884, 0x30,
  0x0885, 0x14,
  0x0886, 0x96,
  0x0887, 0x05,
  0x0888, 0x48,
  0x0889, 0x95,
  0x088A, 0x90,
  0x088B, 0x07,
  0x088C, 0x30,
  0x088D, 0xA8,
  0x088E, 0xA9,
  0x088F, 0x01,
  0x0890, 0x22,
  0x0891, 0x04,
  0x0892, 0x90,
  0x0893, 0x08,
  0x0894, 0x47,
  0x0895, 0xA0,
  0x0896, 0xA8,
  0x0897, 0x05,
  0x0898, 0x21,
  0x0899, 0x85,
  0x089A, 0x4F,
  0x089B, 0x92,
  0x089C, 0x20,
  0x089D, 0x04,
  0x089E, 0x68,
  0x089F, 0x3A,
  0x08A0, 0x06,
  0x08A1, 0x00,
  0x08A2, 0x47,
  0x08A3, 0x90,
  0x08A4, 0x00,
  0x08A5, 0x76,
  0x08A6, 0xD5,
  0x08A7, 0x01,
  0x08A8, 0x26,
  0x08A9, 0x01,
  0x08AA, 0xE0,
  0x08AB, 0x01,
  0x08AC, 0x21,
  0x08AD, 0x87,
  0x08AE, 0x03,
  0x08AF, 0xB6,
  0x08B0, 0x20,
  0x08B1, 0x04,
  0x08B2, 0x68,
  0x08B3, 0x3A,
  0x08B4, 0x06,
  0x08B5, 0x00,
  0x08B6, 0x47,
  0x08B7, 0x90,
  0x08B8, 0x20,
  0x08B9, 0x03,
  0x08BA, 0xD5,
  0x08BB, 0x02,
  0x08BC, 0xE0,
  0x08BD, 0x01,
  0x08BE, 0x07,
  0x08BF, 0x80,
  0x08C0, 0x07,
  0x08C1, 0x80,
  0x08C2, 0x20,
  0x08C3, 0x01,
  0x08C4, 0x96,
  0x08C5, 0x01,
  0x08C6, 0x43,
  0x08C7, 0x06,
  0x08C8, 0x20,
  0x08C9, 0xA2,
  0x08CA, 0x68,
  0x08CB, 0x29,
  0x08CC, 0x90,
  0x08CD, 0x02,
  0x08CE, 0x47,
  0x08CF, 0x88,
  0x08D0, 0x30,
  0x08D1, 0x28,
  0x08D2, 0x48,
  0x08D3, 0x83,
  0x08D4, 0x48,
  0x08D5, 0x82,
  0x08D6, 0x90,
  0x08D7, 0x05,
  0x08D8, 0x22,
  0x08D9, 0x02,
  0x08DA, 0x30,
  0x08DB, 0xA4,
  0x08DC, 0xE0,
  0x08DD, 0xF1,
  0x08DE, 0x90,
  0x08DF, 0x06,
  0x08E0, 0xAF,
  0x08E1, 0x05,
  0x08E2, 0x20,
  0x08E3, 0x00,
  0x08E4, 0x46,
  0x08E5, 0x0A,
  0x08E6, 0x00,
  0x08E7, 0x81,
  0x08E8, 0x1C,
  0x08E9, 0x40,
  0x08EA, 0x43,
  0x08EB, 0x31,
  0x08EC, 0x28,
  0x08ED, 0x03,
  0x08EE, 0x50,
  0x08EF, 0xB9,
  0x08F0, 0x4D,
  0x08F1, 0x78,
  0x08F2, 0xD3,
  0x08F3, 0xF8,
  0x08F4, 0x68,
  0x08F5, 0x29,
  0x08F6, 0x20,
  0x08F7, 0x6A,
  0x08F8, 0x90,
  0x08F9, 0x01,
  0x08FA, 0x47,
  0x08FB, 0x88,
  0x08FC, 0x20,
  0x08FD, 0x68,
  0x08FE, 0x68,
  0x08FF, 0x29,
  0x0900, 0x90,
  0x0901, 0x02,
  0x0902, 0x47,
  0x0903, 0x88,
  0x0904, 0x20,
  0x0905, 0x6E,
  0x0906, 0x68,
  0x0907, 0x29,
  0x0908, 0x90,
  0x0909, 0x03,
  0x090A, 0x47,
  0x090B, 0x88,
  0x090C, 0xA9,
  0x090D, 0x01,
  0x090E, 0x22,
  0x090F, 0x03,
  0x0910, 0xE0,
  0x0911, 0x00,
  0x0912, 0xA8,
  0x0913, 0x05,
  0x0914, 0x47,
  0x0915, 0xA0,
  0x0916, 0xE0,
  0x0917, 0xD9,
  0x0918, 0x4D,
  0x0919, 0x73,
  0x091A, 0x48,
  0x091B, 0x71,
  0x091C, 0x90,
  0x091D, 0x05,
  0x091E, 0x38,
  0x091F, 0x20,
  0x0920, 0x90,
  0x0921, 0x06,
  0x0922, 0x1D,
  0x0923, 0x00,
  0x0924, 0x95,
  0x0925, 0x01,
  0x0926, 0x48,
  0x0927, 0x6E,
  0x0928, 0x90,
  0x0929, 0x07,
  0x092A, 0x38,
  0x092B, 0x14,
  0x092C, 0xA9,
  0x092D, 0x01,
  0x092E, 0x48,
  0x092F, 0x6C,
  0x0930, 0x90,
  0x0931, 0x08,
  0x0932, 0x38,
  0x0933, 0x0C,
  0x0934, 0x90,
  0x0935, 0x02,
  0x0936, 0x48,
  0x0937, 0x6D,
  0x0938, 0x90,
  0x0939, 0x03,
  0x093A, 0x48,
  0x093B, 0x6D,
  0x093C, 0x95,
  0x093D, 0x04,
  0x093E, 0x22,
  0x093F, 0x04,
  0x0940, 0x47,
  0x0941, 0xA0,
  0x0942, 0xA8,
  0x0943, 0x05,
  0x0944, 0xA9,
  0x0945, 0x01,
  0x0946, 0x48,
  0x0947, 0x66,
  0x0948, 0x90,
  0x0949, 0x05,
  0x094A, 0x38,
  0x094B, 0x18,
  0x094C, 0x95,
  0x094D, 0x04,
  0x094E, 0x48,
  0x094F, 0x64,
  0x0950, 0x90,
  0x0951, 0x06,
  0x0952, 0x38,
  0x0953, 0x10,
  0x0954, 0x38,
  0x0955, 0x08,
  0x0956, 0x48,
  0x0957, 0x62,
  0x0958, 0x1D,
  0x0959, 0x00,
  0x095A, 0x90,
  0x095B, 0x07,
  0x095C, 0x20,
  0x095D, 0x00,
  0x095E, 0x90,
  0x095F, 0x08,
  0x0960, 0x90,
  0x0961, 0x02,
  0x0962, 0x90,
  0x0963, 0x01,
  0x0964, 0x22,
  0x0965, 0x04,
  0x0966, 0x90,
  0x0967, 0x03,
  0x0968, 0x47,
  0x0969, 0xA0,
  0x096A, 0xA8,
  0x096B, 0x05,
  0x096C, 0x20,
  0x096D, 0x00,
  0x096E, 0x25,
  0x096F, 0x00,
  0x0970, 0x18,
  0x0971, 0x19,
  0x0972, 0x00,
  0x0973, 0xAB,
  0x0974, 0x22,
  0x0975, 0x03,
  0x0976, 0x00,
  0x0977, 0x89,
  0x0978, 0x03,
  0x0979, 0x52,
  0x097A, 0x31,
  0x097B, 0x2C,
  0x097C, 0x00,
  0x097D, 0x86,
  0x097E, 0x43,
  0x097F, 0x11,
  0x0980, 0x1C,
  0x0981, 0x40,
  0x0982, 0xAA,
  0x0983, 0x05,
  0x0984, 0x28,
  0x0985, 0x04,
  0x0986, 0x51,
  0x0987, 0x91,
  0x0988, 0x4F,
  0x0989, 0x52,
  0x098A, 0xD3,
  0x098B, 0xF3,
  0x098C, 0x46,
  0x098D, 0x30,
  0x098E, 0x00,
  0x098F, 0xEE,
  0x0990, 0x30,
  0x0991, 0x6C,
  0x0992, 0x68,
  0x0993, 0x39,
  0x0994, 0x90,
  0x0995, 0x01,
  0x0996, 0x47,
  0x0997, 0x88,
  0x0998, 0x68,
  0x0999, 0x39,
  0x099A, 0x46,
  0x099B, 0x30,
  0x099C, 0x47,
  0x099D, 0x88,
  0x099E, 0x30,
  0x099F, 0x72,
  0x09A0, 0x46,
  0x09A1, 0x30,
  0x09A2, 0x90,
  0x09A3, 0x02,
  0x09A4, 0x30,
  0x09A5, 0x70,
  0x09A6, 0x68,
  0x09A7, 0x39,
  0x09A8, 0x90,
  0x09A9, 0x03,
  0x09AA, 0x47,
  0x09AB, 0x88,
  0x09AC, 0x68,
  0x09AD, 0x39,
  0x09AE, 0x46,
  0x09AF, 0x30,
  0x09B0, 0x47,
  0x09B1, 0x88,
  0x09B2, 0x30,
  0x09B3, 0x76,
  0x09B4, 0x22,
  0x09B5, 0x04,
  0x09B6, 0x90,
  0x09B7, 0x04,
  0x09B8, 0xA8,
  0x09B9, 0x05,
  0x09BA, 0xA9,
  0x09BB, 0x01,
  0x09BC, 0x1C,
  0x09BD, 0x6D,
  0x09BE, 0x47,
  0x09BF, 0xA0,
  0x09C0, 0xD3,
  0x09C1, 0xD4,
  0x09C2, 0x2D,
  0x09C3, 0x02,
  0x09C4, 0x26,
  0x09C5, 0x03,
  0x09C6, 0x20,
  0x09C7, 0x00,
  0x09C8, 0x03,
  0x09C9, 0x76,
  0x09CA, 0xAD,
  0x09CB, 0x05,
  0x09CC, 0x46,
  0x09CD, 0x0A,
  0x09CE, 0x00,
  0x09CF, 0x81,
  0x09D0, 0x43,
  0x09D1, 0x32,
  0x09D2, 0x32,
  0x09D3, 0x4C,
  0x09D4, 0x50,
  0x09D5, 0x6A,
  0x09D6, 0x1C,
  0x09D7, 0x40,
  0x09D8, 0xD3,
  0x09D9, 0xF7,
  0x09DA, 0x28,
  0x09DB, 0x04,
  0x09DC, 0x20,
  0x09DD, 0x7C,
  0x09DE, 0x4F,
  0x09DF, 0x3D,
  0x09E0, 0x47,
  0x09E1, 0x88,
  0x09E2, 0x68,
  0x09E3, 0x39,
  0x09E4, 0x68,
  0x09E5, 0x39,
  0x09E6, 0x90,
  0x09E7, 0x01,
  0x09E8, 0x47,
  0x09E9, 0x88,
  0x09EA, 0x20,
  0x09EB, 0x82,
  0x09EC, 0x90,
  0x09ED, 0x02,
  0x09EE, 0x4A,
  0x09EF, 0x41,
  0x09F0, 0x68,
  0x09F1, 0x12,
  0x09F2, 0x21,
  0x09F3, 0x84,
  0x09F4, 0x47,
  0x09F5, 0x90,
  0x09F6, 0x20,
  0x09F7, 0x04,
  0x09F8, 0x90,
  0x09F9, 0x03,
  0x09FA, 0x4A,
  0x09FB, 0x3E,
  0x09FC, 0x68,
  0x09FD, 0x12,
  0x09FE, 0x21,
  0x09FF, 0x88,
  0x0A00, 0x47,
  0x0A01, 0x90,
  0x0A02, 0x20,
  0x0A03, 0x04,
  0x0A04, 0x22,
  0x0A05, 0x04,
  0x0A06, 0x90,
  0x0A07, 0x04,
  0x0A08, 0xA8,
  0x0A09, 0x05,
  0x0A0A, 0xA9,
  0x0A0B, 0x01,
  0x0A0C, 0x48,
  0x0A0D, 0x34,
  0x0A0E, 0x47,
  0x0A0F, 0xA0,
  0x0A10, 0x90,
  0x0A11, 0x05,
  0x0A12, 0x30,
  0x0A13, 0x30,
  0x0A14, 0x20,
  0x0A15, 0x80,
  0x0A16, 0x68,
  0x0A17, 0x39,
  0x0A18, 0x90,
  0x0A19, 0x01,
  0x0A1A, 0x47,
  0x0A1B, 0x88,
  0x0A1C, 0xA9,
  0x0A1D, 0x01,
  0x0A1E, 0x22,
  0x0A1F, 0x01,
  0x0A20, 0x47,
  0x0A21, 0xA0,
  0x0A22, 0xA8,
  0x0A23, 0x05,
  0x0A24, 0x00,
  0x0A25, 0x8A,
  0x0A26, 0x21,
  0x0A27, 0x00,
  0x0A28, 0x30,
  0x0A29, 0xD0,
  0x0A2A, 0x46,
  0x0A2B, 0x10,
  0x0A2C, 0x1C,
  0x0A2D, 0x49,
  0x0A2E, 0x43,
  0x0A2F, 0x30,
  0x0A30, 0x29,
  0x0A31, 0x04,
  0x0A32, 0x50,
  0x0A33, 0xA8,
  0x0A34, 0x22,
  0x0A35, 0x02,
  0x0A36, 0xD3,
  0x0A37, 0xF7,
  0x0A38, 0x4D,
  0x0A39, 0x2F,
  0x0A3A, 0x1D,
  0x0A3B, 0x28,
  0x0A3C, 0x68,
  0x0A3D, 0x2B,
  0x0A3E, 0xA9,
  0x0A3F, 0x02,
  0x0A40, 0x68,
  0x0A41, 0x2B,
  0x0A42, 0x47,
  0x0A43, 0x98,
  0x0A44, 0xA9,
  0x0A45, 0x04,
  0x0A46, 0x22,
  0x0A47, 0x01,
  0x0A48, 0x47,
  0x0A49, 0x98,
  0x0A4A, 0xA8,
  0x0A4B, 0x08,
  0x0A4C, 0x20,
  0x0A4D, 0x0E,
  0x0A4E, 0x4D,
  0x0A4F, 0x2B,
  0x0A50, 0x46,
  0x0A51, 0x06,
  0x0A52, 0x47,
  0x0A53, 0xA8,
  0x0A54, 0x47,
  0x0A55, 0xA8,
  0x0A56, 0x20,
  0x0A57, 0x0C,
  0x0A58, 0x43,
  0x0A59, 0x06,
  0x0A5A, 0x04,
  0x0A5B, 0x00,
  0x0A5C, 0x48,
  0x0A5D, 0x28,
  0x0A5E, 0x96,
  0x0A5F, 0x01,
  0x0A60, 0x4D,
  0x0A61, 0x20,
  0x0A62, 0x9E,
  0x0A63, 0x02,
  0x0A64, 0x21,
  0x0A65, 0x03,
  0x0A66, 0x40,
  0x0A67, 0x06,
  0x0A68, 0x20,
  0x0A69, 0x04,
  0x0A6A, 0x68,
  0x0A6B, 0x2A,
  0x0A6C, 0x46,
  0x0A6D, 0x07,
  0x0A6E, 0x47,
  0x0A6F, 0x90,
  0x0A70, 0x68,
  0x0A71, 0x2A,
  0x0A72, 0x21,
  0x0A73, 0x02,
  0x0A74, 0x47,
  0x0A75, 0x90,
  0x0A76, 0x20,
  0x0A77, 0x04,
  0x0A78, 0x43,
  0x0A79, 0x07,
  0x0A7A, 0x04,
  0x0A7B, 0x00,
  0x0A7C, 0x48,
  0x0A7D, 0x20,
  0x0A7E, 0x43,
  0x0A7F, 0x37,
  0x0A80, 0x97,
  0x0A81, 0x02,
  0x0A82, 0x9E,
  0x0A83, 0x03,
  0x0A84, 0x21,
  0x0A85, 0x07,
  0x0A86, 0x40,
  0x0A87, 0x06,
  0x0A88, 0x20,
  0x0A89, 0x04,
  0x0A8A, 0x68,
  0x0A8B, 0x2A,
  0x0A8C, 0x46,
  0x0A8D, 0x07,
  0x0A8E, 0x47,
  0x0A8F, 0x90,
  0x0A90, 0x68,
  0x0A91, 0x2A,
  0x0A92, 0x21,
  0x0A93, 0x06,
  0x0A94, 0x47,
  0x0A95, 0x90,
  0x0A96, 0x20,
  0x0A97, 0x04,
  0x0A98, 0x43,
  0x0A99, 0x07,
  0x0A9A, 0x04,
  0x0A9B, 0x00,
  0x0A9C, 0x48,
  0x0A9D, 0x18,
  0x0A9E, 0x43,
  0x0A9F, 0x37,
  0x0AA0, 0x97,
  0x0AA1, 0x03,
  0x0AA2, 0x9E,
  0x0AA3, 0x04,
  0x0AA4, 0x40,
  0x0AA5, 0x06,
  0x0AA6, 0x43,
  0x0AA7, 0xC0,
  0x0AA8, 0x68,
  0x0AA9, 0x2A,
  0x0AAA, 0x21,
  0x0AAB, 0x01,
  0x0AAC, 0x47,
  0x0AAD, 0x90,
  0x0AAE, 0x20,
  0x0AAF, 0x04,
  0x0AB0, 0x21,
  0x0AB1, 0x00,
  0x0AB2, 0x02,
  0x0AB3, 0x07,
  0x0AB4, 0x20,
  0x0AB5, 0x04,
  0x0AB6, 0x68,
  0x0AB7, 0x2A,
  0x0AB8, 0x06,
  0x0AB9, 0x00,
  0x0ABA, 0x47,
  0x0ABB, 0x90,
  0x0ABC, 0x43,
  0x0ABD, 0x37,
  0x0ABE, 0x43,
  0x0ABF, 0x07,
  0x0AC0, 0x97,
  0x0AC1, 0x04,
  0x0AC2, 0x22,
  0x0AC3, 0x04,
  0x0AC4, 0xA8,
  0x0AC5, 0x05,
  0x0AC6, 0xA9,
  0x0AC7, 0x01,
  0x0AC8, 0xB0,
  0x0AC9, 0x09,
  0x0ACA, 0x47,
  0x0ACB, 0xA0,
  0x0ACC, 0x00,
  0x0ACD, 0x00,
  0x0ACE, 0xBD,
  0x0ACF, 0xF0,
  0x0AD0, 0x10,
  0x0AD1, 0x00,
  0x0AD2, 0x0E,
  0x0AD3, 0x40,
  0x0AD4, 0x10,
  0x0AD5, 0x00,
  0x0AD6, 0x0E,
  0x0AD7, 0x48,
  0x0AD8, 0x7F,
  0x0AD9, 0xFF,
  0x0ADA, 0x7F,
  0x0ADB, 0xFF,
  0x0ADC, 0x00,
  0x0ADD, 0x80,
  0x0ADE, 0x00,
  0x0ADF, 0x80,
  0x0AE0, 0x00,
  0x0AE1, 0x00,
  0x0AE2, 0x60,
  0x0AE3, 0x2C,
  0x0AE4, 0x10,
  0x0AE5, 0x00,
  0x0AE6, 0x0E,
  0x0AE7, 0x54,
  0x0AE8, 0x40,
  0x0AE9, 0x00,
  0x0AEA, 0x40,
  0x0AEB, 0x00,
  0x0AEC, 0xC0,
  0x0AED, 0x00,
  0x0AEE, 0xC0,
  0x0AEF, 0x00,
  0x0AF0, 0x3C,
  0x0AF1, 0x00,
  0x0AF2, 0x3C,
  0x0AF3, 0x00,
  0x0AF4, 0x10,
  0x0AF5, 0x00,
  0x0AF6, 0x0E,
  0x0AF7, 0x4C,
  0x0AF8, 0x10,
  0x0AF9, 0x00,
  0x0AFA, 0x0E,
  0x0AFB, 0x44,
  0x0AFC, 0x00,
  0x0AFD, 0x00,
  0x0AFE, 0x07,
  0x0AFF, 0x7B,
  0x0B00, 0xFF,
  0x0B01, 0x00,
  0x0B02, 0xFF,
  0x0B03, 0x00,
  0x0B04, 0x28,
  0x0B05, 0x01,
  0x0B06, 0xB5,
  0x0B07, 0xF8,
  0x0B08, 0x28,
  0x0B09, 0x02,
  0x0B0A, 0xD0,
  0x0B0B, 0x10,
  0x0B0C, 0x25,
  0x0B0D, 0x01,
  0x0B0E, 0xD1,
  0x0B0F, 0x54,
  0x0B10, 0x21,
  0x0B11, 0x00,
  0x0B12, 0x4C,
  0x0B13, 0x2C,
  0x0B14, 0x46,
  0x0B15, 0x26,
  0x0B16, 0x77,
  0x0B17, 0xA1,
  0x0B18, 0x8F,
  0x0B19, 0x72,
  0x0B1A, 0x3E,
  0x0B1B, 0x20,
  0x0B1C, 0x0C,
  0x0B1D, 0x00,
  0x0B1E, 0x06,
  0x0B1F, 0x10,
  0x0B20, 0x43,
  0x0B21, 0x10,
  0x0B22, 0x0A,
  0x0B23, 0x12,
  0x0B24, 0xD8,
  0x0B25, 0x03,
  0x0B26, 0x28,
  0x0B27, 0xCD,
  0x0B28, 0xE0,
  0x0B29, 0x44,
  0x0B2A, 0x20,
  0x0B2B, 0xFF,
  0x0B2C, 0xE7,
  0x0B2D, 0xEF,
  0x0B2E, 0x25,
  0x0B2F, 0x00,
  0x0B30, 0x60,
  0x0B31, 0x38,
  0x0B32, 0x4F,
  0x0B33, 0x25,
  0x0B34, 0x07,
  0x0B35, 0x40,
  0x0B36, 0x7F,
  0x0B37, 0x20,
  0x0B38, 0x28,
  0x0B39, 0x05,
  0x0B3A, 0x0F,
  0x0B3B, 0x40,
  0x0B3C, 0x20,
  0x0B3D, 0x01,
  0x0B3E, 0xD9,
  0x0B3F, 0x00,
  0x0B40, 0x02,
  0x0B41, 0x52,
  0x0B42, 0x22,
  0x0B43, 0x01,
  0x0B44, 0x01,
  0x0B45, 0x13,
  0x0B46, 0x40,
  0x0B47, 0xC2,
  0x0B48, 0x60,
  0x0B49, 0x13,
  0x0B4A, 0x4A,
  0x0B4B, 0x20,
  0x0B4C, 0x1A,
  0x0B4D, 0x12,
  0x0B4E, 0x22,
  0x0B4F, 0x09,
  0x0B50, 0x02,
  0x0B51, 0x80,
  0x0B52, 0x20,
  0x0B53, 0x01,
  0x0B54, 0x4A,
  0x0B55, 0x1E,
  0x0B56, 0x40,
  0x0B57, 0xD0,
  0x0B58, 0x7F,
  0x0B59, 0x60,
  0x0B5A, 0x60,
  0x0B5B, 0x10,
  0x0B5C, 0x07,
  0x0B5D, 0x40,
  0x0B5E, 0x22,
  0x0B5F, 0x0A,
  0x0B60, 0x28,
  0x0B61, 0x04,
  0x0B62, 0x0F,
  0x0B63, 0x40,
  0x0B64, 0x20,
  0x0B65, 0x02,
  0x0B66, 0xD9,
  0x0B67, 0x00,
  0x0B68, 0x60,
  0x0B69, 0xB9,
  0x0B6A, 0x1A,
  0x0B6B, 0x10,
  0x0B6C, 0x60,
  0x0B6D, 0xF9,
  0x0B6E, 0x60,
  0x0B6F, 0x78,
  0x0B70, 0xF0,
  0x0B71, 0x00,
  0x0B72, 0x46,
  0x0B73, 0x28,
  0x0B74, 0x48,
  0x0B75, 0x15,
  0x0B76, 0xF8,
  0x0B77, 0xAB,
  0x0B78, 0x68,
  0x0B79, 0x00,
  0x0B7A, 0x49,
  0x0B7B, 0x16,
  0x0B7C, 0x42,
  0x0B7D, 0x82,
  0x0B7E, 0x68,
  0x0B7F, 0x0A,
  0x0B80, 0x46,
  0x0B81, 0x28,
  0x0B82, 0xD3,
  0x0B83, 0xFC,
  0x0B84, 0xF9,
  0x0B85, 0x04,
  0x0B86, 0xF0,
  0x0B87, 0x00,
  0x0B88, 0x46,
  0x0B89, 0x39,
  0x0B8A, 0x68,
  0x0B8B, 0xF8,
  0x0B8C, 0x28,
  0x0B8D, 0x00,
  0x0B8E, 0x4F,
  0x0B8F, 0x12,
  0x0B90, 0x68,
  0x0B91, 0x88,
  0x0B92, 0xD0,
  0x0B93, 0x13,
  0x0B94, 0x01,
  0x0B95, 0xC0,
  0x0B96, 0x68,
  0x0B97, 0xC9,
  0x0B98, 0xFB,
  0x0B99, 0xF4,
  0x0B9A, 0xF7,
  0x0B9B, 0xFF,
  0x0B9C, 0x1C,
  0x0B9D, 0x79,
  0x0B9E, 0x00,
  0x0B9F, 0x40,
  0x0BA0, 0xD3,
  0x0BA1, 0x00,
  0x0BA2, 0x42,
  0x0BA3, 0x88,
  0x0BA4, 0x06,
  0x0BA5, 0x01,
  0x0BA6, 0x46,
  0x0BA7, 0x38,
  0x0BA8, 0x0A,
  0x0BA9, 0x00,
  0x0BAA, 0x0C,
  0x0BAB, 0x09,
  0x0BAC, 0x87,
  0x0BAD, 0x31,
  0x0BAE, 0x43,
  0x0BAF, 0x01,
  0x0BB0, 0xD0,
  0x0BB1, 0x04,
  0x0BB2, 0x2D,
  0x0BB3, 0x00,
  0x0BB4, 0x77,
  0x0BB5, 0xA0,
  0x0BB6, 0x20,
  0x0BB7, 0x02,
  0x0BB8, 0x87,
  0x0BB9, 0x37,
  0x0BBA, 0xBD,
  0x0BBB, 0xF8,
  0x0BBC, 0x20,
  0x0BBD, 0x01,
  0x0BBE, 0xE7,
  0x0BBF, 0xB4,
  0x0BC0, 0x00,
  0x0BC1, 0x00,
  0x0BC2, 0xE7,
  0x0BC3, 0xF9,
  0x0BC4, 0x10,
  0x0BC5, 0x00,
  0x0BC6, 0x19,
  0x0BC7, 0x60,
  0x0BC8, 0x10,
  0x0BC9, 0x00,
  0x0BCA, 0x0E,
  0x0BCB, 0xD0,
  0x0BCC, 0x10,
  0x0BCD, 0x00,
  0x0BCE, 0x0E,
  0x0BCF, 0x7C,
  0x0BD0, 0x10,
  0x0BD1, 0x00,
  0x0BD2, 0x0E,
  0x0BD3, 0xEC,
  0x0BD4, 0x10,
  0x0BD5, 0x00,
  0x0BD6, 0x0E,
  0x0BD7, 0x78,
  0x0BD8, 0x00,
  0x0BD9, 0x00,
  0x0BDA, 0xFF,
  0x0BDB, 0xFF,
  0x0BDC, 0x48,
  0x0BDD, 0x24,
  0x0BDE, 0xB5,
  0x0BDF, 0xF0,
  0x0BE0, 0x4A,
  0x0BE1, 0x26,
  0x0BE2, 0x7C,
  0x0BE3, 0x01,
  0x0BE4, 0x4E,
  0x0BE5, 0x24,
  0x0BE6, 0x4D,
  0x0BE7, 0x23,
  0x0BE8, 0x69,
  0x0BE9, 0x13,
  0x0BEA, 0x68,
  0x0BEB, 0xD4,
  0x0BEC, 0xD0,
  0x0BED, 0x16,
  0x0BEE, 0x07,
  0x0BEF, 0xC8,
  0x0BF0, 0x42,
  0x0BF1, 0xA8,
  0x0BF2, 0x88,
  0x0BF3, 0x50,
  0x0BF4, 0x46,
  0x0BF5, 0x30,
  0x0BF6, 0xD1,
  0x0BF7, 0x01,
  0x0BF8, 0x00,
  0x0BF9, 0x40,
  0x0BFA, 0xE0,
  0x0BFB, 0x01,
  0x0BFC, 0x88,
  0x0BFD, 0xD7,
  0x0BFE, 0x5A,
  0x0BFF, 0x20,
  0x0C00, 0xD0,
  0x0C01, 0x01,
  0x0C02, 0x2F,
  0x0C03, 0x02,
  0x0C04, 0xD1,
  0x0C05, 0x00,
  0x0C06, 0x2F,
  0x0C07, 0x03,
  0x0C08, 0x43,
  0x0C09, 0x58,
  0x0C0A, 0x42,
  0x0C0B, 0x40,
  0x0C0C, 0x06,
  0x0C0D, 0x0F,
  0x0C0E, 0x11,
  0x0C0F, 0xC0,
  0x0C10, 0x11,
  0x0C11, 0x00,
  0x0C12, 0xD5,
  0x0C13, 0x01,
  0x0C14, 0x11,
  0x0C15, 0x40,
  0x0C16, 0xE0,
  0x0C17, 0x00,
  0x0C18, 0x19,
  0x0C19, 0xC0,
  0x0C1A, 0x69,
  0x0C1B, 0x57,
  0x0C1C, 0x69,
  0x0C1D, 0xD0,
  0x0C1E, 0xE0,
  0x0C1F, 0x00,
  0x0C20, 0xD5,
  0x0C21, 0x14,
  0x0C22, 0x07,
  0x0C23, 0x8F,
  0x0C24, 0x42,
  0x0C25, 0xAF,
  0x0C26, 0x88,
  0x0C27, 0x97,
  0x0C28, 0x00,
  0x0C29, 0x7D,
  0x0C2A, 0xD0,
  0x0C2B, 0x01,
  0x0C2C, 0x89,
  0x0C2D, 0x14,
  0x0C2E, 0x5B,
  0x0C2F, 0x66,
  0x0C30, 0xD0,
  0x0C31, 0x01,
  0x0C32, 0x2C,
  0x0C33, 0x02,
  0x0C34, 0xD1,
  0x0C35, 0x00,
  0x0C36, 0x2C,
  0x0C37, 0x03,
  0x0C38, 0x43,
  0x0C39, 0x5E,
  0x0C3A, 0x42,
  0x0C3B, 0x76,
  0x0C3C, 0x06,
  0x0C3D, 0x09,
  0x0C3E, 0x11,
  0x0C3F, 0xF3,
  0x0C40, 0x11,
  0x0C41, 0x19,
  0x0C42, 0xD5,
  0x0C43, 0x01,
  0x0C44, 0x11,
  0x0C45, 0x59,
  0x0C46, 0xE0,
  0x0C47, 0x00,
  0x0C48, 0x18,
  0x0C49, 0x89,
  0x0C4A, 0x69,
  0x0C4B, 0x92,
  0x0C4C, 0x6A,
  0x0C4D, 0x11,
  0x0C4E, 0xE0,
  0x0C4F, 0x00,
  0x0C50, 0xFB,
  0x0C51, 0x7C,
  0x0C52, 0xF7,
  0x0C53, 0xFF,
  0x0C54, 0x22,
  0x0C55, 0x01,
  0x0C56, 0x49,
  0x0C57, 0x09,
  0x0C58, 0x1F,
  0x0C59, 0x08,
  0x0C5A, 0x1D,
  0x0C5B, 0x89,
  0x0C5C, 0xFB,
  0x0C5D, 0x56,
  0x0C5E, 0xF7,
  0x0C5F, 0xFF,
  0x0C60, 0x22,
  0x0C61, 0x01,
  0x0C62, 0x49,
  0x0C63, 0x06,
  0x0C64, 0x1F,
  0x0C65, 0x08,
  0x0C66, 0x31,
  0x0C67, 0x08,
  0x0C68, 0xFB,
  0x0C69, 0x50,
  0x0C6A, 0xF7,
  0x0C6B, 0xFF,
  0x0C6C, 0x00,
  0x0C6D, 0x00,
  0x0C6E, 0xBD,
  0x0C6F, 0xF0,
  0x0C70, 0x10,
  0x0C71, 0x00,
  0x0C72, 0x1A,
  0x0C73, 0x60,
  0x0C74, 0x00,
  0x0C75, 0x00,
  0x0C76, 0x01,
  0x0C77, 0x00,
  0x0C78, 0x00,
  0x0C79, 0x00,
  0x0C7A, 0x7F,
  0x0C7B, 0xFF,
  0x0C7C, 0x10,
  0x0C7D, 0x00,
  0x0C7E, 0x0E,
  0x0C7F, 0xAC,
  0x0C80, 0x4A,
  0x0C81, 0x0E,
  0x0C82, 0xB5,
  0x0C83, 0x10,
  0x0C84, 0x1F,
  0x0C85, 0xC1,
  0x0C86, 0x88,
  0x0C87, 0x50,
  0x0C88, 0xD1,
  0x0C89, 0x01,
  0x0C8A, 0x39,
  0x0C8B, 0xF9,
  0x0C8C, 0xE0,
  0x0C8D, 0x02,
  0x0C8E, 0x48,
  0x0C8F, 0x0C,
  0x0C90, 0x00,
  0x0C91, 0x40,
  0x0C92, 0x68,
  0x0C93, 0x91,
  0x0C94, 0x88,
  0x0C95, 0x91,
  0x0C96, 0x5A,
  0x0C97, 0x08,
  0x0C98, 0xD0,
  0x0C99, 0x01,
  0x0C9A, 0x29,
  0x0C9B, 0x02,
  0x0C9C, 0xD1,
  0x0C9D, 0x00,
  0x0C9E, 0x29,
  0x0C9F, 0x03,
  0x0CA0, 0x49,
  0x0CA1, 0x08,
  0x0CA2, 0x42,
  0x0CA3, 0x40,
  0x0CA4, 0x40,
  0x0CA5, 0x88,
  0x0CA6, 0x68,
  0x0CA7, 0x09,
  0x0CA8, 0x62,
  0x0CA9, 0xC8,
  0x0CAA, 0x49,
  0x0CAB, 0x07,
  0x0CAC, 0x68,
  0x0CAD, 0xD2,
  0x0CAE, 0x49,
  0x0CAF, 0x03,
  0x0CB0, 0x1E,
  0x0CB1, 0x88,
  0x0CB2, 0x1D,
  0x0CB3, 0x09,
  0x0CB4, 0xFB,
  0x0CB5, 0x2A,
  0x0CB6, 0xF7,
  0x0CB7, 0xFF,
  0x0CB8, 0x00,
  0x0CB9, 0x00,
  0x0CBA, 0xBD,
  0x0CBB, 0x10,
  0x0CBC, 0x10,
  0x0CBD, 0x00,
  0x0CBE, 0x0E,
  0x0CBF, 0xE0,
  0x0CC0, 0x00,
  0x0CC1, 0x00,
  0x0CC2, 0x7F,
  0x0CC3, 0xFF,
  0x0CC4, 0x10,
  0x0CC5, 0x00,
  0x0CC6, 0x0E,
  0x0CC7, 0xD4,
  0x0CC8, 0x40,
  0x0CC9, 0x00,
  0x0CCA, 0x00,
  0x0CCB, 0x40,
  0x0CCC, 0x4B,
  0x0CCD, 0x27,
  0x0CCE, 0xB4,
  0x0CCF, 0x30,
  0x0CD0, 0x29,
  0x0CD1, 0x01,
  0x0CD2, 0x78,
  0x0CD3, 0x19,
  0x0CD4, 0x21,
  0x0CD5, 0x19,
  0x0CD6, 0xD0,
  0x0CD7, 0x48,
  0x0CD8, 0x68,
  0x0CD9, 0x0A,
  0x0CDA, 0x06,
  0x0CDB, 0xC9,
  0x0CDC, 0x03,
  0x0CDD, 0x24,
  0x0CDE, 0x24,
  0x0CDF, 0x01,
  0x0CE0, 0x60,
  0x0CE1, 0x0A,
  0x0CE2, 0x43,
  0x0CE3, 0xA2,
  0x0CE4, 0x80,
  0x0CE5, 0x5C,
  0x0CE6, 0x24,
  0x0CE7, 0x00,
  0x0CE8, 0x00,
  0x0CE9, 0xC9,
  0x0CEA, 0x80,
  0x0CEB, 0x9C,
  0x0CEC, 0x2A,
  0x0CED, 0x00,
  0x0CEE, 0x6A,
  0x0CEF, 0x4A,
  0x0CF0, 0x6A,
  0x0CF1, 0x8A,
  0x0CF2, 0xD1,
  0x0CF3, 0xFC,
  0x0CF4, 0xD1,
  0x0CF5, 0xFC,
  0x0CF6, 0x2A,
  0x0CF7, 0x00,
  0x0CF8, 0x4D,
  0x0CF9, 0x1D,
  0x0CFA, 0x6A,
  0x0CFB, 0x0A,
  0x0CFC, 0x62,
  0x0CFD, 0x0A,
  0x0CFE, 0x43,
  0x0CFF, 0x2A,
  0x0D00, 0xD0,
  0x0D01, 0x29,
  0x0D02, 0x28,
  0x0D03, 0x01,
  0x0D04, 0x4D,
  0x0D05, 0x1C,
  0x0D06, 0x4A,
  0x0D07, 0x1B,
  0x0D08, 0x6A,
  0x0D09, 0x4A,
  0x0D0A, 0x60,
  0x0D0B, 0x6A,
  0x0D0C, 0x43,
  0x0D0D, 0x2A,
  0x0D0E, 0x25,
  0x0D0F, 0x20,
  0x0D10, 0x28,
  0x0D11, 0x01,
  0x0D12, 0x62,
  0x0D13, 0x4A,
  0x0D14, 0x4A,
  0x0D15, 0x17,
  0x0D16, 0xD0,
  0x0D17, 0x22,
  0x0D18, 0x4D,
  0x0D19, 0x17,
  0x0D1A, 0x1F,
  0x0D1B, 0x12,
  0x0D1C, 0x61,
  0x0D1D, 0x2A,
  0x0D1E, 0x35,
  0x0D1F, 0x40,
  0x0D20, 0xD0,
  0x0D21, 0x1E,
  0x0D22, 0x28,
  0x0D23, 0x01,
  0x0D24, 0x1D,
  0x0D25, 0x00,
  0x0D26, 0x48,
  0x0D27, 0x13,
  0x0D28, 0x6A,
  0x0D29, 0x88,
  0x0D2A, 0x61,
  0x0D2B, 0x68,
  0x0D2C, 0x43,
  0x0D2D, 0x10,
  0x0D2E, 0x22,
  0x0D2F, 0x30,
  0x0D30, 0x48,
  0x0D31, 0x12,
  0x0D32, 0x62,
  0x0D33, 0x88,
  0x0D34, 0x20,
  0x0D35, 0x01,
  0x0D36, 0x60,
  0x0D37, 0x04,
  0x0D38, 0x49,
  0x0D39, 0x11,
  0x0D3A, 0x70,
  0x0D3B, 0x18,
  0x0D3C, 0x60,
  0x0D3D, 0x08,
  0x0D3E, 0x20,
  0x0D3F, 0x02,
  0x0D40, 0x06,
  0x0D41, 0x40,
  0x0D42, 0x20,
  0x0D43, 0x63,
  0x0D44, 0x22,
  0x0D45, 0x0C,
  0x0D46, 0x6A,
  0x0D47, 0x81,
  0x0D48, 0x62,
  0x0D49, 0x81,
  0x0D4A, 0x43,
  0x0D4B, 0x91,
  0x0D4C, 0x68,
  0x0D4D, 0x00,
  0x0D4E, 0x48,
  0x0D4F, 0x0D,
  0x0D50, 0x68,
  0x0D51, 0x09,
  0x0D52, 0x49,
  0x0D53, 0x0D,
  0x0D54, 0x47,
  0x0D55, 0x08,
  0x0D56, 0xBC,
  0x0D57, 0x30,
  0x0D58, 0xE7,
  0x0D59, 0xD4,
  0x0D5A, 0x4A,
  0x0D5B, 0x0C,
  0x0D5C, 0x1F,
  0x0D5D, 0x12,
  0x0D5E, 0x4A,
  0x0D5F, 0x0B,
  0x0D60, 0x48,
  0x0D61, 0x0A,
  0x0D62, 0xE7,
  0x0D63, 0xDB,
  0x0D64, 0xE7,
  0x0D65, 0xDF,
  0x0D66, 0x1D,
  0x0D67, 0x00,
  0x0D68, 0x47,
  0x0D69, 0x70,
  0x0D6A, 0xBC,
  0x0D6B, 0x30,
  0x0D6C, 0x10,
  0x0D6D, 0x00,
  0x0D6E, 0x0E,
  0x0D6F, 0xE0,
  0x0D70, 0x00,
  0x0D71, 0x00,
  0x0D72, 0xC0,
  0x0D73, 0x20,
  0x0D74, 0x00,
  0x0D75, 0x00,
  0x0D76, 0x41,
  0x0D77, 0x08,
  0x0D78, 0x40,
  0x0D79, 0x00,
  0x0D7A, 0x00,
  0x0D7B, 0x40,
  0x0D7C, 0x10,
  0x0D7D, 0x00,
  0x0D7E, 0x0E,
  0x0D7F, 0x78,
  0x0D80, 0xE0,
  0x0D81, 0x00,
  0x0D82, 0xE1,
  0x0D83, 0x00,
  0x0D84, 0x10,
  0x0D85, 0x00,
  0x0D86, 0x0E,
  0x0D87, 0xD0,
  0x0D88, 0x10,
  0x0D89, 0x00,
  0x0D8A, 0x0E,
  0x0D8B, 0x5C,
  0x0D8C, 0x00,
  0x0D8D, 0x00,
  0x0D8E, 0x51,
  0x0D8F, 0x08,
  0x0D90, 0x46,
  0x0D91, 0x04,
  0x0D92, 0xB5,
  0x0D93, 0x1C,
  0x0D94, 0x68,
  0x0D95, 0x00,
  0x0D96, 0x48,
  0x0D97, 0x19,
  0x0D98, 0x20,
  0x0D99, 0x63,
  0x0D9A, 0x47,
  0x0D9B, 0x80,
  0x0D9C, 0x6A,
  0x0D9D, 0x81,
  0x0D9E, 0x06,
  0x0D9F, 0x40,
  0x0DA0, 0x43,
  0x0DA1, 0x91,
  0x0DA2, 0x22,
  0x0DA3, 0x0C,
  0x0DA4, 0x62,
  0x0DA5, 0x81,
  0x0DA6, 0x1D,
  0x0DA7, 0x09,
  0x0DA8, 0x20,
  0x0DA9, 0x02,
  0x0DAA, 0x49,
  0x0DAB, 0x15,
  0x0DAC, 0x48,
  0x0DAD, 0x15,
  0x0DAE, 0x60,
  0x0DAF, 0x08,
  0x0DB0, 0x70,
  0x0DB1, 0x02,
  0x0DB2, 0x22,
  0x0DB3, 0x00,
  0x0DB4, 0x07,
  0x0DB5, 0x89,
  0x0DB6, 0x21,
  0x0DB7, 0x01,
  0x0DB8, 0x23,
  0x0DB9, 0x20,
  0x0DBA, 0x6A,
  0x0DBB, 0x48,
  0x0DBC, 0x62,
  0x0DBD, 0x48,
  0x0DBE, 0x43,
  0x0DBF, 0x98,
  0x0DC0, 0x23,
  0x0DC1, 0x30,
  0x0DC2, 0x6A,
  0x0DC3, 0x88,
  0x0DC4, 0x62,
  0x0DC5, 0x88,
  0x0DC6, 0x43,
  0x0DC7, 0x98,
  0x0DC8, 0x4B,
  0x0DC9, 0x0F,
  0x0DCA, 0x6A,
  0x0DCB, 0x08,
  0x0DCC, 0x62,
  0x0DCD, 0x08,
  0x0DCE, 0x40,
  0x0DCF, 0x18,
  0x0DD0, 0x06,
  0x0DD1, 0xC9,
  0x0DD2, 0x21,
  0x0DD3, 0x19,
  0x0DD4, 0x23,
  0x0DD5, 0x01,
  0x0DD6, 0x68,
  0x0DD7, 0x08,
  0x0DD8, 0x43,
  0x0DD9, 0x18,
  0x0DDA, 0x03,
  0x0DDB, 0x1B,
  0x0DDC, 0x2C,
  0x0DDD, 0x01,
  0x0DDE, 0x60,
  0x0DDF, 0x08,
  0x0DE0, 0x48,
  0x0DE1, 0x0A,
  0x0DE2, 0xD0,
  0x0DE3, 0x09,
  0x0DE4, 0x92,
  0x0DE5, 0x01,
  0x0DE6, 0x4B,
  0x0DE7, 0x0A,
  0x0DE8, 0x68,
  0x0DE9, 0x1B,
  0x0DEA, 0x90,
  0x0DEB, 0x00,
  0x0DEC, 0xA9,
  0x0DED, 0x01,
  0x0DEE, 0x22,
  0x0DEF, 0x01,
  0x0DF0, 0x47,
  0x0DF1, 0x98,
  0x0DF2, 0x46,
  0x0DF3, 0x68,
  0x0DF4, 0x48,
  0x0DF5, 0x07,
  0x0DF6, 0xBD,
  0x0DF7, 0x1C,
  0x0DF8, 0x00,
  0x0DF9, 0x00,
  0x0DFA, 0xE7,
  0x0DFB, 0xF4,
  0x0DFC, 0x10,
  0x0DFD, 0x00,
  0x0DFE, 0x0E,
  0x0DFF, 0x60,
  0x0E00, 0xE0,
  0x0E01, 0x00,
  0x0E02, 0xE1,
  0x0E03, 0x80,
  0x0E04, 0x10,
  0x0E05, 0x00,
  0x0E06, 0x0E,
  0x0E07, 0xE0,
  0x0E08, 0xFF,
  0x0E09, 0xFF,
  0x0E0A, 0x3F,
  0x0E0B, 0xDF,
  0x0E0C, 0x00,
  0x0E0D, 0x00,
  0x0E0E, 0x41,
  0x0E0F, 0x08,
  0x0E10, 0x10,
  0x0E11, 0x00,
  0x0E12, 0x0E,
  0x0E13, 0x40,
  0x0E14, 0x00,
  0x0E15, 0x00,
  0x0E16, 0x51,
  0x0E17, 0x08,
  0x0E18, 0x00,
  0x0E19, 0x00,
  0x0E1A, 0x06,
  0x0E1B, 0x6F,
  0x0E1C, 0x00,
  0x0E1D, 0x00,
  0x0E1E, 0x03,
  0x0E1F, 0x37,
  0x0E20, 0x00,
  0x0E21, 0x00,
  0x0E22, 0x02,
  0x0E23, 0x23,
  0x0E24, 0x00,
  0x0E25, 0x00,
  0x0E26, 0x01,
  0x0E27, 0x9B,
  0x0E28, 0x00,
  0x0E29, 0x00,
  0x0E2A, 0x01,
  0x0E2B, 0x48,
  0x0E2C, 0x00,
  0x0E2D, 0x00,
  0x0E2E, 0x01,
  0x0E2F, 0x10,
  0x0E30, 0x00,
  0x0E31, 0x00,
  0x0E32, 0x00,
  0x0E33, 0xE9,
  0x0E34, 0x00,
  0x0E35, 0x00,
  0x0E36, 0x00,
  0x0E37, 0xCD,
  0x0E38, 0x00,
  0x0E39, 0x00,
  0x0E3A, 0x00,
  0x0E3B, 0xB5,
  0x0E3C, 0x00,
  0x0E3D, 0x00,
  0x0E3E, 0x00,
  0x0E3F, 0xA3,
  0x0E40, 0x10,
  0x0E41, 0x00,
  0x0E42, 0x02,
  0x0E43, 0x51,
  0x0E44, 0x10,
  0x0E45, 0x00,
  0x0E46, 0x02,
  0x0E47, 0x51,
  0x0E48, 0x10,
  0x0E49, 0x00,
  0x0E4A, 0x02,
  0x0E4B, 0x51,
  0x0E4C, 0x10,
  0x0E4D, 0x00,
  0x0E4E, 0x02,
  0x0E4F, 0x51,
  0x0E50, 0x10,
  0x0E51, 0x00,
  0x0E52, 0x02,
  0x0E53, 0x51,
  0x0E54, 0x10,
  0x0E55, 0x00,
  0x0E56, 0x02,
  0x0E57, 0x51,
  0x0E58, 0x10,
  0x0E59, 0x00,
  0x0E5A, 0x02,
  0x0E5B, 0x51,
  0x0E5C, 0x10,
  0x0E5D, 0x00,
  0x0E5E, 0x02,
  0x0E5F, 0x51,
  0x0E60, 0x10,
  0x0E61, 0x00,
  0x0E62, 0x02,
  0x0E63, 0x51,
  0x0E64, 0x10,
  0x0E65, 0x00,
  0x0E66, 0x02,
  0x0E67, 0x51,
  0x0E68, 0x10,
  0x0E69, 0x00,
  0x0E6A, 0x02,
  0x0E6B, 0x51,
  0x0E6C, 0x10,
  0x0E6D, 0x00,
  0x0E6E, 0x02,
  0x0E6F, 0x51,
  0x0E70, 0x10,
  0x0E71, 0x00,
  0x0E72, 0x02,
  0x0E73, 0x51,
  0x0E74, 0x00,
  0x0E75, 0x00,
  0x0E76, 0x00,
  0x0E77, 0x00,
  0x0E78, 0x00,
  0x0E79, 0x00,
  0x0E7A, 0x00,
  0x0E7B, 0x00,
  0x0E7C, 0x00,
  0x0E7D, 0x00,
  0x0E7E, 0x00,
  0x0E7F, 0x00,
  0x0E80, 0x00,
  0x0E81, 0x00,
  0x0E82, 0x00,
  0x0E83, 0x00,
  0x0E84, 0x00,
  0x0E85, 0x00,
  0x0E86, 0x00,
  0x0E87, 0x00,
  0x0E88, 0x00,
  0x0E89, 0x00,
  0x0E8A, 0x00,
  0x0E8B, 0x00,
  0x0E8C, 0x00,
  0x0E8D, 0x00,
  0x0E8E, 0x00,
  0x0E8F, 0x00,
  0x0E90, 0x00,
  0x0E91, 0x00,
  0x0E92, 0x00,
  0x0E93, 0x00,
  0x0E94, 0x00,
  0x0E95, 0x00,
  0x0E96, 0x00,
  0x0E97, 0x00,
  0x0E98, 0x00,
  0x0E99, 0x00,
  0x0E9A, 0x00,
  0x0E9B, 0x00,
  0x0E9C, 0x00,
  0x0E9D, 0x00,
  0x0E9E, 0x00,
  0x0E9F, 0x00,
  0x0EA0, 0x00,
  0x0EA1, 0x00,
  0x0EA2, 0x00,
  0x0EA3, 0x00,
  0x0EA4, 0x00,
  0x0EA5, 0x00,
  0x0EA6, 0x00,
  0x0EA7, 0x00,
  0x0EA8, 0x00,
  0x0EA9, 0x00,
  0x0EAA, 0x00,
  0x0EAB, 0x00,
  0x0EAC, 0x00,
  0x0EAD, 0x00,
  0x0EAE, 0x00,
  0x0EAF, 0x00,
  0x0EB0, 0x00,
  0x0EB1, 0x00,
  0x0EB2, 0x00,
  0x0EB3, 0x00,
  0x0EB4, 0x00,
  0x0EB5, 0x00,
  0x0EB6, 0x00,
  0x0EB7, 0x00,
  0x0EB8, 0x00,
  0x0EB9, 0x00,
  0x0EBA, 0x5E,
  0x0EBB, 0x00,
  0x0EBC, 0x00,
  0x0EBD, 0x00,
  0x0EBE, 0x00,
  0x0EBF, 0x00,
  0x0EC0, 0x00,
  0x0EC1, 0x00,
  0x0EC2, 0x00,
  0x0EC3, 0x00,
  0x0EC4, 0x00,
  0x0EC5, 0x00,
  0x0EC6, 0x00,
  0x0EC7, 0x00,
  0x0EC8, 0x00,
  0x0EC9, 0x00,
  0x0ECA, 0x00,
  0x0ECB, 0x00,
  0x0ECC, 0x00,
  0x0ECD, 0x00,
  0x0ECE, 0x00,
  0x0ECF, 0x00,
  0x0ED0, 0x00,
  0x0ED1, 0x00,
  0x0ED2, 0x00,
  0x0ED3, 0x00,
  0x0ED4, 0x00,
  0x0ED5, 0x00,
  0x0ED6, 0x00,
  0x0ED7, 0x00,
  0x0ED8, 0x00,
  0x0ED9, 0x00,
  0x0EDA, 0x00,
  0x0EDB, 0x00,
  0x0EDC, 0x00,
  0x0EDD, 0x00,
  0x0EDE, 0x00,
  0x0EDF, 0x00,
  0x0EE0, 0x00,
  0x0EE1, 0x00,
  0x0EE2, 0x00,
  0x0EE3, 0x00,
  0x0EE4, 0x00,
  0x0EE5, 0x00,
  0x0EE6, 0x00,
  0x0EE7, 0x00,
  0x0EE8, 0x00,
  0x0EE9, 0x00,
  0x0EEA, 0x5E,
  0x0EEB, 0x00,
  0x0EEC, 0x00,
  0x0EED, 0x00,
  0x0EEE, 0x00,
  0x0EEF, 0x00,
  0xffff, 0xff,
};
const T_I2CRegWriteData SensorFocusWnd::i2cdat_SineTest_dp2_[] = {
  0x1C00, 0x00,
  0x1C01, 0x00,
  0x1C02, 0x77,
  0x1C03, 0x7E,
  0x1C04, 0x02,
  0x1C05, 0x0F,
  0x1C06, 0x24,
  0x1C07, 0x1B,
  0x1C08, 0x00,
  0x1C09, 0x01,
  0x1C0A, 0x00,
  0x1C0B, 0x00,
  0x1C0C, 0x00,
  0x1C0D, 0x22,
  0x1C0E, 0x04,
  0x1C0F, 0x0B,
  0x1C10, 0x80,
  0x1C11, 0x45,
  0x1C12, 0x00,
  0x1C13, 0x00,
  0x1C14, 0x00,
  0x1C15, 0x00,
  0x1C16, 0x02,
  0x1C17, 0x14,
  0x1C18, 0x00,
  0x1C19, 0x01,
  0x1C1A, 0x00,
  0x1C1B, 0x01,
  0x1C1C, 0x08,
  0x1C1D, 0x00,
  0x1C1E, 0x00,
  0x1C1F, 0x02,
  0x1C20, 0x10,
  0x1C21, 0x10,
  0x1C22, 0x10,
  0x1C23, 0x03,
  0x1C24, 0x0C,
  0x1C25, 0x00,
  0x1C26, 0x0C,
  0x1C27, 0x00,
  0x1C28, 0x04,
  0x1C29, 0x00,
  0x1C2A, 0x04,
  0x1C2B, 0x00,
  0x1C2C, 0x0C,
  0x1C2D, 0x30,
  0x1C2E, 0x01,
  0x1C2F, 0x12,
  0x1C30, 0x00,
  0x1C31, 0xF0,
  0x1C32, 0x22,
  0x1C33, 0x71,
  0x1C34, 0x02,
  0x1C35, 0x27,
  0x1C36, 0x00,
  0x1C37, 0xF0,
  0x1C38, 0x83,
  0x1C39, 0xC0,
  0x1C3A, 0x05,
  0x1C3B, 0x00,
  0x1C3C, 0x02,
  0x1C3D, 0x32,
  0x1C3E, 0x00,
  0x1C3F, 0x02,
  0x1C40, 0x00,
  0x1C41, 0xC0,
  0x1C42, 0x00,
  0x1C43, 0x00,
  0x1C44, 0x00,
  0x1C45, 0x20,
  0x1C46, 0x00,
  0x1C47, 0x00,
  0x1C48, 0xD0,
  0x1C49, 0x00,
  0x1C4A, 0xD0,
  0x1C4B, 0x00,
  0x1C4C, 0xD0,
  0x1C4D, 0x00,
  0x1C4E, 0xD0,
  0x1C4F, 0x00,
  0x1C50, 0x50,
  0x1C51, 0x50,
  0x1C52, 0x50,
  0x1C53, 0x50,
  0x1C54, 0x3F,
  0x1C55, 0x3F,
  0x1C56, 0x07,
  0x1C57, 0x80,
  0x1C58, 0x04,
  0x1C59, 0x00,
  0x1C5A, 0x04,
  0x1C5B, 0x00,
  0x1C5C, 0x00,
  0x1C5D, 0x00,
  0x1C5E, 0x00,
  0x1C5F, 0x00,
  0x1C60, 0x20,
  0x1C61, 0x00,
  0x1C62, 0x20,
  0x1C63, 0x00,
  0x1C64, 0x04,
  0x1C65, 0x00,
  0x1C66, 0x04,
  0x1C67, 0x00,
  0x1C68, 0x00,
  0x1C69, 0x04,
  0x1C6A, 0x00,
  0x1C6B, 0x04,
  0x1C6C, 0x00,
  0x1C6D, 0x00,
  0x1C6E, 0x00,
  0x1C6F, 0x04,
  0x1C70, 0x00,
  0x1C71, 0x02,
  0x1C72, 0x02,
  0x1C73, 0x00,
  0x1C74, 0x83,
  0x1C75, 0x91,
  0x1C76, 0x3C,
  0x1C77, 0x6F,
  0x1C78, 0x83,
  0x1C79, 0x9A,
  0x1C7A, 0x3C,
  0x1C7B, 0x66,
  0x1C7C, 0x3F,
  0x1C7D, 0xFE,
  0x1C7E, 0x3F,
  0x1C7F, 0xFC,
  0x1C80, 0x3F,
  0x1C81, 0xFF,
  0x1C82, 0x3F,
  0x1C83, 0xFF,
  0x1C84, 0x40,
  0x1C85, 0x00,
  0x1C86, 0x00,
  0x1C87, 0x00,
  0x1C88, 0x00,
  0x1C89, 0x00,
  0x1C8A, 0x00,
  0x1C8B, 0x00,
  0x1C8C, 0x3E,
  0x1C8D, 0xFC,
  0x1C8E, 0xC1,
  0x1C8F, 0x0A,
  0x1C90, 0x7F,
  0x1C91, 0xF8,
  0x1C92, 0x7F,
  0x1C93, 0xF8,
  0x1C94, 0x1A,
  0x1C95, 0x5E,
  0x1C96, 0x01,
  0x1C97, 0xA5,
  0x1C98, 0x05,
  0x1C99, 0xDE,
  0x1C9A, 0x01,
  0x1C9B, 0x7A,
  0x1C9C, 0x01,
  0x1C9D, 0x0E,
  0x1C9E, 0x28,
  0x1C9F, 0x00,
  0x1CA0, 0x00,
  0x1CA1, 0x02,
  0x1CA2, 0x04,
  0x1CA3, 0x65,
  0x1CA4, 0x01,
  0x1CA5, 0x51,
  0x1CA6, 0x01,
  0x1CA7, 0x00,
  0x1CA8, 0x3F,
  0x1CA9, 0xF2,
  0x1CAA, 0x3F,
  0x1CAB, 0xF2,
  0x1CAC, 0x3F,
  0x1CAD, 0xE0,
  0x1CAE, 0x7F,
  0x1CAF, 0xFF,
  0x1CB0, 0x05,
  0x1CB1, 0xDE,
  0x1CB2, 0x04,
  0x1CB3, 0x65,
  0x1CB4, 0x01,
  0x1CB5, 0x00,
  0x1CB6, 0x11,
  0x1CB7, 0x9A,
  0x1CB8, 0x50,
  0x1CB9, 0x46,
  0x1CBA, 0xB0,
  0x1CBB, 0x9E,
  0x1CBC, 0x7F,
  0x1CBD, 0x74,
  0x1CBE, 0x13,
  0x1CBF, 0x26,
  0x1CC0, 0xDD,
  0x1CC1, 0x45,
  0x1CC2, 0x10,
  0x1CC3, 0x1C,
  0x1CC4, 0x7C,
  0x1CC5, 0xD5,
  0x1CC6, 0x85,
  0x1CC7, 0xD0,
  0x1CC8, 0x21,
  0x1CC9, 0x87,
  0x1CCA, 0xBE,
  0x1CCB, 0x1E,
  0x1CCC, 0x20,
  0x1CCD, 0x65,
  0x1CCE, 0x7E,
  0x1CCF, 0xDA,
  0x1CD0, 0x82,
  0x1CD1, 0x38,
  0x1CD2, 0x50,
  0x1CD3, 0x00,
  0x1CD4, 0x30,
  0x1CD5, 0x00,
  0x1CD6, 0x30,
  0x1CD7, 0x00,
  0x1CD8, 0x30,
  0x1CD9, 0x00,
  0x1CDA, 0x30,
  0x1CDB, 0x00,
  0x1CDC, 0x00,
  0x1CDD, 0x00,
  0x1CDE, 0x00,
  0x1CDF, 0x00,
  0x1CE0, 0x00,
  0x1CE1, 0x00,
  0x1CE2, 0x00,
  0x1CE3, 0x00,
  0x1CE4, 0x00,
  0x1CE5, 0x00,
  0x1CE6, 0x00,
  0x1CE7, 0x00,
  0x1CE8, 0x10,
  0x1CE9, 0x00,
  0x1CEA, 0x10,
  0x1CEB, 0x00,
  0x1CEC, 0x01,
  0x1CED, 0x18,
  0x1CEE, 0x10,
  0x1CEF, 0x00,
  0x1CF0, 0x00,
  0x1CF1, 0x30,
  0x1CF2, 0x00,
  0x1CF3, 0x40,
  0x1CF4, 0x00,
  0x1CF5, 0x05,
  0x1CF6, 0x00,
  0x1CF7, 0x02,
  0x1CF8, 0x00,
  0x1CF9, 0x00,
  0x1CFA, 0x07,
  0x1CFB, 0xC0,
  0x1CFC, 0x00,
  0x1CFD, 0x01,
  0x1CFE, 0x00,
  0x1CFF, 0x00,
  0x1D00, 0xB4,
  0x1D01, 0x70,
  0x1D02, 0x12,
  0x1D03, 0x00,
  0x1D04, 0x07,
  0x1D05, 0x80,
  0x1D06, 0x01,
  0x1D07, 0xFF,
  0x1D08, 0x00,
  0x1D09, 0x00,
  0x1D0A, 0x18,
  0x1D0B, 0x00,
  0x1D0C, 0x18,
  0x1D0D, 0x00,
  0x1D0E, 0x50,
  0x1D0F, 0x00,
  0x1D10, 0x40,
  0x1D11, 0x00,
  0x1D12, 0xC0,
  0x1D13, 0x00,
  0x1D14, 0x3C,
  0x1D15, 0x00,
  0x1D16, 0x40,
  0x1D17, 0x00,
  0x1D18, 0x00,
  0x1D19, 0x00,
  0x1D1A, 0x40,
  0x1D1B, 0x00,
  0x1D1C, 0x06,
  0x1D1D, 0x78,
  0x1D1E, 0x06,
  0x1D1F, 0x78,
  0x1D20, 0x00,
  0x1D21, 0x00,
  0x1D22, 0x33,
  0x1D23, 0x18,
  0x1D24, 0x00,
  0x1D25, 0x00,
  0x1D26, 0xF9,
  0x1D27, 0x88,
  0x1D28, 0xF9,
  0x1D29, 0x88,
  0x1D2A, 0x00,
  0x1D2B, 0x00,
  0x1D2C, 0x33,
  0x1D2D, 0x18,
  0x1D2E, 0x00,
  0x1D2F, 0x00,
  0x1D30, 0x7F,
  0x1D31, 0xFF,
  0x1D32, 0x00,
  0x1D33, 0x00,
  0x1D34, 0x00,
  0x1D35, 0x00,
  0x1D36, 0xC0,
  0x1D37, 0x00,
  0x1D38, 0x00,
  0x1D39, 0x00,
  0x1D3A, 0x00,
  0x1D3B, 0x00,
  0x1D3C, 0x00,
  0x1D3D, 0x00,
  0x1D3E, 0x00,
  0x1D3F, 0x00,
  0x1D40, 0x00,
  0x1D41, 0x00,
  0x1D42, 0x00,
  0x1D43, 0x00,
  0x1D44, 0x00,
  0x1D45, 0x00,
  0x1D46, 0x00,
  0x1D47, 0x00,
  0x1D48, 0x00,
  0x1D49, 0x00,
  0x1D4A, 0x00,
  0x1D4B, 0x00,
  0x1D4C, 0x00,
  0x1D4D, 0x00,
  0x1D4E, 0x00,
  0x1D4F, 0x00,
  0x1D50, 0x00,
  0x1D51, 0x00,
  0x1D52, 0x00,
  0x1D53, 0x00,
  0x1D54, 0x00,
  0x1D55, 0x00,
  0x1D56, 0x00,
  0x1D57, 0x00,
  0x1D58, 0x00,
  0x1D59, 0x00,
  0x1D5A, 0x00,
  0x1D5B, 0x00,
  0x1D5C, 0x00,
  0x1D5D, 0x00,
  0x1D5E, 0x00,
  0x1D5F, 0x00,
  0x1D60, 0x00,
  0x1D61, 0x00,
  0x1D62, 0x00,
  0x1D63, 0x00,
  0x1D64, 0x03,
  0x1D65, 0x6D,
  0x1D66, 0x01,
  0x1D67, 0x23,
  0x1D68, 0x06,
  0x1D69, 0x54,
  0x1D6A, 0x05,
  0x1D6B, 0x42,
  0x1D6C, 0x24,
  0x1D6D, 0x01,
  0x1D6E, 0x1E,
  0x1D6F, 0x12,
  0x1D70, 0x01,
  0x1D71, 0x01,
  0x1D72, 0x18,
  0x1D73, 0x14,
  0x1D74, 0x01,
  0x1D75, 0x01,
  0x1D76, 0x18,
  0x1D77, 0x14,
  0x1D78, 0x0F,
  0x1D79, 0x0F,
  0x1D7A, 0x18,
  0x1D7B, 0x10,
  0x1D7C, 0x0F,
  0x1D7D, 0x0F,
  0x1D7E, 0x18,
  0x1D7F, 0x10,
  0x1D80, 0x00,
  0x1D81, 0x01,
  0x1D82, 0x00,
  0x1D83, 0x05,
  0x1D84, 0x00,
  0x1D85, 0x01,
  0x1D86, 0x00,
  0x1D87, 0x05,
  0x1D88, 0x00,
  0x1D89, 0x01,
  0x1D8A, 0x00,
  0x1D8B, 0x02,
  0x1D8C, 0x00,
  0x1D8D, 0x01,
  0x1D8E, 0x00,
  0x1D8F, 0x05,
  0x1D90, 0x00,
  0x1D91, 0x00,
  0x1D92, 0x01,
  0x1D93, 0x84,
  0x1D94, 0xE0,
  0x1D95, 0xE0,
  0x1D96, 0x00,
  0x1D97, 0x80,
  0x1D98, 0x00,
  0x1D99, 0xC3,
  0x1D9A, 0xC3,
  0x1D9B, 0x00,
  0x1D9C, 0x00,
  0x1D9D, 0x00,
  0x1D9E, 0x00,
  0x1D9F, 0x00,
  0x1DA0, 0x00,
  0x1DA1, 0x00,
  0x1DA2, 0x00,
  0x1DA3, 0x00,
  0x1DA4, 0x00,
  0x1DA5, 0x00,
  0x1DA6, 0x00,
  0x1DA7, 0x00,
  0x1DA8, 0x00,
  0x1DA9, 0x00,
  0x1DAA, 0x00,
  0x1DAB, 0x00,
  0x1DAC, 0x00,
  0x1DAD, 0x10,
  0x1DAE, 0x00,
  0x1DAF, 0x00,
  0x1DB0, 0x00,
  0x1DB1, 0x00,
  0x1DB2, 0x00,
  0x1DB3, 0x00,
  0x1DB4, 0x7F,
  0x1DB5, 0xF4,
  0x1DB6, 0x41,
  0x1DB7, 0xEB,
  0x1DB8, 0x00,
  0x1DB9, 0x00,
  0x1DBA, 0x00,
  0x1DBB, 0x00,
  0x1DBC, 0x00,
  0x1DBD, 0x00,
  0x1DBE, 0x00,
  0x1DBF, 0x00,
  0xffff, 0xff,
};
